//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



#include "StdAfx.h"
#include "ClientCredits.h"
#include "opcodes.h"
#include "otherfunctions.h"
#include "emule.h"
#include "safefile.h"
#include <math.h>
#include "crypto51/base64.h"
#include "crypto51/osrng.h"
#include "crypto51/files.h"
#include "crypto51/sha.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define CLIENTS_MET_FILENAME	_T("clients.met")
#define ELEMENT_COUNT(X) (sizeof(X) / sizeof(X[0]))


CClientCredits::CClientCredits(CreditStruct* in_credits)
{
	m_pCredits = in_credits;
	InitalizeIdent();
	m_dwUnSecureWaitTime = 0;
	m_dwSecureWaitTime = 0;
	m_dwWaitTimeIP = 0;
}

CClientCredits::CClientCredits(const uchar* key)
{
	m_pCredits = new CreditStruct;
	memzero(m_pCredits, sizeof(CreditStruct));
	md4cpy(m_pCredits->abyKey, key);
	InitalizeIdent();
	m_dwUnSecureWaitTime = ::GetTickCount();
	m_dwSecureWaitTime = ::GetTickCount();
	m_dwWaitTimeIP = 0;
}

CClientCredits::~CClientCredits()
{
	delete m_pCredits;
}

void CClientCredits::AddDownloaded(uint32 dwBytes, uint32 dwForIP)
{
	EIdentState		eIdentState = GetCurrentIdentState(dwForIP);

	if ( ( (eIdentState != IS_IDFAILED) && (eIdentState != IS_IDBADGUY) &&
		(eIdentState != IS_IDNEEDED) ) || !g_eMuleApp.m_pClientCreditList->CryptoAvailable() )
	{
		uint32		dwLow = m_pCredits->nDownloadedLo;

		m_pCredits->nDownloadedLo += dwBytes;
		if (m_pCredits->nDownloadedLo < dwLow)
			m_pCredits->nDownloadedHi++;
	}
}

void CClientCredits::AddUploaded(uint32 dwBytes, uint32 dwForIP)
{
	EIdentState		eIdentState = GetCurrentIdentState(dwForIP);

	if ( ( (eIdentState != IS_IDFAILED) && (eIdentState != IS_IDBADGUY) &&
		(eIdentState != IS_IDNEEDED) ) || !g_eMuleApp.m_pClientCreditList->CryptoAvailable() )
	{
		uint32		dwLow = m_pCredits->nUploadedLo;

		m_pCredits->nUploadedLo += dwBytes;
		if (m_pCredits->nUploadedLo < dwLow)
			m_pCredits->nUploadedHi++;
	}
}

uint64	CClientCredits::GetUploadedTotal(){
	return ( (uint64)m_pCredits->nUploadedHi<<32)+m_pCredits->nUploadedLo;
}

uint64	CClientCredits::GetDownloadedTotal(){
	return ( (uint64)m_pCredits->nDownloadedHi<<32)+m_pCredits->nDownloadedLo;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CClientCredits::GetScoreRatio(uint32 dwForIP,bool bInverted)
{
    EMULE_TRY

	if ( ( GetCurrentIdentState(dwForIP) == IS_IDFAILED || GetCurrentIdentState(dwForIP) == IS_IDBADGUY || GetCurrentIdentState(dwForIP) == IS_IDNEEDED) && g_eMuleApp.m_pClientCreditList->CryptoAvailable())
	{
	//	Bad guy - no credits for you
		return 1.0;
	}

//	By Maverick
//	Added Inversion to calculate the clients own ratio
//		at the remote client from its own data	
	uint64		qwDown = GetDownloadedTotal();
	uint64		qwUp = GetUploadedTotal();

  	if (bInverted)
	{
		qwDown = GetUploadedTotal();
		qwUp = GetDownloadedTotal();
	}

	if (qwDown < 1000000)
		return 1.0;

	double		dblResult = 0.0;

	if (!qwUp)
		dblResult = 10.0;
	else
		dblResult = static_cast<double>((static_cast<double>(qwDown)*2.0)/static_cast<double>(qwUp));

	double		dblResult2 = 0.0;

	dblResult2 = static_cast<double>(qwDown)/1048576.0;
	dblResult2 += 2.0;
	dblResult2 = sqrt(dblResult2);

	if (dblResult > dblResult2)
		dblResult = dblResult2;

	if (dblResult < 1.0)
		return 1.0;
	else if (dblResult > 10.0)
		return 10.0;

	return dblResult;

    EMULE_CATCH

    return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientCreditsList::CClientCreditsList(CPreferences* in_prefs)
{
	m_pAppPrefs = in_prefs;
	m_nLastSaved = ::GetTickCount();
	LoadList();
	
	InitalizeCrypting();
}

CClientCreditsList::~CClientCreditsList()
{
	SaveList();
	CClientCredits* cur_credit;
	CCKey tmpkey(0);
	POSITION pos = m_mapClients.GetStartPosition();
	while (pos){
		m_mapClients.GetNextAssoc(pos, tmpkey, cur_credit);
		delete cur_credit;
	}
	m_mapClients.RemoveAll();
	if (m_pSignkey){
		delete m_pSignkey;
		m_pSignkey = NULL;
	}
}

void CClientCreditsList::LoadList()
{
	CString strFileName = m_pAppPrefs->GetConfigDir() + CString(CLIENTS_MET_FILENAME);
	const int iOpenFlags = CFile::modeRead|CFile::osSequentialScan|CFile::typeBinary;
	CSafeBufferedFile file;
	CFileException fexp;
	if (!file.Open(strFileName, iOpenFlags, &fexp)){
		if (fexp.m_cause != CFileException::fileNotFound){
			CString strError(GetResString(IDS_ERR_LOADCREDITFILE));
			TCHAR szError[MAX_CFEXP_ERRORMSG];
			if (fexp.GetErrorMessage(szError, ELEMENT_COUNT(szError))){
				strError += _T(" - ");
				strError += szError;
			}
			g_eMuleApp.m_pdlgEmule->AddLogLine(true, RGB_LOG_ERROR + strError);
		}
		return;
	}
	setvbuf(file.m_pStream, NULL, _IOFBF, 16384);
	
	try{
		byte version;
		file.Read(&version, 1);
		if (version != CREDITFILE_VERSION && version != CREDITFILE_VERSION_29){
			g_eMuleApp.m_pdlgEmule->AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_CREDITFILEOLD));
			file.Close();
			return;
		}

		// everything is ok, lets see if the backup exist...
		CString strBakFileName;
		strBakFileName.Format(_T("%s") CLIENTS_MET_FILENAME _T(".bak"), m_pAppPrefs->GetConfigDir());
		DWORD dwBakFileSize = 0;
		BOOL bCreateBackup = TRUE;

		HANDLE hBakFile = ::CreateFile(strBakFileName, GENERIC_READ, FILE_SHARE_READ, NULL,
										OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hBakFile != INVALID_HANDLE_VALUE)
		{
			// Ok, the backup exist, get the size
			dwBakFileSize = ::GetFileSize(hBakFile, NULL); //debug
			if (dwBakFileSize > (DWORD)file.GetLength())
			{
				// the size of the backup was larger then the org. file, something is wrong here, don't overwrite old backup..
				bCreateBackup = FALSE;
			}
			//else: backup is smaller or the same size as org. file, proceed with copying of file
			::CloseHandle(hBakFile);
		}
		//else: the backup doesn't exist, create it

		if (bCreateBackup)
		{
			file.Close(); // close the file before copying

			if (!::CopyFile(strFileName, strBakFileName, FALSE))
				g_eMuleApp.m_pdlgEmule->AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_MAKEBAKCREDITFILE));

			// reopen file
			CFileException fexp;
			if (!file.Open(strFileName, iOpenFlags, &fexp)){
				CString strError(RGB_LOG_ERROR + GetResString(IDS_ERR_LOADCREDITFILE));
				TCHAR szError[MAX_CFEXP_ERRORMSG];
				if (fexp.GetErrorMessage(szError, ELEMENT_COUNT(szError))){
					strError += _T(" - ");
					strError += szError;
				}
				g_eMuleApp.m_pdlgEmule->AddLogLine(true, RGB_LOG_ERROR + strError);
				return;
			}
			setvbuf(file.m_pStream, NULL, _IOFBF, 16384);
			file.Seek(1, CFile::begin); //set filepointer behind file version byte
		}

		uint32 count;
		file.Read(&count, 4);
		m_mapClients.InitHashTable(count+5000); // TODO: should be prime number... and 20% larger

		const uint32 dwExpired = time(NULL) - 12960000; // today - 150 day
		uint32 cDeleted = 0;
		for (uint32 i = 0; i < count; i++){
			CreditStruct* newcstruct = new CreditStruct;
			memzero(newcstruct, sizeof(CreditStruct));
			if (version == CREDITFILE_VERSION_29)
				file.Read(newcstruct, sizeof(CreditStruct_29a));
			else
				file.Read(newcstruct, sizeof(CreditStruct));
			
			if (newcstruct->nLastSeen < dwExpired){
				cDeleted++;
				delete newcstruct;
				continue;
			}

			CClientCredits* newcredits = new CClientCredits(newcstruct);
			m_mapClients.SetAt(CCKey(newcredits->GetKey()), newcredits);
		}
		file.Close();

		if (cDeleted>0) g_eMuleApp.m_pdlgEmule->AddLogLine(false, GetResString(IDS_CREDITFILELOADED) + GetResString(IDS_CREDITSEXPIRED), count-cDeleted,cDeleted);
			else g_eMuleApp.m_pdlgEmule->AddLogLine(false, GetResString(IDS_CREDITFILELOADED), count);

	}
	catch(CFileException* error){
		OUTPUT_DEBUG_TRACE();
		if (error->m_cause == CFileException::endOfFile)
			g_eMuleApp.m_pdlgEmule->AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_CREDITFILECORRUPT));
		else{
			char buffer[MAX_CFEXP_ERRORMSG];
			error->GetErrorMessage(buffer, MAX_CFEXP_ERRORMSG);
			g_eMuleApp.m_pdlgEmule->AddLogLine(true, GetResString(IDS_ERR_CREDITFILEREAD), buffer);
		}
		error->Delete();
	}
}

void CClientCreditsList::SaveList()
{
//	DEBUG_ONLY(g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, "Saved Credit list"));
	m_nLastSaved = ::GetTickCount();

	CString strFileName = m_pAppPrefs->GetConfigDir() + CString(CLIENTS_MET_FILENAME);
	CFile file;// no buffering needed here since we swap out the entire array
	BYTE* pbyteBuffer = NULL;
	uint32 iCount;
	
	try 
	{
	//	Let's open the file, since "Open" generates an exception, no need for any check
		file.Open(strFileName, CFile::modeWrite|CFile::modeCreate|CFile::typeBinary);

		iCount = m_mapClients.GetCount();
			
		pbyteBuffer = new BYTE[iCount*sizeof(CreditStruct)];
		CClientCredits* cur_credit;
		CCKey tempkey(0);
		POSITION pos = m_mapClients.GetStartPosition();
		iCount = 0;
		while (pos)
		{
			m_mapClients.GetNextAssoc(pos, tempkey, cur_credit);
			if (cur_credit->GetUploadedTotal() || cur_credit->GetDownloadedTotal())
			{
				memcpy2(pbyteBuffer+(iCount*sizeof(CreditStruct)), cur_credit->GetDataStruct(), sizeof(CreditStruct));
				iCount++; 
			}
		}

		byte version = CREDITFILE_VERSION;
		file.Write(&version, 1);
		file.Write(&iCount, 4);
		file.Write(pbyteBuffer, iCount*sizeof(CreditStruct));
		//if (g_eMuleApp.m_pGlobPrefs->GetCommitFiles() >= 2 || (g_eMuleApp.m_pGlobPrefs->GetCommitFiles() >= 1 && !g_eMuleApp.m_pdlgEmule->IsRunning()))
		if (!g_eMuleApp.m_pdlgEmule->IsRunning())
			file.Flush();

		safe_delete(pbyteBuffer);
			
		file.Close();
	}
	catch(CFileException* error)
	{
		g_eMuleApp.m_pdlgEmule->AddLogLine(RGB_LOG_ERROR + _T("%s - %s"), GetResString(IDS_ERR_FAILED_CREDITSAVE), GetErrorMessage(error));
		error->Delete();
		if (pbyteBuffer)
			delete pbyteBuffer;
	}
	catch(CException *err)
	{ 
		g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__,__FILE__, __LINE__,GetErrorMessage(err));
		err->Delete();
	}

}

CClientCredits* CClientCreditsList::GetCredit(const uchar* key)
{
	CClientCredits* result;
	CCKey tkey(key);
	if (!m_mapClients.Lookup(tkey, result)){
		result = new CClientCredits(key);
		m_mapClients.SetAt(CCKey(result->GetKey()), result);
	}
	result->SetLastSeen();
	return result;
}

void CClientCreditsList::Process()
{
	if (::GetTickCount() - m_nLastSaved > MIN2MS(18))
		SaveList();
}

void CClientCredits::InitalizeIdent(){
	if (m_pCredits->nKeySize == 0 ){
		memzero(m_abyPublicKey, 80); // for debugging
		m_nPublicKeyLen = 0;
		IdentState = IS_NOTAVAILABLE;
	}
	else{
		m_nPublicKeyLen = m_pCredits->nKeySize;
		memcpy2(m_abyPublicKey, m_pCredits->abySecureIdent, m_nPublicKeyLen);
		IdentState = IS_IDNEEDED;
	}
	m_dwCryptRndChallengeFor = 0;
	m_dwCryptRndChallengeFrom = 0;
	m_dwIdentIP = 0;
}

void CClientCredits::Verified(uint32 dwForIP){
	m_dwIdentIP = dwForIP;
	// client was verified, copy the keyto store him if not done already
	if (m_pCredits->nKeySize == 0){
		m_pCredits->nKeySize = m_nPublicKeyLen; 
		memcpy2(m_pCredits->abySecureIdent, m_abyPublicKey, m_nPublicKeyLen);
		if (GetDownloadedTotal() > 0){
			// for security reason, we have to delete all prior credits here
			m_pCredits->nDownloadedHi = 0;
			m_pCredits->nDownloadedLo = 1;
			m_pCredits->nUploadedHi = 0;
			m_pCredits->nUploadedLo = 1; // in order to safe this client, set 1 byte
			DEBUG_ONLY(g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, "Credits deleted due to new SecureIdent"));
		}
	}
	IdentState = IS_IDENTIFIED;
#ifdef DEBUG_SHOW_SECUREID
DEBUG_ONLY(g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, "client verified!"));
#endif
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientCredits::SetSecureIdent(uchar* pachIdent, byte nIdentLen)
{
//	A verified Public key cannot change. If there is already a public key...
	if (MAXPUBKEYSIZE < nIdentLen || m_pCredits->nKeySize != 0 )
		return false;
	memcpy2(m_abyPublicKey,pachIdent, nIdentLen);
	m_nPublicKeyLen = nIdentLen;
	IdentState = IS_IDNEEDED;
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EIdentState	CClientCredits::GetCurrentIdentState(uint32 dwForIP){

	//DonGato (we're using this function before we have an IP?!)
	if(dwForIP == NULL)
		return IS_IDNEEDED;

	if (IdentState != IS_IDENTIFIED)
		return IdentState;
	else{
		if (dwForIP == m_dwIdentIP)
			return IS_IDENTIFIED;
		else
			return IS_IDBADGUY; 
			// mod note: clients which just reconnected after an IP change and have to ident yet will also have this state for 1-2 seconds
			//		 so don't try to spam such clients with "bad guy" messages (besides: spam messages are always bad)
	}
}


void CClientCreditsList::InitalizeCrypting()
{
	m_nMyPublicKeyLen = 0;
	memzero(m_abyMyPublicKey, 80); // not really needed; better for debugging tho
	m_pSignkey = NULL;
	if (!m_pAppPrefs->IsSecureIdentEnabled())
		return;
//
//	Check if the key file exists
	bool		bCreateNewKey = false;
	HANDLE		hKeyFile = ::CreateFile(m_pAppPrefs->GetConfigDir() + CString("cryptkey.dat"), GENERIC_READ,
										FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
//
//	If the key file doesn't exists or it is empty, we need to create a new one
	if (hKeyFile != INVALID_HANDLE_VALUE)
	{
		if (::GetFileSize(hKeyFile, NULL) == 0)
			bCreateNewKey = true;
		::CloseHandle(hKeyFile);
	}
	else
	{
		bCreateNewKey = true;
	}
	if (bCreateNewKey)
		CreateKeyPair();
	
//	Load key
	try
	{
	//	Load private key
		FileSource		filesource(CString(m_pAppPrefs->GetConfigDir() + CString("cryptkey.dat")), true,new Base64Decoder);

		m_pSignkey = new RSASSA_PKCS1v15_SHA_Signer(filesource);

	//	Calculate and store public key
		RSASSA_PKCS1v15_SHA_Verifier		pubkey(*m_pSignkey);
		ArraySink							asink(m_abyMyPublicKey, 80);

		pubkey.DEREncode(asink);
		m_nMyPublicKeyLen = asink.TotalPutLength();
		asink.MessageEnd();
	}
	catch(...)
	{
		if (m_pSignkey){
			delete m_pSignkey;
			m_pSignkey = NULL;
		}
		g_eMuleApp.m_pdlgEmule->AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_CRYPT_INITFAILED));
	}
//	Debug_CheckCrypting();
}

bool CClientCreditsList::CreateKeyPair(){
	try{
		AutoSeededRandomPool rng;
		InvertibleRSAFunction privkey;
		privkey.Initialize(rng,RSAKEYSIZE);

		Base64Encoder privkeysink(new FileSink(CString(m_pAppPrefs->GetConfigDir())+"cryptkey.dat"));
		//Base64Encoder privkeysink(new FileSink(CString(m_pAppPrefs->GetAppDir())+"cryptkey.dat"));
		privkey.DEREncode(privkeysink);
		privkeysink.MessageEnd();

		AddDebugLogLine(false, "Created new RSA keypair");
	}
	catch(...)
	{
		AddDebugLogLine(false, RGB_LOG_WARNING + _T("Failed to create new RSA keypair"));
		ASSERT ( false );
		return false;
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
byte CClientCreditsList::CreateSignature( CClientCredits *pTarget, uchar *pachOutput, byte nMaxSize, uint32 ChallengeIP,
										   byte byChaIPKind, RSASSA_PKCS1v15_SHA_Signer *sigkey /*= NULL*/ )
{
//	sigkey param is used for debug only
	if (sigkey == NULL)
		sigkey = m_pSignkey;

//	Create a signature of the public key from pTarget
	ASSERT( pTarget );
	ASSERT( pachOutput );

	byte		nResult;

	if ( !CryptoAvailable() )
	{
		return 0;
	}
	try
	{
		
		SecByteBlock			sbbSignature(sigkey->SignatureLength());
		AutoSeededRandomPool	rng;

		byte		abyteSignature[MAXPUBKEYSIZE+9];
		uint32		keylen = pTarget->GetSecIDKeyLen();

	//	Start by adding our public key which we are going to use as the signature "string".
		memcpy2(abyteSignature,pTarget->GetSecureIdent(),keylen);

	//	Append 4 additional bytes of data sent by the other client as extra verification
	//		that the signature is from us.
		uint32		challenge = pTarget->m_dwCryptRndChallengeFrom;

		ASSERT ( challenge != 0 );
		memcpy2(abyteSignature+keylen,&challenge,4);

		uint16		ChIpLen = 0;

		if (byChaIPKind != 0)
		{
			ChIpLen = sizeof(ChallengeIP) + sizeof(byChaIPKind);
		//	Append the challenge IP to the signature
			memcpy2(abyteSignature+keylen+4,&ChallengeIP,4);
		//	Append the challenge IP type (i.e. local client IP or remote client IP) to the signature
			memcpy2(abyteSignature+keylen+4+4,&byChaIPKind,1);
		}
	//	Encrypt the signature using our private key (only our public key can decrypt)
		sigkey->SignMessage(rng, abyteSignature, keylen+sizeof(challenge)+ChIpLen, sbbSignature.begin());

		ArraySink		asink(pachOutput, nMaxSize);

		asink.Put(sbbSignature.begin(), sbbSignature.size());
		nResult = asink.TotalPutLength();			
	}
	catch(...)
	{
		ASSERT ( false );
		nResult = 0;
	}
	return nResult;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	VerifyIdent() verifies that the sender of the signature is who he says he is given a signature string encrypted with
//		the client's private key
bool CClientCreditsList::VerifyIdent(CClientCredits* pTarget, uchar* pachSignature, byte nInputSize, uint32 dwForIP, byte byChaIPKind)
{
	ASSERT( pTarget );
	ASSERT( pachSignature );
	if ( !CryptoAvailable() )
	{
		pTarget->IdentState = IS_NOTAVAILABLE;
		return false;
	}

	bool		bResult = false;

	try
	{
	//
	//	We have all the information that the client did when its signature was created and we know how it was
	//		created. Given that we can reproduce the unencrypted signature that we should receive.
	//
		StringSource					ss_Pubkey((byte*)pTarget->GetSecureIdent(),pTarget->GetSecIDKeyLen(),true,0);
		RSASSA_PKCS1v15_SHA_Verifier	pubkey(ss_Pubkey);

		byte		abyteTestSig[MAXPUBKEYSIZE+9];

		memcpy2(abyteTestSig,m_abyMyPublicKey,m_nMyPublicKeyLen);

		uint32		challenge = pTarget->m_dwCryptRndChallengeFor;

		ASSERT ( challenge != 0 );

		memcpy2(abyteTestSig+m_nMyPublicKeyLen,&challenge,4);
	//
	//	v2 security improvements (not supported by 29b, not used as default by 29c)
		byte		nChIpSize = 0;

		if (byChaIPKind != 0)
		{
			uint32		ChallengeIP = 0;

			nChIpSize = sizeof(ChallengeIP) + sizeof(byChaIPKind);

			switch (byChaIPKind)
			{
				case CRYPT_CIP_LOCALCLIENT:
					ChallengeIP = dwForIP;
					break;
				case CRYPT_CIP_REMOTECLIENT:
#ifdef OLD_SOCKETS_ENABLED
					if (g_eMuleApp.m_pServerConnect->GetClientID() == 0 || g_eMuleApp.m_pServerConnect->IsLowID())
					{
						g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, RGB_LOG_WARNING + _T("Warning: Maybe SecureHash Ident fails because LocalIP is unknown"));
						ChallengeIP = g_eMuleApp.m_pServerConnect->GetLocalIP();
					}
					else
					{
						ChallengeIP = g_eMuleApp.m_pServerConnect->GetClientID();
					}
#endif OLD_SOCKETS_ENABLED
					break;
				case CRYPT_CIP_NONECLIENT: // maybe not supported in future versions
					ChallengeIP = 0;
					break;
			}
			memcpy2(abyteTestSig+m_nMyPublicKeyLen+4,&ChallengeIP,4);
			memcpy2(abyteTestSig+m_nMyPublicKeyLen+4+4,&byChaIPKind,1);
		}
		//v2 end
	//	Verify that when we decrypt the signature sent to us that it matches our test signature.
    	bResult = pubkey.VerifyMessage(abyteTestSig, m_nMyPublicKeyLen+4+nChIpSize, pachSignature, nInputSize);
	}
	catch(...)
	{
		bResult = false;
	}
	if (!bResult)
	{
		if (pTarget->IdentState == IS_IDNEEDED)
			pTarget->IdentState = IS_IDFAILED;
	}
	else
	{
		pTarget->Verified(dwForIP);
	}
	return bResult;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CryptoAvailable() returns true if we have sufficient information to verify a signature
bool CClientCreditsList::CryptoAvailable()
{
	return (m_nMyPublicKeyLen > 0 && m_pSignkey != NULL && m_pAppPrefs->IsSecureIdentEnabled());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef _DEBUG
bool CClientCreditsList::Debug_CheckCrypting(){
	// create random key
	AutoSeededRandomPool rng;

	RSASSA_PKCS1v15_SHA_Signer priv(rng, 384);
	RSASSA_PKCS1v15_SHA_Verifier pub(priv);

	byte abyPublicKey[80];
	ArraySink asink(abyPublicKey, 80);
	pub.DEREncode(asink);
	byte PublicKeyLen = asink.TotalPutLength();
	asink.MessageEnd();
	uint32 challenge = rand();
	// create fake client which pretends to be this emule
	CreditStruct* newcstruct = new CreditStruct;
	memzero(newcstruct, sizeof(CreditStruct));
	CClientCredits* newcredits = new CClientCredits(newcstruct);
	newcredits->SetSecureIdent(m_abyMyPublicKey,m_nMyPublicKeyLen);
	newcredits->m_dwCryptRndChallengeFrom = challenge;
	// create signature with fake priv key
	uchar pachSignature[200];
	memzero(pachSignature, 200);
	byte sigsize = CreateSignature(newcredits,pachSignature,200,0,false, &priv);


	// next fake client uses the random created public key
	CreditStruct* newcstruct2 = new CreditStruct;
	memzero(newcstruct2, sizeof(CreditStruct));
	CClientCredits* newcredits2 = new CClientCredits(newcstruct2);
	newcredits2->m_dwCryptRndChallengeFor = challenge;

	// if you uncomment one of the following lines the check has to fail
	//abyPublicKey[5] = 34;
	//m_abyMyPublicKey[5] = 22;
	//pachSignature[5] = 232;

	newcredits2->SetSecureIdent(abyPublicKey,PublicKeyLen);

	//now verify this signature - if it's true everything is fine
	bool bResult = VerifyIdent(newcredits2,pachSignature,sigsize,0,0);

	delete newcredits;
	delete newcredits2;

	return bResult;
}
#endif

uint32 CClientCredits::GetSecureWaitStartTime(uint32 dwForIP){
	if (m_dwUnSecureWaitTime == 0 || m_dwSecureWaitTime == 0)
		SetSecWaitStartTime(dwForIP);

//	this client is a SecureHash Client
	if (m_pCredits->nKeySize != 0)
	{
	//	good boy
		if (GetCurrentIdentState(dwForIP) == IS_IDENTIFIED)
		{
			return m_dwSecureWaitTime;
		}
	//	not so good boy
		else
		{
			if (dwForIP == m_dwWaitTimeIP)
			{
				return m_dwUnSecureWaitTime;
			}
		//	bad boy
			else
			{
			//	this can also happen if the client has not identified himself yet, but will do later - so maybe he is not a bad boy :) .
			/*
				CString buffer2, buffer;
				for (uint16 i = 0;i != 16;i++){
					buffer2.Format("%02X",this->m_pCredits->abyKey[i]);
					buffer+=buffer2;
				}
				g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, RGB_LOG_WARNING + _T("Warning: WaitTime resetted due to Invalid Ident for Userhash %s"),buffer.GetBuffer());
			*/
				
				m_dwUnSecureWaitTime = ::GetTickCount();
				m_dwWaitTimeIP = dwForIP;
				return m_dwUnSecureWaitTime;
			}	
		}
	}
	else{	// not a SecureHash Client - handle it like before for now (no security checks)
		return m_dwUnSecureWaitTime;
	}
}

void CClientCredits::SetSecWaitStartTime(uint32 dwForIP){
	m_dwUnSecureWaitTime = ::GetTickCount()-1;
	m_dwSecureWaitTime = ::GetTickCount()-1;
	m_dwWaitTimeIP = dwForIP;
}

void CClientCredits::ClearWaitStartTime(){
	m_dwUnSecureWaitTime = 0;
	m_dwSecureWaitTime = 0;
}
