//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// ClientListCtrl.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ClientListCtrl.h"
#include "otherfunctions.h"
#include "opcodes.h"
#include ".\\Details\\ClientDetails.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

BEGIN_MESSAGE_MAP(CClientListCtrl, CMuleListCtrl)
	ON_WM_CONTEXTMENU()
	ON_NOTIFY_REFLECT(LVN_COLUMNCLICK, OnColumnClick)
	ON_NOTIFY_REFLECT(NM_CLICK, OnNMClick)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnNMDblclk)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnGetDispInfo)
END_MESSAGE_MAP()

// CClientListCtrl

IMPLEMENT_DYNAMIC(CClientListCtrl, CMuleListCtrl)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientListCtrl::CClientListCtrl()
{
	memzero(&m_bSortAscending, sizeof(m_bSortAscending));

//	Refresh the list every 10 secs
	m_hTimer = ::SetTimer( NULL,				// Window to associate with
						   NULL,				// ID of timer (ignored if window = null)
						   30000,				// Time out duration
						   (TIMERPROC)CClientListCtrl::ListUpdateTimer);	// Proc to run
	if (!m_hTimer)
		g_eMuleApp.m_pdlgEmule->AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_TIMERCREATEFAILED));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientListCtrl::~CClientListCtrl()
{
	m_imageList.DeleteImageList();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::Init()
{
	CImageList		ilDummyImageList;

	ilDummyImageList.Create(1, 17,ILC_COLOR32 | ILC_MASK, 1, 1);
	SetImageList(&ilDummyImageList, LVSIL_SMALL);
	ilDummyImageList.Detach();

	for (int i = 0; i < ULCOL_NUMCOLUMNS; i++)
	{
		m_iColumnMaxWidths[i] = 0;
	}
//	Index of the column being measured. -1 for none.
	m_iMeasuringColumn = -1;

	SetExtendedStyle(LVS_EX_FULLROWSELECT);
	InsertColumn(CLCOL_USERNAME,GetResString(IDS_QL_USERNAME),LVCFMT_LEFT,150,0);
	InsertColumn(CLCOL_UPLOADSTATUS,GetResString(IDS_CL_UPLOADSTATUS),LVCFMT_LEFT,100,1);
	InsertColumn(CLCOL_TRANSFERREDUP,GetResString(IDS_CL_TRANSFUP),LVCFMT_LEFT,100,2);
	InsertColumn(CLCOL_DOWNLOADSTATUS,GetResString(IDS_CL_DOWNLSTATUS),LVCFMT_LEFT,100,3);
	InsertColumn(CLCOL_TRANSFERREDDOWN,GetResString(IDS_CL_TRANSFDOWN),LVCFMT_LEFT,100,4);

	CString		strColTemp;

	strColTemp = GetResString(IDS_CD_CSOFT);
	strColTemp.Remove(':');
	InsertColumn(CLCOL_CLIENTSOFTWARE,strColTemp,LVCFMT_LEFT,120,5);
	InsertColumn(CLCOL_CONNECTEDTIME,GetResString(IDS_CONNECTED),LVCFMT_LEFT,80,6);
	strColTemp = GetResString(IDS_CD_UHASH);
	strColTemp.Remove(':');
	InsertColumn(CLCOL_USERHASH,strColTemp,LVCFMT_LEFT,210,7);

	m_imageList.Create(16,16,ILC_COLOR32|ILC_MASK,0,10);
	m_imageList.SetBkColor(CLR_NONE);
	m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_NORMAL),16,16));			// 0
	m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_COMPROT),16,16));		// 1
	m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_MLDONKEY),16,16));		// 2
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_EDONKEYHYBRID),16,16));  // 3
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_CREDIT_ONLY),16,16));    // 4
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_FRIEND_ONLY),16,16));    // 5
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_BANNED_ONLY),16,16));    // 6
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_UNKNOWN),16,16));  		// 7
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_CDONKEY),16,16));  		// 8
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_LMULE),16,16));  		// 9
    m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_SHAREAZA),16,16));  		// 10
	m_imageList.Add(LoadImgLstIcon(MAKEINTRESOURCE(IDI_SECUREHASH),16,16));  	// 11

	LoadImgLstIcon(NULL,0,0);
	SetImageList(&m_imageList,LVSIL_SMALL);

	Localize();

	LoadSettings(CPreferences::TABLE_CLIENTLIST);

	if (g_eMuleApp.m_pGlobPrefs->DoUseSort())
	{
		SortInit(g_eMuleApp.m_pGlobPrefs->GetClientListSortCol());
	}
	else
	{
		int		iSortItem = g_eMuleApp.m_pGlobPrefs->GetColumnSortItem(CPreferences::TABLE_CLIENTLIST);
		bool	bSortAscending = g_eMuleApp.m_pGlobPrefs->GetColumnSortAscending(CPreferences::TABLE_CLIENTLIST);

		SetSortArrow(iSortItem, bSortAscending);
		SortItems(SortProc, iSortItem + (bSortAscending ? MLC_SORTASC : MLC_SORTDESC));
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::SortInit(int iSortCode)
{
//	Get the sort column
	int		iSortColumn = (iSortCode & MLC_COLUMNMASK);
//	Get the sort order
	bool	bSortAscending = (iSortCode & MLC_SORTDESC) == 0;

	SetSortArrow(iSortColumn,bSortAscending);
	SortItems(&SortProc, iSortCode);
	m_bSortAscending[iSortColumn] = bSortAscending;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::Localize()
{
	CHeaderCtrl	   *pHeaderCtrl = GetHeaderCtrl();
	HDITEM			hdi;

	hdi.mask = HDI_TEXT;

	if (pHeaderCtrl->GetItemCount() != 0)
	{
		CString		strRes;

		strRes = GetResString(IDS_QL_USERNAME);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_USERNAME, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_CL_UPLOADSTATUS);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_UPLOADSTATUS, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_CL_TRANSFUP);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_TRANSFERREDUP, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_CL_DOWNLSTATUS);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_DOWNLOADSTATUS, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_CL_TRANSFDOWN);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_TRANSFERREDDOWN, &hdi);
		strRes.ReleaseBuffer();

		strRes=GetResString(IDS_CD_CSOFT);
		strRes.Remove(':');
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_CLIENTSOFTWARE, &hdi);
		strRes.ReleaseBuffer();

		strRes = GetResString(IDS_CONNECTED);
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_CONNECTEDTIME, &hdi);
		strRes.ReleaseBuffer();

		strRes=GetResString(IDS_CD_UHASH);
		strRes.Remove(':');
		hdi.pszText = strRes.GetBuffer();
		pHeaderCtrl->SetItem(CLCOL_USERHASH, &hdi);
		strRes.ReleaseBuffer();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::ShowKnownClients()
{
	DeleteAllItems();

	int					i = 0;
	CString				strTemp;
	CUpDownClient	   *pClient;

	for (POSITION pos = g_eMuleApp.m_pClientList->m_clientList.GetHeadPosition(); pos != NULL; )
	{
		pClient = g_eMuleApp.m_pClientList->m_clientList.GetNext(pos);

		InsertItem(LVIF_TEXT|LVIF_PARAM,i,LPSTR_TEXTCALLBACK,0,0,0,reinterpret_cast<LPARAM>(pClient));
		UpdateClient(pClient);
		i++;
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateKnown();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::AddClient(CUpDownClient *pClient)
{
	LVFINDINFO		find;

	find.flags = LVFI_PARAM;
	find.lParam = (LPARAM)pClient;

	sint32		iResult = FindItem(&find);

	if (iResult == (-1) && pClient != NULL)
	{
		uint32		iNumItems = GetItemCount();

		iNumItems = InsertItem(LVIF_TEXT|LVIF_PARAM,iNumItems,LPSTR_TEXTCALLBACK,0,0,1,(LPARAM)pClient);
		UpdateClient(pClient);
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateKnown();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::RemoveClient(CUpDownClient* pClient)
{
	if (!g_eMuleApp.m_pdlgEmule->IsRunning())
		return;
	if (pClient == NULL)
		return;

	if (g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.GetType() == INFOLISTTYPE_SOURCE)
	{
		if (g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.GetClient() == pClient)
		{
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.SetList(INFOLISTTYPE_NONE);
		}
	}

//g_eMuleApp.AddDebugLogLine(false, _T("%s: Removing client %s addr(0x%08x) from client list ctrl"),__FUNCTION__,pClient->GetUserName(),pClient);
	sint32		iResult = 0;

	while (iResult != -1)
	{
		LVFINDINFO		find;

		find.flags = LVFI_PARAM;
		find.lParam = (LPARAM)pClient;
		iResult = FindItem(&find);
		if (iResult != (-1))
			DeleteItem(iResult);
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateKnown();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::UpdateClient(CUpDownClient* pClient)
{
//	There is some type of timing issue here.. If you click on item in the queue or upload and leave
//	the focus on it when you exit the cient, it breaks on line 854 of emuleDlg.cpp..
//	I added this IsRunning() check to this function and the DrawItem method and
//	this seems to keep it from crashing. This is not the fix but a patch until
//	someone points out what is going wrong.. Also, it will still assert in debug mode..
	if (!g_eMuleApp.m_pdlgEmule->IsRunning())
		return;
	if (pClient == NULL)
		return;

	LVFINDINFO		find;

	find.flags = LVFI_PARAM;
	find.lParam = reinterpret_cast<LPARAM>(pClient);

	sint16		iResult = FindItem(&find);

	if (iResult != -1)
		Update(iResult);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define DLC_DT_TEXT (DT_LEFT|DT_SINGLELINE|DT_VCENTER|DT_NOPREFIX|DT_END_ELLIPSIS)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	if (!g_eMuleApp.m_pdlgEmule->IsRunning())
		return;

	if (!lpDrawItemStruct->itemData)
		return;

	CDC			   *odc = CDC::FromHandle(lpDrawItemStruct->hDC);
	BOOL			bCtrlFocused = ((GetFocus() == this ) || (GetStyle() & LVS_SHOWSELALWAYS));

	if( odc && (lpDrawItemStruct->itemAction | ODA_SELECT) && (lpDrawItemStruct->itemState & ODS_SELECTED ))
	{
		if(bCtrlFocused)
			odc->SetBkColor(m_crHighlight);
		else
			odc->SetBkColor(m_crNoHighlight);
	}
	else
		odc->SetBkColor(GetBkColor());

	CUpDownClient	   *pClient = (CUpDownClient*)lpDrawItemStruct->itemData;
	CMemDC				dc(CDC::FromHandle(lpDrawItemStruct->hDC),&CRect(lpDrawItemStruct->rcItem));
	CFont			   *pOldFont = dc.SelectObject(GetFont());
	COLORREF			crOldTextColor = dc.SetTextColor(m_crWindowText);
	int					iWidth = 0;
	bool				bMeasuring = (m_iMeasuringColumn >= 0);
	UINT				iCalcFlag = bMeasuring ? (DT_LEFT|DT_SINGLELINE|DT_VCENTER|DT_NOPREFIX|DT_CALCRECT) : (DT_LEFT|DT_SINGLELINE|DT_VCENTER|DT_NOPREFIX|DT_END_ELLIPSIS);

	if (IsRightToLeftLanguage())
		iCalcFlag |= DT_RTLREADING;

	RECT			r = lpDrawItemStruct->rcItem;
	CString			strBuffer;

	CHeaderCtrl	   *pHeaderCtrl = GetHeaderCtrl();
	int				iNumColumns = pHeaderCtrl->GetItemCount();

	r.right = r.left - 8;
	r.left += 4;
	iWidth = 12;

	for (int iCurrent = 0; iCurrent < iNumColumns; iCurrent++)
	{
		int		iColumn = pHeaderCtrl->OrderToIndex(iCurrent);

		if (!IsColumnHidden(iColumn) && !(bMeasuring && iColumn != m_iMeasuringColumn))
		{
			r.right += GetColumnWidth(iColumn);
			switch (iColumn)
			{
				case CLCOL_USERNAME:
				{
					if (!bMeasuring)
					{
						POINT				point = {r.left, r.top+1};
						EnumClientListIcons	enumClientIcon;

						// Display pClient icon
						switch (pClient->GetClientSoft())
						{
							case SO_CDONKEY:
								enumClientIcon = CL_ICON_CDONKEY;
								break;
							case SO_LMULE:
								enumClientIcon = CL_ICON_LMULE;
								break;
							case SO_SHAREAZA:
								enumClientIcon = CL_ICON_SHAREAZA;
								break;
							case SO_EDONKEYHYBRID:
								enumClientIcon = CL_ICON_EDONKEYHYBRID;
								break;
							case SO_MLDONKEY:
								enumClientIcon = CL_ICON_MLDONKEY;
								break;
							case SO_EMULE:
							case SO_OLDEMULE:
								if (pClient->m_pCredits->GetCurrentIdentState(pClient->GetIP()) != IS_IDENTIFIED)
									enumClientIcon = CL_ICON_SECUREHASH;
								else
									enumClientIcon = CL_ICON_COMPROT;
								break;
							case SO_EDONKEY:
								enumClientIcon = CL_ICON_NORMAL;
								break;
							default:
								enumClientIcon = CL_ICON_UNKNOWN;
								break;
						}

						m_imageList.Draw(dc, enumClientIcon, point, ILD_TRANSPARENT);

					//	Display any Extra Overlay icons
						if (pClient->IsBanned())
							m_imageList.Draw(dc, CL_ICON_BANNED_ONLY, point, ILD_TRANSPARENT);
						else if (pClient->IsFriend())
							m_imageList.Draw(dc, CL_ICON_FRIEND_ONLY, point, ILD_TRANSPARENT);
						else if (pClient->m_pCredits->GetScoreRatio(pClient->GetIP()) > 1)
							m_imageList.Draw(dc, CL_ICON_CREDIT_ONLY, point, ILD_TRANSPARENT);

						r.left += 20;
					}
					else
						iWidth += 20;

					if (pClient->GetUserName() == _T(""))
						strBuffer.Format(_T("(%s)"), GetResString(IDS_UNKNOWN));
					else
						strBuffer = pClient->GetUserName();

					RECT	r2 = r;

					dc->DrawText(strBuffer,strBuffer.GetLength(),&r2,iCalcFlag);

					if (bMeasuring && strBuffer.GetLength() != 0)
						iWidth += (r2.right - r2.left + 1);

					r.left -=20;
					break;
				}
				case CLCOL_UPLOADSTATUS:
				{
					switch (pClient->GetUploadState())
					{
						case US_ONUPLOADQUEUE:
							strBuffer = GetResString(IDS_ONQUEUE);
							break;
						case US_PENDING:
							strBuffer = GetResString(IDS_CL_PENDING);
							break;
						case US_LOWTOLOWIP:
							strBuffer = GetResString(IDS_CL_LOW2LOW);
							break;
						case US_BANNED:
							strBuffer = GetResString(IDS_BANNED);
							break;
						case US_ERROR:
							strBuffer = GetResString(IDS_ERRORLIKE);
							break;
						case US_CONNECTING:
							strBuffer = GetResString(IDS_CONNECTING);
							break;
						case US_WAITCALLBACK:
							strBuffer = GetResString(IDS_CONNVIASERVER);
							break;
						case US_UPLOADING:
							strBuffer = GetResString(IDS_TRANSFERRING);
							break;
						default:
							strBuffer.Empty();
					}
					break;
				}
				case CLCOL_TRANSFERREDUP:
				{
					if (pClient->m_pCredits)
						strBuffer = CastItoXBytes(pClient->m_pCredits->GetUploadedTotal());
					else
						strBuffer.Empty();
					break;
				}
				case CLCOL_DOWNLOADSTATUS:
				{
					switch (pClient->GetDownloadState())
					{
						case DS_CONNECTING:
							strBuffer = GetResString(IDS_CONNECTING);
							break;
						case DS_CONNECTED:
							strBuffer = GetResString(IDS_ASKING);
							break;
						case DS_WAITCALLBACK:
							strBuffer = GetResString(IDS_CONNVIASERVER);
							break;
						case DS_ONQUEUE:
							if (pClient->IsRemoteQueueFull())
								strBuffer = GetResString(IDS_QUEUEFULL);
							else
								strBuffer = GetResString(IDS_ONQUEUE);
							break;
						case DS_DOWNLOADING:
							strBuffer = GetResString(IDS_TRANSFERRING);
							break;
						case DS_REQHASHSET:
							strBuffer = GetResString(IDS_RECHASHSET);
							break;
						case DS_NONEEDEDPARTS:
							strBuffer = GetResString(IDS_NONEEDEDPARTS);
							break;
						case DS_LOWTOLOWIP:
							strBuffer = GetResString(IDS_NOCONNECTLOW2LOW);
							break;
						case DS_TOOMANYCONNS:
							strBuffer = GetResString(IDS_TOOMANYCONNS);
							break;
						default:
							strBuffer.Empty();
					}
					break;
				}
				case CLCOL_TRANSFERREDDOWN:
				{
					if(pClient->m_pCredits)
						strBuffer = CastItoXBytes(pClient->m_pCredits->GetDownloadedTotal());
					else
						strBuffer.Empty();
					break;
				}
				case CLCOL_CLIENTSOFTWARE:
				{
					strBuffer.Format(_T("%s"),GetClientNameAndVersionString(pClient));
					break;
				}
				case CLCOL_CONNECTEDTIME:
				{
#ifdef OLD_SOCKETS_ENABLED
					if (pClient->m_pRequestSocket != NULL)
					{
						if (pClient->m_pRequestSocket->IsConnected())
						{
							strBuffer = GetResString(IDS_YES);
							break;
						}
					}
#endif //OLD_SOCKETS_ENABLED
					strBuffer = GetResString(IDS_NO);
					break;
				}
				case CLCOL_USERHASH:
					strBuffer = HashToString(pClient->GetUserHash());
					break;
			}
			if (iColumn != CLCOL_USERNAME)
			{
				RECT		r2 = r;

				dc->DrawText(strBuffer,strBuffer.GetLength(),&r2,iCalcFlag);
				if (bMeasuring && strBuffer.GetLength() != 0)
					iWidth += r2.right - r2.left + 1;
			}
			r.left += GetColumnWidth(iColumn);

			if (bMeasuring)
			{
			//	Pin the column widths at some reasonable value.
				if (iWidth < 40 && iWidth != 0)
					iWidth = 40;
				if (iWidth > m_iColumnMaxWidths[m_iMeasuringColumn])
					m_iColumnMaxWidths[m_iMeasuringColumn] = iWidth;
			}
		}
	}
//	Draw rectangle around selected item(s)
	if (!bMeasuring && (lpDrawItemStruct->itemAction | ODA_SELECT) && (lpDrawItemStruct->itemState & ODS_SELECTED))
	{
		RECT		outline_rec = lpDrawItemStruct->rcItem;

		outline_rec.top--;
		outline_rec.bottom++;
		CBrush gdiBrush(GetBkColor());
		CBrush *pOldBrush = dc->SelectObject(&gdiBrush);
		dc->FrameRect(&outline_rec, &gdiBrush);
		dc->SelectObject(pOldBrush);
		outline_rec.top++;
		outline_rec.bottom--;
		outline_rec.left++;
		outline_rec.right--;

		if (bCtrlFocused)
		{
			CBrush		fl_Brush(m_crFocusLine);
			dc->SelectObject(&fl_Brush);
			dc->FrameRect(&outline_rec, &fl_Brush);
		}
		else
		{
			CBrush nfl_Brush(m_crNoFocusLine);
			dc->SelectObject(&nfl_Brush);
			dc->FrameRect(&outline_rec, &nfl_Brush);
		}

		dc->SelectObject(pOldBrush);
	}

	if (pOldFont)
		dc.SelectObject(pOldFont);
	if (crOldTextColor)
		dc.SetTextColor(crOldTextColor);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CClientListCtrl::OnWndMsg(UINT iMessage,WPARAM wParam, LPARAM lParam, LRESULT *pResult)
{
	BOOL		bHandled = FALSE;

	EMULE_TRY

	switch (iMessage)
	{
		case WM_CL_REFRESH:
		{
			CClientList::ClientVector	   *pvecDirtyClients = g_eMuleApp.m_pClientList->GetDirtyClients();

			if (pvecDirtyClients != NULL)
			{
				int		iNumDirtyClients = pvecDirtyClients->size();

				for (int i = 0; i < iNumDirtyClients; i++)
				{
					UpdateClient((*pvecDirtyClients)[i]);
				}

				delete pvecDirtyClients;
			}
			bHandled = true;
			break;
		}
	}

	EMULE_CATCH

	if (!bHandled)
	{
		bHandled = CMuleListCtrl::OnWndMsg(iMessage,wParam,lParam,pResult);
	}

	return bHandled;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CClientListCtrl::OnNMDividerDoubleClick(NMHEADER *pHeader, LRESULT *pResult)
{
	EMULE_TRY

	int		iColumn = pHeader->iItem;

	m_iColumnMaxWidths[iColumn] = 0;
	m_iMeasuringColumn = iColumn;
	Invalidate();
	UpdateWindow();
	m_iMeasuringColumn = -1;
	if (m_iColumnMaxWidths[iColumn] > 0)
		SetColumnWidth(iColumn,m_iColumnMaxWidths[iColumn]);
	else
		CMuleListCtrl::OnNMDividerDoubleClick(pHeader,pResult);
	Invalidate();

	EMULE_CATCH

	return *pResult = TRUE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// CClientListCtrl message handlers

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnContextMenu(CWnd *pWnd, CPoint point)
{
	EMULE_TRY

	CTitleMenu		menuClient;
	POSITION		posSelClient = GetFirstSelectedItemPosition();
	UINT			dwMenuFlags = MF_STRING | MF_GRAYED;
	CUpDownClient	*pClient = NULL;

	if (posSelClient)
	{
		dwMenuFlags = MF_STRING;
		pClient = reinterpret_cast<CUpDownClient*>(GetItemData(GetNextSelectedItem(posSelClient)));

		if (pClient)
		{
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.SetList(INFOLISTTYPE_SOURCE,pClient);
		}
	}

	menuClient.CreatePopupMenu();
	menuClient.AddMenuTitle(GetResString(IDS_CLIENTLIST));
	menuClient.AppendMenu(dwMenuFlags,MP_DETAIL, GetResString(IDS_SHOWDETAILS) + GetStringFromShortcutCode(g_eMuleApp.m_pGlobPrefs->GetShortcutCode(IndexULDetails),SSP_TAB_PREFIX));

	UINT_PTR	dwRes = MP_ADDFRIEND;
	UINT		dwResStrId = IDS_ADDFRIEND;

	if (pClient && pClient->IsFriend())
	{
		dwRes = MP_REMOVEFRIEND;
		dwResStrId = IDS_REMOVEFRIEND;
	}
	menuClient.AppendMenu(dwMenuFlags, dwRes, GetResString(dwResStrId) + GetStringFromShortcutCode(g_eMuleApp.m_pGlobPrefs->GetShortcutCode(IndexULFriend), SSP_TAB_PREFIX));
	menuClient.AppendMenu(dwMenuFlags,MP_MESSAGE, GetResString(IDS_SEND_MSG) + GetStringFromShortcutCode(g_eMuleApp.m_pGlobPrefs->GetShortcutCode(IndexULMessage),SSP_TAB_PREFIX));
	menuClient.AppendMenu(dwMenuFlags,MP_SHOWLIST, GetResString(IDS_VIEWFILES) + GetStringFromShortcutCode(g_eMuleApp.m_pGlobPrefs->GetShortcutCode(IndexULFiles),SSP_TAB_PREFIX));
	menuClient.AppendMenu(dwMenuFlags | (((pClient && pClient->IsBanned()) ? MF_ENABLED : MF_GRAYED)), MP_UNBAN, GetResString(IDS_UNBAN) );

	menuClient.AppendMenu(MF_SEPARATOR);
	menuClient.AppendMenu( MF_STRING |
		((g_eMuleApp.m_pGlobPrefs->GetUpdateKnownList()) ? MF_CHECKED : MF_UNCHECKED),
		MP_UPDATEQQUEUE, GetResString(IDS_UPDATEQUEUE) );

	menuClient.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CClientListCtrl::OnCommand(WPARAM wParam,LPARAM lParam)
{
	POSITION posSelClient = GetFirstSelectedItemPosition();

	if (posSelClient)
	{
		CUpDownClient* 	pClient = reinterpret_cast<CUpDownClient*>(GetItemData(GetNextSelectedItem(posSelClient)));

		switch (wParam)
		{
			case MP_SHOWLIST:
			{
				pClient->RequestSharedFileList();
				break;
			}
			case MP_MESSAGE:
			{
				g_eMuleApp.m_pdlgEmule->m_wndChat.StartSession(pClient);
				break;
			}
			case MP_ADDFRIEND:
			{
				g_eMuleApp.m_pFriendList->AddFriend(pClient);
				UpdateClient(pClient);
				break;
			}
			case MP_REMOVEFRIEND:
			{
				g_eMuleApp.m_pFriendList->RemoveFriend(pClient);
				UpdateClient(pClient);
				break;
			}
			case MP_UNBAN:
			{
				if (pClient->IsBanned())
				{
					pClient->UnBan();
					UpdateClient(pClient);
				}
				break;
			}
			case MP_DETAIL:
			{
				CClientDetails		dialog(GetResString(IDS_CD_TITLE), pClient, this, 0);
				dialog.DoModal();
				break;
			}


		}
	}
	switch(wParam)
	{
		case MP_UPDATEQQUEUE:
		{
			g_eMuleApp.m_pGlobPrefs->SetUpdateKnownList(!g_eMuleApp.m_pGlobPrefs->GetUpdateKnownList());
			break;
		}
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnColumnClick(NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW		*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR); //<= Yeeeech

//	If it's a second click on the same column then reverse the sort order,
//		otherwise sort the new column in ascending order.
	m_bSortAscending[pNMListView->iSubItem] = !m_bSortAscending[pNMListView->iSubItem];
	SetSortArrow(pNMListView->iSubItem, m_bSortAscending[pNMListView->iSubItem]);
	SortItems(SortProc, pNMListView->iSubItem + ((m_bSortAscending[pNMListView->iSubItem]) ? MLC_SORTASC : MLC_SORTDESC));
	*pResult = 0;
	g_eMuleApp.m_pGlobPrefs->SetColumnSortItem(CPreferences::TABLE_CLIENTLIST, pNMListView->iSubItem);
	g_eMuleApp.m_pGlobPrefs->SetColumnSortAscending(CPreferences::TABLE_CLIENTLIST, m_bSortAscending[pNMListView->iSubItem]);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CClientListCtrl::SortProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	CUpDownClient*	pClient1 = reinterpret_cast<CUpDownClient*>(lParam1);
	CUpDownClient*	pClient2 = reinterpret_cast<CUpDownClient*>(lParam2);

	if (pClient1 == NULL || pClient2 == NULL)
		return 0;

	int	iCompare = 0;
	int	iSortMod = ((lParamSort & MLC_SORTDESC) == 0) ? 1 : -1;
	int	iSortColumn = (lParamSort & MLC_COLUMNMASK);

	switch (iSortColumn)
	{
		case CLCOL_USERNAME:
			if (pClient1->GetUserName() && pClient2->GetUserName())
				iCompare = _tcsicmp(pClient1->GetUserName(),pClient2->GetUserName());
			else
			{
				iCompare = (!pClient1->GetUserName()) ? ( (!pClient2->GetUserName())? 0:1) : -1;
				iSortMod = 1;
			}
			break;
		case CLCOL_UPLOADSTATUS:
			iCompare = pClient1->GetUploadState()-pClient2->GetUploadState();
			if ((pClient1->GetUploadState() == US_NONE) || (pClient2->GetUploadState() == US_NONE))
				iSortMod = 1;
			break;
		//eklmn: we are using a 64 bit variables for the credits
		case CLCOL_TRANSFERREDUP:
			if (pClient1->m_pCredits && pClient2->m_pCredits)
			{
				uint64	qwClient1Credits = pClient1->m_pCredits->GetUploadedTotal();
				uint64	qwClient2Credits = pClient2->m_pCredits->GetUploadedTotal();

				if (qwClient1Credits > qwClient2Credits)
					iCompare = 1;
				else if (qwClient1Credits < qwClient2Credits)
					iCompare = -1;
				else
					iCompare = 0;
			}
			else
			{
				iCompare = (!pClient1->m_pCredits) ? ( (!pClient2->m_pCredits)? 0:1) : -1;
				iSortMod = 1;
			}
			break;
		case CLCOL_DOWNLOADSTATUS:
			if (pClient1->GetDownloadState() == pClient2->GetDownloadState())
				iCompare = pClient1->IsRemoteQueueFull() ? 1 : (pClient2->IsRemoteQueueFull() ? -1: 0);
			else
				iCompare = pClient1->GetDownloadState() - pClient2->GetDownloadState();
			if ((pClient1->GetDownloadState() == DS_NONE) || (pClient2->GetDownloadState() == DS_NONE))
				iSortMod = 1;
			break;
		//eklmn: we are using a 64 bit variables for the credits
		case CLCOL_TRANSFERREDDOWN:
			if (pClient1->m_pCredits && pClient2->m_pCredits)
			{
				uint64	qwClient1Credits = pClient1->m_pCredits->GetDownloadedTotal();
				uint64	qwClient2Credits = pClient2->m_pCredits->GetDownloadedTotal();

				if (qwClient1Credits > qwClient2Credits)
					iCompare = 1;
				else if (qwClient1Credits < qwClient2Credits)
					iCompare = -1;
				else
					iCompare = 0;
			}
			else
			{
				iCompare = (!pClient1->m_pCredits) ? ( (!pClient2->m_pCredits)? 0:1) : -1;
				iSortMod = 1;
			}
			break;
		case CLCOL_CLIENTSOFTWARE:
			iCompare = SortClient(pClient1, pClient2, iSortMod);
			break;
		case CLCOL_CONNECTEDTIME:
#ifdef OLD_SOCKETS_ENABLED
			if (pClient1->m_pRequestSocket != NULL && pClient2->m_pRequestSocket != NULL)
				iCompare = pClient1->m_pRequestSocket->IsConnected() - pClient2->m_pRequestSocket->IsConnected();
			else
			{
				iCompare = (pClient1->m_pRequestSocket != NULL) ? 1: ((pClient2->m_pRequestSocket != NULL) ? -1 : 0);
				iSortMod = 1;
			}
#endif //OLD_SOCKETS_ENABLED
			break;
		case CLCOL_USERHASH:
			iCompare = memcmp(pClient1->GetUserHash(), pClient2->GetUserHash(), 16);
			break;
		default:
			return 0;
	}

	return iCompare * iSortMod;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnNMClick(NMHDR *pNMHDR, LRESULT *pResult)
{

	//eklmn: reset selection in DL list
	POSITION posSelClient = g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.GetFirstSelectedItemPosition();

	if (posSelClient)
	{
		while (posSelClient)
		{
			int iSelClientListIndex = g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.GetNextSelectedItem(posSelClient);
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.SetItemState(iSelClientListIndex, ~LVIS_SELECTED, LVIS_SELECTED);
		}

	}

	RefreshInfo();
	*pResult = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnNMDblclk(NMHDR *pNMHDR, LRESULT *pResult)
{
	POSITION posSelClient = GetFirstSelectedItemPosition();

	if (posSelClient)
	{
		CUpDownClient* pClient = reinterpret_cast<CUpDownClient*>(GetItemData(GetNextSelectedItem(posSelClient)));

		if (g_eMuleApp.m_pGlobPrefs->GetDetailsOnClick())
		{
			CClientDetails			dialog(GetResString(IDS_CD_TITLE), pClient, this, 0);
			dialog.DoModal();
		}
		else
		{
			g_eMuleApp.m_pdlgEmule->m_wndChat.StartSession(pClient);
		}
	}
	*pResult = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	RefreshInfo();
	CMuleListCtrl::OnKeyUp(nChar, nRepCnt, nFlags);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::RefreshInfo(void)
{
	if (GetFocus() == this)
	{
		POSITION		posSelClient = GetFirstSelectedItemPosition();

		if (posSelClient == 0)
		{
	 		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.SetList(INFOLISTTYPE_NONE);
		}
		else
		{
			CUpDownClient* pClient = reinterpret_cast<CUpDownClient*>(GetItemData(GetNextSelectedItem(posSelClient)));

	 		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlInfoList.SetList(INFOLISTTYPE_SOURCE,pClient);
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientListCtrl::OnGetDispInfo(NMHDR *pNMHDR, LRESULT *pResult)
{
	NMLVDISPINFO		*pDispInfo = reinterpret_cast<NMLVDISPINFO*>(pNMHDR);

	if (g_eMuleApp.m_pdlgEmule->IsRunning())
	{
	// Although we have an owner drawn listview control we store the text for the primary item in the listview, to be
	// capable of quick searching those items via the keyboard. Because our listview items may change their contents,
	// we do this via a text callback function. The listview control will send us the LVN_DISPINFO notification if
	// it needs to know the contents of the primary item.
	//
	// But, the listview control sends this notification all the time, even if we do not search for an item. At least
	// this notification is only sent for the visible items and not for all items in the list. Though, because this
	// function is invoked *very* often, no *NOT* put any time consuming code here in.

		if (pDispInfo->item.mask & LVIF_TEXT)
		{
			CUpDownClient		*pClient = reinterpret_cast<CUpDownClient*>(pDispInfo->item.lParam);

			if (pClient != NULL)
			{
				switch (pDispInfo->item.iSubItem)
				{
					case 0:
						if (pClient->GetUserName() != _T("") && pDispInfo->item.cchTextMax > 0)
						{
							_tcsncpy(pDispInfo->item.pszText, pClient->GetUserName(), pDispInfo->item.cchTextMax);
							pDispInfo->item.pszText[pDispInfo->item.cchTextMax-1] = _T('\0');
						}
						break;
					default:
					//	Shouldn't happen
						pDispInfo->item.pszText[0] = _T('\0');
						break;
				}
			}
		}
	}
	*pResult = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SortClient() compares two clients by software type/version/mod and returns -1,0,1. If 'iSortMode' is
//		1 then the sort is ascending, if -1 then descending. If the two clients can't be sorted by
//		software then they're sorted by user name.
int CClientListCtrl::SortClient(CUpDownClient *pClient1, CUpDownClient *pClient2, int iSortMode)
{
	int		iCompare = 0;

//	If either client is using unknown software, sort them by software ID (putting them at the end)
	if (pClient1->GetClientSoft() == SO_UNKNOWN || pClient2->GetClientSoft() == SO_UNKNOWN )
	{
		iCompare = (pClient1->GetClientSoft()-pClient2->GetClientSoft())*iSortMode;		//Cax2 append all unknown ones at the end
	}
//	If the clients are using the same software...
	else if (pClient1->GetClientSoft() == pClient2->GetClientSoft())
	{
	//	... and they're both eMule...
		if (pClient1->IsEmuleClient())
		{
			uint32		plus1 = (pClient1->GetPlusVersion() < 0x2600) ? pClient1->GetPlusVersion() : 0;
			uint32		plus2 = (pClient2->GetPlusVersion() < 0x2600) ? pClient2->GetPlusVersion() : 0;

		//	If they're both e+...
			if (plus1 && plus2)
				iCompare = plus2 - plus1;
		//	If only one is e+...
			else if (plus1 || plus2)
				iCompare = (plus1) ? -1 : 1;
		//	If neither is e+ and they're the same eMule version...
			else if (pClient1->GetMuleVersion() == pClient2->GetMuleVersion())
			{
			//	Sort by mod version
				if (!pClient1->GetModVersion().IsEmpty() && !pClient2->GetModVersion().IsEmpty())
					iCompare = _tcsicmp(pClient1->GetModVersion(), pClient2->GetModVersion());
				else if (!pClient1->GetModVersion().IsEmpty() || !pClient2->GetModVersion().IsEmpty())
					iCompare = (pClient1->GetModVersion().IsEmpty()) ? 1 : -1;
			}
			else
				iCompare = pClient2->GetMuleVersion() - pClient1->GetMuleVersion();
		}
	//	... but they're not eMule...
		else
		{
		//	Sort them by software _version_
			iCompare = pClient2->GetVersion() - pClient1->GetVersion();
		}
	}
//	Sort the clients by software ID
	else
	{
		iCompare = pClient1->GetClientSoft() - pClient2->GetClientSoft();
	}

//	If the sort is still ambiguous, do a secondary sort on user name
	if (iCompare == 0)
	{
		if (pClient1->GetUserName() && pClient2->GetUserName())
			iCompare = _tcsicmp(pClient1->GetUserName(),pClient2->GetUserName());
		else
			iCompare = (!pClient1->GetUserName()) ? ( (!pClient1->GetUserName())? 0:1) : -1;
		iCompare *= iSortMode;
	}
	return iCompare;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	PostRefreshMessage() puts a message into the list control's message queue telling it it needs to
//		refresh dirty clients and then immediately returns.
void CClientListCtrl::PostRefreshMessage()
{
	if (::IsWindow(GetSafeHwnd()))
	{
		MSG		msg;

	//	If there's no refresh message already in the message queue... (don't want to flood it)
		if (!::PeekMessage(&msg,m_hWnd,WM_CL_REFRESH,WM_CL_REFRESH,false))
		{
		//	... then post one with the current time so we know how long the request has been pending when we get it.
			PostMessage(WM_CL_REFRESH,0,static_cast<LPARAM>(::GetCurrentTime()));
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Refresh the queue every 10 secs
void CALLBACK CClientListCtrl::ListUpdateTimer(HWND hwnd, UINT uiMsg, UINT idEvent, DWORD dwTime)
{
	// Don't do anything if the app is shutting down - can cause unhandled exceptions
	if ((g_eMuleApp.m_pClientList == NULL) || !g_eMuleApp.m_pdlgEmule->IsRunning() || !g_eMuleApp.m_pGlobPrefs->GetUpdateKnownList())
		return;

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.DeleteAllItems();

	int					i = 0;
	CUpDownClient	   *pClient;

	for ( POSITION pos = g_eMuleApp.m_pClientList->m_clientList.GetHeadPosition(); pos != NULL; )
	{
		pClient = g_eMuleApp.m_pClientList->m_clientList.GetNext(pos);

		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.InsertItem(LVIF_TEXT|LVIF_PARAM,i,LPSTR_TEXTCALLBACK,0,0,0,reinterpret_cast<LPARAM>(pClient));
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(pClient);
		i++;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CClientListCtrl::PreTranslateMessage(MSG* pMsg)
{
	if (pMsg->message == WM_KEYDOWN || pMsg->message == WM_SYSKEYDOWN)
	{
		int		 iMessage	  = 0;
		short	 nCode		  = GetCodeFromPressedKeys(pMsg);
		POSITION posSelClient = GetFirstSelectedItemPosition();

		if (posSelClient != NULL)
		{
			CUpDownClient* pSource = reinterpret_cast<CUpDownClient*>(GetItemData(GetNextSelectedItem(posSelClient)));

			iMessage = GetClientListActionFromShortcutCode(nCode, pSource);

			if (iMessage > 0)
			{
				PostMessage(WM_COMMAND, static_cast<WPARAM>(iMessage));
				return TRUE;
			}
		}
	}

	return CMuleListCtrl::PreTranslateMessage(pMsg);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
