//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include "StdAfx.h"
#include "zlib/zlib.h"
#include "updownclient.h"
#include "partfile.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


//	members of CUpDownClient
//	which are mainly used for downloading functions

void CUpDownClient::DrawStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool bFlat)
{
	EMULE_TRY

	const COLORREF crBoth = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);	//Cax2 - same green used for files & clients
	const COLORREF crNeither = (bFlat) ? RGB(224, 224, 224) : RGB(240, 240, 240);
	const COLORREF crClientOnly = (bFlat) ? RGB(0, 0, 0) : RGB(95, 95, 95);
	const COLORREF crPending = RGB(255, 208, 0);
	const COLORREF crNextPending = RGB(255, 255, 100);
//--- xrmb:seetheneed ---
	const COLORREF crMeOnly = RGB(190, 190, 255);
//--- :xrmb ---

	uint32 filesize = (m_pReqPartFile) ? m_pReqPartFile->GetFileSize() : 1;

	CBarShader statusBar(rect->bottom - rect->top, rect->right - rect->left);
	statusBar.SetFileSize(filesize);
	statusBar.Fill(crNeither);

	if (m_pReqPartFile)
	{
		if (!onlygreyrect && m_pReqPartFile && m_pbytePartStatuses)
		{
			_Bvector gettingParts;
			bool isDLing = (m_eDownloadState == DS_DOWNLOADING);

			if (isDLing)	//Cax2 - no unneded calls to showdownloadingparts...
				ShowDownloadingParts(gettingParts);		// Barry - was only showing one part from client, even when reserved bits from 2 parts

			uint32 uEnd; //Cax2 less cpu usage

			for (uint32 i = 0;i < m_uPartCount;i++)
			{
			//--- xrmb:seetheneed --- & Cax2 graphic glitch fix

				if (PARTSIZE * (i + 1) > filesize)
					uEnd = filesize;
				else
					uEnd = PARTSIZE * (i + 1);

				if (m_pbytePartStatuses[i])
				{ //they have this part
				//	do i have it ?

					if (m_pReqPartFile->IsComplete(PARTSIZE * i, PARTSIZE * (i + 1) - 1))
						statusBar.FillRange(PARTSIZE * i, uEnd, crBoth);
					else if (isDLing && m_dwLastBlockOffset < uEnd && m_dwLastBlockOffset >= PARTSIZE * i)
						statusBar.FillRange(PARTSIZE * i, uEnd, crPending);
					else if (isDLing && gettingParts[i])	//Cax2 - no more flashes of yellow across all clients...
						statusBar.FillRange(PARTSIZE * i, uEnd, crNextPending);
					else
						statusBar.FillRange(PARTSIZE * i, uEnd, crClientOnly);
				}
				else //they don't have a part. Do I have it?
					if (m_pReqPartFile->IsComplete(PARTSIZE * i, PARTSIZE * (i + 1) - 1))
						statusBar.FillRange(PARTSIZE * i, uEnd, crMeOnly);

			//--- :xrmb --- & Cax2 graphic glitch fix
			}
		}
	}

	statusBar.Draw(dc, rect->left, rect->top, bFlat);

	EMULE_CATCH
}


bool CUpDownClient::Compare(CUpDownClient* tocomp)
{
	EMULE_TRY

	if (GetIP() && tocomp->GetIP() && GetIP() == tocomp->GetIP() && GetUserPort() == tocomp->GetUserPort())
		return true;

	if (HasValidHash() && tocomp->HasValidHash())
		return !md4cmp(GetUserHash(), tocomp->GetUserHash());

	if (HasLowID())
		return((GetUserID() == tocomp->GetUserID()) && (GetServerIP() == tocomp->GetServerIP()) && (GetUserPort() == tocomp->GetUserPort()));
	else
		return((GetUserID() == tocomp->GetUserID() && GetUserPort() == tocomp->GetUserPort()) || (GetIP() && (GetIP() == tocomp->GetIP() && GetUserPort() == tocomp->GetUserPort())));

	EMULE_CATCH

	return false;
}

//eklmn: return TRUE if client still connected, otherwise return false
bool CUpDownClient::AskForDownload()
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED

	if (g_eMuleApp.m_pListenSocket->TooManySockets() && !(m_pRequestSocket && m_pRequestSocket->IsConnected()))
	{
		if (GetDownloadState() != DS_TOOMANYCONNS && GetDownloadState() != DS_DOWNLOADING)
			SetDownloadState(DS_TOOMANYCONNS);

	//	Added by Tarod [Dnowo]
	//	Vorlost halved the code: asking time random only at the beginning for each new source
		if (m_dwLastAskedTime == 0)
		{
			m_dwLastAskedTime = GetTickCount() - FILEREASKTIME + static_cast<uint32>(ROUND((static_cast<double>(rand()) / RAND_MAX) * 300000.0));
		}

		return true;
	}

#endif //OLD_SOCKETS_ENABLED

	m_bUDPPending = false;

	m_dwLastAskedTime = GetTickCount();

	SetDownloadState(DS_CONNECTING);

	if (TryToConnect())
		return true;
	else
		return false;

	EMULE_CATCH

	return false;
}

bool CUpDownClient::IsSourceRequestAllowed()
{
	EMULE_TRY

	DWORD			dwTickCount = GetTickCount() + CONNECTION_LATENCY;
	unsigned int	nTimePassedClient = dwTickCount - GetLastSrcAnswerTime();
	unsigned int	nTimePassedFile = dwTickCount - m_pReqPartFile->GetLastAnsweredTime();
	bool			bNeverAskedBefore = GetLastSrcReqTime() == 0;

	UINT			dwSources = m_pReqPartFile->GetSourceCount();

	return (
			//	if client has the correct extended protocol
				ExtProtocolAvailable() && m_byteSourceExchangeVer >= 1 &&
			//	AND if we can send sources
				!g_eMuleApp.m_pGlobPrefs->IsDisabledXS() &&
			//	AND if we need more sources
				g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() > m_pReqPartFile->GetSourceCount() &&
			//	AND if...
				(
				//	source is not complete and file is rare, allow once every 10 minutes
					( !m_bCompleteSource
						&& ( dwSources - m_pReqPartFile->GetValidSourcesCount() <= RARE_FILE / 4
							|| dwSources <= RARE_FILE * 2 )
						&& (bNeverAskedBefore || nTimePassedClient > SOURCECLIENTREASK) )
				//	otherwise, allow every 90 minutes, but only if we haven't
				//	asked someone else in last 10 minutes
					|| ( (bNeverAskedBefore || nTimePassedClient > (unsigned)(SOURCECLIENTREASK * m_pReqPartFile->GetCommonFilePenalty())) &&
						(nTimePassedFile > SOURCECLIENTREASK) )
				)
	      );

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendFileRequest()
{
	EMULE_TRY

	ASSERT(m_pReqPartFile != NULL);

	if (m_pReqPartFile == NULL)
		return;

	AddAskedCountDown();

	CSafeMemFile		packetStream1(16 + 16);

	packetStream1.Write(m_pReqPartFile->GetFileHash(), 16);

	if (GetExtendedRequestsVersion() > 0)
		m_pReqPartFile->WritePartStatus(&packetStream1);

	if (GetExtendedRequestsVersion() > 1)
		m_pReqPartFile->WriteCompleteSourcesCount(&packetStream1); //netwolf: complete sources (zegzav) 13.05.03

	Packet		*pPacket = new Packet(&packetStream1);

	pPacket->m_eOpcode = OP_FILEREQUEST;

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

#ifdef OLD_SOCKETS_ENABLED
	m_pRequestSocket->SendPacket(pPacket, true);
#endif //OLD_SOCKETS_ENABLED

//	AddDebugLogLine(false ,"File Status: %s was requested from %s", m_pReqPartFile->GetFileName(), this->GetUserName());

//	Removed requesting the file status for files <= PARTSIZE for better compatibility with ed2k protocol (eDonkeyHybrid).
//	if the remote client answers the OP_FILEREQUEST with OP_FILEREQANSWER the file is shared by the remote client. if we
//	know that the file is shared, we know also that the file is complete and don't need to request the file status.
	if (m_pReqPartFile->GetPartCount() > 1)
	{
		CSafeMemFile	packetStream2(16);

		packetStream2.Write(m_pReqPartFile->GetFileHash(), 16);

		pPacket = new Packet(&packetStream2);

		pPacket->m_eOpcode = OP_SETREQFILEID;

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);

#ifdef OLD_SOCKETS_ENABLED

		m_pRequestSocket->SendPacket(pPacket, true);

#endif //OLD_SOCKETS_ENABLED
	}

	if (IsEmuleClient())
	{
		SetRemoteQueueFull(true);
		SetRemoteQueueRank(0);
	}

	if ( g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() < m_pReqPartFile->GetSourceCount()
	  || g_eMuleApp.m_pGlobPrefs->IsDisabledXS() )
	{
		return;
	}
//
//	DonGato - Disable XS for up to X sources
	if ( g_eMuleApp.m_pGlobPrefs->DisableXSUpTo()
	  && g_eMuleApp.m_pGlobPrefs->XSUpTo() < m_pReqPartFile->GetSourceCount() )
	{
		return;
	}

	if (IsSourceRequestAllowed())
	{
		m_pReqPartFile->SetLastAnsweredTimeTimeout();

		Packet		*pPacket = new Packet(OP_REQUESTSOURCES, 16, OP_EMULEPROT);

		md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadSourceExchange(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
		m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED
	//	AddDebugLogLine(false, "Send:Source Request User(%s) File(%s)", GetUserName(), m_pReqPartFile->GetFileName());
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	ProcessFileInfo() processes the contents of a File Request Answer packet
void CUpDownClient::ProcessFileInfo(char *pcPacketBuf, uint32 dwPacketSize)
{
	EMULE_TRY
//
//	Create a memory stream from the packet and read its contents
//
	CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);

	uchar		fileHash[16];

	packetStream.Read(fileHash, sizeof(fileHash));

	uint16		uNameLength;

	packetStream.Read(&uNameLength, sizeof(uint16));

	packetStream.Read(m_strClientFilename.GetBuffer(uNameLength), uNameLength);
	m_strClientFilename.ReleaseBuffer(uNameLength);

	m_dwLastAskedTime = ::GetTickCount(); //<<-- enkeyDEV(th1) -L2HAC- highid side
//
//	Patch 'fix disappearing m_pReqPartFile' START (taab)
	if (m_pReqPartFile == NULL)
	{
		CKnownFile		*pReqFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(fileHash);

		if (pReqFile != NULL && pReqFile->IsPartFile())
			m_pReqPartFile = dynamic_cast<CPartFile*>(pReqFile);
	}

	if (m_pReqPartFile == NULL)
	{
		AddDebugLogLine(false, RGB_LOG_DIMMED + _T("%s: ") + GetResString(IDS_ERR_WRONGFILEID) + _T(" (m_pReqPartFile==NULL)"), __FUNCTION__);
		return;
	}
	if (md4cmp(fileHash, m_pReqPartFile->GetFileHash()))
	{
		AddDebugLogLine(false, RGB_LOG_DIMMED + _T("%s: ") + GetResString(IDS_ERR_WRONGFILEID) + _T(" (m_pReqPartFile != fileHash)"), __FUNCTION__);
	}

//	Removed requesting the file status for files <= PARTSIZE for better compatibility with ed2k protocol (eDonkeyHybrid).
//	if the remote client answers the OP_FILEREQUEST with OP_FILEREQANSWER the file is shared by the remote client. if we
//	know that the file is shared, we know also that the file is complete and don't need to request the file status.
	if (m_pReqPartFile->GetPartCount() == 1){
		if (m_pbytePartStatuses)
		{
			delete[] m_pbytePartStatuses;
			m_pbytePartStatuses = NULL;
		}

		m_uPartCount = m_pReqPartFile->GetPartCount();
		m_pbytePartStatuses = new byte[m_uPartCount];
		memset2(m_pbytePartStatuses, 1, m_uPartCount);
		m_bCompleteSource = true;

		m_pReqPartFile->NewSrcPartsInfo();
		UpdateDisplayedInfo();

	//	even if the file is <= PARTSIZE, we _may_ need the hashset for that file (if the file size == PARTSIZE)
		if (m_pReqPartFile->m_bHashSetNeeded)
		{
			Packet * pPacket = new Packet(OP_HASHSETREQUEST, 16);
			md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

			SetDownloadState(DS_REQHASHSET);
			m_pReqPartFile->m_bHashSetNeeded = false;
		}
		else
		{
			SetDownloadState(DS_ONQUEUE);
			CSafeMemFile data1(16);	//TwoBottle Mod
			data1.Write(m_pReqPartFile->GetFileHash(), 16);
			Packet* pPacket = new Packet(&data1);
			pPacket->m_eOpcode = OP_STARTUPLOADREQ;
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
			m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

			m_dwLastAskedTime = ::GetTickCount();
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ProcessFileStatus(char* packet, uint32 size)
{
	EMULE_TRY

	CSafeMemFile packetStream((BYTE*)packet, size);	//TwoBottle Mod (exceptions!)
	uchar pcFileHash[16];
	packetStream.Read(pcFileHash, 16);

	if (!m_pReqPartFile)
	{
		m_pReqPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(pcFileHash);

		if (!m_pReqPartFile)
		{
			AddDebugLogLine(false, RGB_LOG_DIMMED + _T("%s: ") + GetResString(IDS_ERR_WRONGFILEID) + _T(" (m_pReqPartFile==NULL)"), __FUNCTION__);
			return;
		}
	}
	if (md4cmp(pcFileHash, m_pReqPartFile->GetFileHash()))
	{
		AddDebugLogLine(false, RGB_LOG_DIMMED + _T("%s: ") + GetResString(IDS_ERR_WRONGFILEID) + _T(" (m_pReqPartFile!=pcFileHash)"), __FUNCTION__);
		return;
	}

	uint16 uED2KPartCount;

	packetStream.Read(&uED2KPartCount, 2);

	if (m_pbytePartStatuses)
	{
		delete[] m_pbytePartStatuses;
		m_pbytePartStatuses = NULL;
	}

	bool bPartsNeeded = false;

	m_uPartCount = m_pReqPartFile->GetPartCount();
	if (uED2KPartCount == 0)
	{
		m_pbytePartStatuses = new byte[m_uPartCount];
		memset2(m_pbytePartStatuses, 1, m_uPartCount);
		bPartsNeeded = true;
		m_bCompleteSource = true;
	}
	else
	{
		if (m_pReqPartFile->GetED2KPartCount() != uED2KPartCount)
		{
			m_uPartCount = 0;
			throw GetResString(IDS_ERR_WRONGPARTNUMBER);
		}

		m_bCompleteSource = false;
		uint16 nCompleteSourcePart = 0;	// zegzav:completesrc (add)
		m_pbytePartStatuses = new byte[m_uPartCount];
		uint16 done = 0;

		while (done != m_uPartCount)
		{
			byte toread;
			packetStream.Read(&toread, 1);

			for (sint32 i = 0;i != 8;i++)
			{
				m_pbytePartStatuses[done] = ((toread >> i) & 1) ? 1 : 0;
				if (m_pbytePartStatuses[done])
				{
					nCompleteSourcePart++;
					if (!m_pReqPartFile->IsComplete(done * PARTSIZE, ((done + 1) * PARTSIZE) - 1))
						bPartsNeeded = true;
				}
				done++;

				if (done == m_uPartCount)
					break;
			}
		}

		if (nCompleteSourcePart == m_uPartCount)
			m_bCompleteSource = true;
	}

	m_pReqPartFile->NewSrcPartsInfo(); // eklmn: changes(00): NewSrcPartsInfo & UpdateAvailablePartsCount merged (CPU load)
	UpdateDisplayedInfo();

	//CheckForUpPartsInfo(); // obaldin

	if (!bPartsNeeded)
		SetDownloadState(DS_NONEEDEDPARTS);
	else if (m_pReqPartFile->m_bHashSetNeeded)
	{
		Packet * pPacket = new Packet(OP_HASHSETREQUEST, 16);
		md4cpy(pPacket->m_pcBuffer, m_pReqPartFile->GetFileHash());
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
		m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

		SetDownloadState(DS_REQHASHSET);
		m_pReqPartFile->m_bHashSetNeeded = false;
	}
	else
	{
		SetDownloadState(DS_ONQUEUE);
		CSafeMemFile data1(16);	//TwoBottle Mod
		data1.Write(m_pReqPartFile->GetFileHash(), 16);
		Packet* pPacket = new Packet(&data1);
		pPacket->m_eOpcode = OP_STARTUPLOADREQ;
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

		m_dwLastAskedTime = ::GetTickCount(); //<<-- enkeyDEV(th1) -L2HAC- highid side
	//AddDebugLogLine(false,"File Upload: %s was requested from %s", m_pReqPartFile->GetFileName(), this->GetUserName());
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::AddRequestForAnotherFile(CPartFile* file)
{
	EMULE_TRY

	if (file == NULL)
		return false;

//check Other list
	if (m_otherRequestsList.Find(file) != 0)
		return false;
	else
	m_otherRequestsList.AddTail(file);

	if (!file->m_A4AFsrcList.Find(this))
		file->m_A4AFsrcList.AddTail(this);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::RemoveRequestForAnotherFile(CPartFile* pPartFile)
{
	EMULE_TRY

	if (pPartFile == NULL)
		return;

//check Other list
	POSITION posPartFile = m_otherRequestsList.Find(pPartFile);

	if (posPartFile != NULL)
		m_otherRequestsList.RemoveAt(posPartFile);

	m_otherNoNeededMap.RemoveKey(pPartFile);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetDownloadState(EnumDLQState byNewState)
{
	EMULE_TRY

	if (m_eDownloadState != byNewState)
	{
		//clear variables before switching
		switch (m_eDownloadState)
		{
			case DS_DOWNLOADING:
				if (m_pReqPartFile)
					m_pReqPartFile->RemoveDownloadingSource(this);

				if (m_bTransferredDownMini && byNewState != DS_ERROR)
					g_eMuleApp.m_pGlobPrefs->Add2DownSuccessfulSessions(); // Increment our counters for successful sessions (Cumulative AND Session)
				else
					g_eMuleApp.m_pGlobPrefs->Add2DownFailedSessions(); // Increment our counters failed sessions (Cumulative AND Session)

				g_eMuleApp.m_pGlobPrefs->Add2DownSAvgTime((::GetTickCount() - m_dwDownStartTime) / 1000);

				//prevet any action from the client
				m_eDownloadState = byNewState;

				ClearDownloadBlocksList();

				ClearPendingBlocksList();

				m_dwDownDataRate = 0;

				if (byNewState == DS_NONE)
				{
					if (m_pbytePartStatuses)
						delete[] m_pbytePartStatuses;

					m_pbytePartStatuses = 0;

					m_uPartCount = 0;	//SyruS (0.29c)
				}

				m_uLastPartAsked = 0xFFFF;

#ifdef OLD_SOCKETS_ENABLED

				if (m_pRequestSocket && byNewState != DS_ERROR)
					m_pRequestSocket->DisableDownloadLimit();
#endif //OLD_SOCKETS_ENABLED
				break;

			case DS_ERROR:
				return;
		}

	//	switch the state & update info
		m_eDownloadState = byNewState;

		//prepare variables for new state
		switch (byNewState)
		{
			case DS_DOWNLOADING:
			//	Update statistical variables (start time of DL & minimal amount)
				m_dwLastDataRateCalculationTime = m_dwDownStartTime = ::GetTickCount();

				m_bTransferredDownMini = false;

				m_pReqPartFile->AddDownloadingSource(this);

				if (IsEmuleClient())
					SetRemoteQueueFull(false);

				SetAskedCountDown(0);

				SetRemoteQueueRank(0, false);

				break;
			case DS_CONNECTED:
				m_dwEnteredConnectedState = ::GetTickCount();
				break;
		}

		UpdateDisplayedInfo(true);
	}

	EMULE_CATCH
}

void CUpDownClient::ProcessHashSet(char* pPacket, uint32 size)
{
	EMULE_TRY

	if ((!m_pReqPartFile) || md4cmp(pPacket, m_pReqPartFile->GetFileHash()))
		throw GetResString(IDS_ERR_WRONGFILEID);

	CSafeMemFile data1((BYTE*)pPacket, size);

	if (m_pReqPartFile->LoadHashsetFromFile(data1, true))
	{
		SetDownloadState(DS_ONQUEUE);

		CSafeMemFile data2(16);
		data2.Write(this->m_pReqPartFile->GetFileHash(), 16);
		Packet* opacket = new Packet(&data2);
		opacket->m_eOpcode = OP_STARTUPLOADREQ;
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(opacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		m_pRequestSocket->SendPacket(opacket, true, true);
#endif //OLD_SOCKETS_ENABLED

		m_dwLastAskedTime = ::GetTickCount(); //<<-- enkeyDEV(th1) -L2HAC- highid side
	}
	else
	{
		m_pReqPartFile->m_bHashSetNeeded = true;
		AddDebugLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_BADHASHSET), m_pReqPartFile->GetFileName());
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendBlockRequests()
{
	EMULE_TRY

	m_dwLastBlockReceived = GetTickCount();

	if (m_pReqPartFile == NULL)
		return;
//
//	Get blocks from the same chunk first
	if (m_downloadBlocksList.IsEmpty())
	{
		//eklmn: get part for request. if actual part still has empty blocks, the we get the actuall
		// part, otherwise it will be a new part
		m_uLastPartAsked = m_pReqPartFile->GetNextRequestedBlock(this);

	//	Barry - instead of getting 3, just get how many is needed
		uint16		uCount, uSavedCount;

		uSavedCount = uCount = 3 - m_pendingBlocksList.GetCount();

		if (m_uLastPartAsked != 0xFFFF)		//previous active part ?
		{
			Requested_Block_Struct	**ppRequestedBlocksToAdd = new Requested_Block_Struct * [uCount];

			m_pReqPartFile->GetGapsInPart(this, m_uLastPartAsked, ppRequestedBlocksToAdd, &uCount);

			if (uCount == 0)		//finished block (part ?)
			{
			//	check if part was completed

				if ( m_pReqPartFile->GetBytesInPart(m_uLastPartAsked, true, true) == 0
				  || m_pReqPartFile->IsCorruptedPart(m_uLastPartAsked) )
				{
					m_uLastPartAsked = 0xFFFF; //ICR 2.10
				}

				m_uLastPartAsked = m_pReqPartFile->GetNextRequestedBlock(this); //get next part

				if (m_uLastPartAsked != 0xFFFF)		//more parts ?
				{
					uCount = uSavedCount;
					m_pReqPartFile->GetGapsInPart(this, m_uLastPartAsked, ppRequestedBlocksToAdd, &uCount);
				}
			}

			for (int i = 0; i < uCount; i++)
			{
				m_downloadBlocksList.AddTail(ppRequestedBlocksToAdd[i]);
			}
			
			delete[] ppRequestedBlocksToAdd;
		}
		else
		{
			uCount = 0;
		}
	}

//	Barry - Why are unfinished blocks requested again, not just new ones?
	while (m_pendingBlocksList.GetCount() < 3 && !m_downloadBlocksList.IsEmpty())
	{
		Pending_Block_Struct		*pPendingBlock = new Pending_Block_Struct;

		pPendingBlock->m_pRequestedBlock = m_downloadBlocksList.RemoveHead();
		pPendingBlock->m_pZStream = NULL;
		pPendingBlock->iZStreamErr = 0;
		m_pendingBlocksList.AddTail(pPendingBlock);
	}

	if (m_pendingBlocksList.IsEmpty())
	{
		Packet		*pPacket = new Packet(OP_CANCELTRANSFER);

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED
	//SetDownloadState(DS_NONEEDEDPARTS);

		if ((GetTickCount() - GetLastAskedTime()) > FILEREASKTIME)
			AskForDownload(); //<<--enkeyDEV(Ottavio84) modified by lovelace - reask for file

//		AddDebugLogLine(false, "Cancel download from %s due empty block list", this->GetUserName());

		return;
	}

	Packet		*pPacket = new Packet(OP_REQUESTPARTS, 40);

	CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pPacket->m_pcBuffer), 40);

	packetStream.Write(m_pReqPartFile->GetFileHash(), 16);

	POSITION		pos = m_pendingBlocksList.GetHeadPosition();
	uint32			dwNull = 0;

	Requested_Block_Struct		*pRequestedBlock;

	for (uint32 i = 0; i < 3; i++)
	{
		if (pos != NULL)
		{
			pRequestedBlock = m_pendingBlocksList.GetNext(pos)->m_pRequestedBlock;
			packetStream.Write(&pRequestedBlock->m_dwStartOffset, 4);
		}
		else
		{
			packetStream.Write(&dwNull, 4);
		}
	}

	pos = m_pendingBlocksList.GetHeadPosition();

	for (uint32 i = 0; i < 3; i++)
	{
		if (pos != NULL)
		{
			pRequestedBlock = m_pendingBlocksList.GetNext(pos)->m_pRequestedBlock;

			uint32		dwEndOffset = pRequestedBlock->m_dwEndOffset + 1;

			packetStream.Write(&dwEndOffset, 4);
		}
		else
		{
			packetStream.Write(&dwNull, 4);
		}
	}

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Barry - Originally this only wrote to disk when a full 180k block
           had been received from a client, and only asked for data in
		   180k blocks.

		   This meant that on average 90k was lost for every connection
		   to a client data source. That is a lot of wasted data.

		   To reduce the lost data, packets are now written to a buffer
		   and flushed to disk regularly regardless of size downloaded.
		   This includes compressed packets.

		   Data is also requested only where gaps are, not in 180k blocks.
		   The requests will still not exceed 180k, but may be smaller to
		   fill a gap.
*/
void CUpDownClient::ProcessBlockPacket(char *pcPacketBuf, uint32 dwPacketSize, bool bPacked/*=false*/)
{
	EMULE_TRY

	const int		HEADER_SIZE = 24;

//	Update timer (eklmn: don't remove this value cause it use to timeout check)
	m_dwLastBlockReceived = ::GetTickCount();

//	Read data from packet
	CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);

	uchar		fileHash[16];

	packetStream.Read(fileHash, sizeof(fileHash));				//<HASH>

//	Check that this data is for the correct file
	if (m_pReqPartFile == NULL || md4cmp(fileHash, m_pReqPartFile->GetFileHash()))
	{
		throw CString(CString(__FUNCTION__) + _T(": ") + GetResString(IDS_ERR_WRONGFILEID));
	}

//	Find the start & end positions, and size of this chunk of data
	uint32			dwStartPos = 0;
	uint32			dwEndPos = 0;
	uint32			dwBlockSize = 0;
	const uint32 	dwDataTransferdInPacket = dwPacketSize - HEADER_SIZE;

	packetStream.Read(&dwStartPos, sizeof(uint32));				//<BYTEOFFSET:DWORD> start position

	if (bPacked)
	{
		packetStream.Read(&dwBlockSize, sizeof(uint32));		//<BYTELENGTH:DWORD> packed block size
		dwEndPos = dwStartPos + dwDataTransferdInPacket;
		m_bUsedCompressionDown = true;
	}
	else
	{
		packetStream.Read(&dwEndPos, sizeof(uint32));			//<BYTEOFFSET:DWORD> end position
	}

//	Check that packet size matches the declared packetStream packet + header size (24)
	if (dwPacketSize != ((dwEndPos - dwStartPos) + HEADER_SIZE))
	{
		throw CString(CString(__FUNCTION__ ": ") + GetResString(IDS_ERR_BADDATABLOCK));
	}

//	Protect from 0 data size, because it spoils statistics
	if (dwDataTransferdInPacket == 0)
	{
		return;
	}

//	Extended statistics information based on which client and remote port sent this data.
//		The new function adds the bytes to the grand total as well as the given client/port.
//		bFromPF is not relevant to downloaded data.  It is purely an uploads statistic.
	g_eMuleApp.m_pGlobPrefs->Add2SessionTransferData(GetClientSoft(), GetUserPort(), false, TRANSFERTYPE_DOWNLOAD, dwDataTransferdInPacket);

//	Move end back one, should be inclusive
	m_pCredits->AddDownloaded(dwDataTransferdInPacket, this->GetIP());

	dwEndPos--;

	m_dwTransferredDown += dwDataTransferdInPacket;

	m_dwTransferredInLastPeriod += dwDataTransferdInPacket;

//	Loop through to find the reserved block that this is within
	Pending_Block_Struct* cur_block;

	POSITION pos1, pos2;

	for (pos1 = m_pendingBlocksList.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		cur_block = m_pendingBlocksList.GetNext(pos1);
	//	Cax2: I'm not happy with this whole if clause

		if ((cur_block->m_pRequestedBlock->m_dwStartOffset <= dwStartPos) && (cur_block->m_pRequestedBlock->m_dwEndOffset >= dwStartPos))
		{
		//	Found reserved block

			if (cur_block->iZStreamErr != 0)
			{
				m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwEndOffset);
				break;
			}

		//	Remember this start pos, used to draw part downloading in list
			m_dwLastBlockOffset = dwStartPos;

		//	Occasionally packets are duplicated, no point writing it twice
		//	This will be 0 in these cases, or the length written otherwise
			uint32 lenWritten = 0;

		//	Handle differently depending on whether bPacked or not

			if (!bPacked)
			{
			//	Write to disk (will be buffered in part file class)
				lenWritten = m_pReqPartFile->WriteToBuffer( dwDataTransferdInPacket,
				             (BYTE *) (pcPacketBuf + HEADER_SIZE),
				             dwStartPos,
				             dwEndPos,
				             cur_block->m_pRequestedBlock );
#ifdef _DEBUG
				if (lenWritten == 0)
					AddDebugLogLine(false, RGB_LOG_WARNING + "client: %s, type: %s, mod: %s", GetUserName(), GetClientNameString(GetClientSoft()), GetModVersion());
#endif
			}
			else // Packed
			{
			//	Create space to store unzipped packetStream, the dwPacketSize is only an initial guess, will be resized in unzip() if not big enough
				uint32 lenUnzipped = (dwPacketSize * 2);
			//	Don't get too big

				if (lenUnzipped > (EMBLOCKSIZE + 300))
					lenUnzipped = (EMBLOCKSIZE + 300);

				BYTE *unzipped = new BYTE[lenUnzipped];

			//	Try to unzip the pcPacketBuf
				int result = unzip(cur_block, (BYTE *)(pcPacketBuf + HEADER_SIZE), dwDataTransferdInPacket, &unzipped, &lenUnzipped);

				if (result == Z_OK)
				{
				//	Write any unzipped packetStream to disk

					if (lenUnzipped > 0)
					{
					//	Use the current start and end positions for the uncompressed packetStream
						dwStartPos = cur_block->m_pRequestedBlock->m_dwStartOffset + cur_block->m_dwTotalUnzipped - lenUnzipped;
						dwEndPos = cur_block->m_pRequestedBlock->m_dwStartOffset + cur_block->m_dwTotalUnzipped - 1;

						if (dwBlockSize * 25 < lenUnzipped)
						{
							m_uLastPartAsked = 0xFFFF;	//eklmn: feature(00): ICR (reset choosen part in case suspicious block)

							AddDebugLogLine(true, RGB_LOG_DIMMED + _T("Received suspicious block: file '%s', part %u, block %u, blocksize %u, comp. blocksize %u"), m_pReqPartFile->GetFileName(), cur_block->m_pRequestedBlock->m_dwStartOffset / PARTSIZE, cur_block->m_pRequestedBlock->m_dwStartOffset / EMBLOCKSIZE, lenUnzipped, dwBlockSize);
							AddDebugLogLine(false, RGB_LOG_DIMMED + _T("username '%s', ip %u.%u.%u.%u:%u, hash %s"),
								m_strUserName, (byte)m_dwUserIP, (byte)(m_dwUserIP >> 8),
								(byte)(m_dwUserIP >> 16), (byte)(m_dwUserIP >> 24),
								GetUserPort(), HashToString(GetUserHash()));

							if (m_pReqPartFile->GetDiscardSuperCompressed())
							{
							//--- (un)comment the block-dropping as you wish---
								m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwStartOffset + 1);
								uint32 ip1 = 0;
								CString buf, temp;
								int counter = 0;
								buf.Format(_T("%u.%u.%u.%u"), (byte)m_dwUserIP, (byte)(m_dwUserIP >> 8), (byte)(m_dwUserIP >> 16), (byte)(m_dwUserIP >> 24));

								for (int i = 0; i < 4; i++)
								{
									temp = buf.Tokenize(_T("."), counter);
									ip1 += _tstoi(temp) * pow(256, 3 - i);
								}
								//eklmn: add to temporary(24h) banned
								g_eMuleApp.m_pIPFilter->AddTemporaryBannedIP(ip1);
								// set an ERROR state, client will be removed & disconnected by PartFile::Process later on
								SetDownloadState(DS_ERROR);
								AddDebugLogLine(false, RGB_LOG_DIMMED + _T("block dropped & IP locked out!"));
							}

						//	BavarianSnail
							else
							{
								if (dwEndPos > cur_block->m_pRequestedBlock->m_dwEndOffset)
								{
									AddDebugLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_CORRUPTCOMPRPKG), m_pReqPartFile->GetFileName(), 666);
									m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwEndOffset);
								}
								else
								{
								//	Write uncompressed packetStream to file
									lenWritten = m_pReqPartFile->WriteToBuffer(dwDataTransferdInPacket,
									             unzipped,
									             dwStartPos,
									             dwEndPos,
									             cur_block->m_pRequestedBlock );
									if (lenWritten == 0)
										AddDebugLogLine(false, RGB_LOG_WARNING + "client: %s, type: %s, mod: %s", GetUserName(), GetClientNameString(GetClientSoft()), GetModVersion());
								}
							}

						//- BavarianSnail
						}
						else
						{
							if (dwEndPos > cur_block->m_pRequestedBlock->m_dwEndOffset)
							{
								AddDebugLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_CORRUPTCOMPRPKG), m_pReqPartFile->GetFileName(), 666);
								m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwEndOffset);
							}
							else
							{
							//	Write uncompressed packetStream to file
								lenWritten = m_pReqPartFile->WriteToBuffer( dwDataTransferdInPacket,
								             unzipped,
								             dwStartPos,
								             dwEndPos,
								             cur_block->m_pRequestedBlock );
								if (lenWritten == 0)
									AddDebugLogLine(false, RGB_LOG_WARNING + "client: %s, type: %s, mod: %s", GetUserName(), GetClientNameString(GetClientSoft()), GetModVersion());
							}
						}
					}
				}
				else
				{
					cur_block->iZStreamErr = result;
					AddDebugLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_CORRUPTCOMPRPKG), m_pReqPartFile->GetFileName(), result);
					m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwEndOffset);

				//	Release resources in case of decompression error
					if (cur_block->m_pZStream != NULL)
					{
						inflateEnd(cur_block->m_pZStream);
						delete cur_block->m_pZStream;
						cur_block->m_pZStream = NULL;
					}
				}

				try
				{
					if (unzipped)
				delete [] unzipped;
			}
				catch (...)
				{}

			}

			//eklmn: check result of WriteToBuffer, if we save some byte, then this DL session can be
			// counted as succesful
			if (lenWritten > 0)
			{
				m_bTransferredDownMini = true;		// -khaos--+++> For determining whether the current download session was a success or not.
			}

			// eklmn: it needs to check "END of the block" condition every time, cause
			//  if remote client will send us a data that already were written, we will not delete the block & delete a packet

			//	[TwoBottle Mod]: If finished reserved block
			//	in this place because m_pReqPartFile->WriteToBuffer was called
			//	cur_block can be already deleted (if file was completed)
			//	so to avoid access violation here we will check
			//	m_pendingBlocksList for emptyness....possibly there is
			//	better check
			if (!m_pendingBlocksList.IsEmpty() && dwEndPos == cur_block->m_pRequestedBlock->m_dwEndOffset)
			{
				m_pReqPartFile->RemoveBlockFromList(cur_block->m_pRequestedBlock->m_dwStartOffset, cur_block->m_pRequestedBlock->m_dwEndOffset);
				delete cur_block->m_pRequestedBlock;
			//	Not always allocated

				if (cur_block->m_pZStream)
				{
					inflateEnd(cur_block->m_pZStream);
					delete cur_block->m_pZStream;
				}

				delete cur_block;
				m_pendingBlocksList.RemoveAt(pos2);

			//	Request next block
				SendBlockRequests();
			}


		//	Stop looping and exit method
			break;
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CUpDownClient::unzip(Pending_Block_Struct *pRequestedBlock, BYTE *pbyteZipped, uint32 dwLenZipped, BYTE **ppbyteUzipped, uint32 *pdwLenUnzipped)
{
	int iErr = Z_MEM_ERROR;

	EMULE_TRY

//	Save some typing
	z_stream *zS = pRequestedBlock->m_pZStream;

//	Is this the first time this block has been unzipped
	if (zS == NULL)
	{
		pRequestedBlock->m_dwTotalUnzipped = 0;
	//	Create stream
		zS = pRequestedBlock->m_pZStream = new z_stream;

	//	Initialize stream values
		zS->zalloc = (alloc_func)0;
		zS->zfree = (free_func)0;

	//	Set output data streams, do this here to avoid overwriting on recursive calls
		zS->next_out = *ppbyteUzipped;
		zS->avail_out = *pdwLenUnzipped;

	//	Initialise the z_stream
		iErr = inflateInit(zS);

		if (iErr != Z_OK)
			return iErr;
	}

//	Use whatever input is provided
	zS->next_in = pbyteZipped;
	zS->avail_in = dwLenZipped;

	zS->next_out = *ppbyteUzipped;
	zS->avail_out = *pdwLenUnzipped;

	for (;;)
	{
	//	Unzip the data
		iErr = inflate(zS, Z_SYNC_FLUSH);

		if (iErr == Z_OK)
		{
			if (zS->avail_in == 0)
			{
			//	All available input has been successfully processed.
			//	Set the size to the amount unzipped in this call.
				*pdwLenUnzipped = (zS->total_out - pRequestedBlock->m_dwTotalUnzipped);
				pRequestedBlock->m_dwTotalUnzipped = zS->total_out;
			}
			else if (zS->avail_out == 0)
			{
			//	Output buffer is full
				uint32		dwNewLen = (*pdwLenUnzipped << 1);	//	increase output buffer size

				iErr = Z_MEM_ERROR;	//	be ready for lack of memory

			//	Copy unzipped data to a new buffer
				BYTE		*pbyteNewBuff = new BYTE[dwNewLen];
				uint32		dwUnpackedSz = zS->total_out - pRequestedBlock->m_dwTotalUnzipped;

				memcpy2(pbyteNewBuff, *ppbyteUzipped, dwUnpackedSz);

				delete [] (*ppbyteUzipped);

				*ppbyteUzipped = pbyteNewBuff;
				*pdwLenUnzipped = dwNewLen;

			//	Set correct position inside a new output buffer
				zS->next_out = pbyteNewBuff + dwUnpackedSz;
				zS->avail_out = dwNewLen - dwUnpackedSz;
				continue;	//	try again
			}
		}
	//	Is zip finished reading all currently available input and writing all generated output
		else if (iErr == Z_STREAM_END)
		{
		//	Finish up
			iErr = inflateEnd(zS);
			pRequestedBlock->m_pZStream = NULL;

			if (iErr == Z_OK)
			{
			//	Got the good result, set the size to the unzipped amount
				*pdwLenUnzipped = (zS->total_out - pRequestedBlock->m_dwTotalUnzipped);
				pRequestedBlock->m_dwTotalUnzipped = zS->total_out;
			}
			delete zS;
		}
		else
		{
		//	Should not get here unless input data is corrupted
			*pdwLenUnzipped = 0;
		//	Error will be logged by caller on return
		}
		break;
	}

	EMULE_CATCH

	return iErr;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::CalculateDownloadRate()
{
	EMULE_TRY

	uint32	dwCurTick = GetTickCount();

	if ((dwCurTick - m_dwLastBlockReceived) > DOWNLOADTIMEOUT)
	{
#ifdef OLD_SOCKETS_ENABLED
		if (m_pRequestSocket != NULL)
		{
			Packet * pPacket = new Packet(OP_CANCELTRANSFER);
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
			m_pRequestSocket->SendPacket(pPacket, true, true);
		}
#endif //OLD_SOCKETS_ENABLED

		SetDownloadState(DS_ONQUEUE);
		m_dwDownDataRate = 0;

//		if (this != NULL)
//		{
//			AddDebugLogLine(false, "Client %s (%s) was dropped due to timeout %u. Send him a CANCEL", GetUserName(), GetClientNameAndVersionString(this), dwCurTick - m_dwLastBlockReceived);
//		}
	}
	else
	{
		uint32 time_diff_ms = dwCurTick - m_dwLastDataRateCalculationTime;

	//update average sum
		m_qwClientSumDLDataRateOverLastNMeasurements += m_dwTransferredInLastPeriod;
		m_qwClientSumDLTickOverLastNMeasurements += time_diff_ms;
	//	update a lists
		m_averageDLDataRateList.push_front(m_dwTransferredInLastPeriod);
		m_averageDLTickList.push_front(time_diff_ms);
	//	check lists size

		while (m_averageDLTickList.size() > 500)
		{
			m_qwClientSumDLDataRateOverLastNMeasurements -= m_averageDLDataRateList.back();
			m_averageDLDataRateList.pop_back();
			m_qwClientSumDLTickOverLastNMeasurements -= m_averageDLTickList.back();
			m_averageDLTickList.pop_back();
		}

	//	calculate average trasfered data (bytes)
		uint64 AverageTransferredData = m_qwClientSumDLDataRateOverLastNMeasurements / m_averageDLDataRateList.size();

	//	calculate average time period between measuremets (ms)
		uint64 AverageTransferTimePeriod = m_qwClientSumDLTickOverLastNMeasurements / m_averageDLTickList.size();

	//	calculate a DL rate (b/s)
		if (AverageTransferTimePeriod != 0)
			m_dwDownDataRate = 1000 * AverageTransferredData / AverageTransferTimePeriod;
		else
			m_dwDownDataRate = 0;

	//update a time & reset data transferred over last period
		m_dwLastDataRateCalculationTime = dwCurTick;

		m_dwTransferredInLastPeriod = 0;
	}

	UpdateDisplayedInfo();

	return m_dwDownDataRate;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CUpDownClient::GetAvailablePartCount()
{
	EMULE_TRY

	uint16 result = 0;

	for (int i = 0;i < m_uPartCount;i++)
	{
		if (IsPartAvailable(i))
			result++;
	}

	return result;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetRemoteQueueRank(uint16 nr, bool bUpdateGUI /*= true*/)
{
	EMULE_TRY

	m_iDeltaQueueRank = static_cast<int>(nr) - static_cast<int>(m_uRemoteQueueRank);
	if (m_uRemoteQueueRank != nr)
	{
		m_uRemoteQueueRank = nr;
		if (bUpdateGUI)
			UpdateDisplayedInfo();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//eklmn: bugfix(08): A4AF switching
bool CUpDownClient::SwapToAnotherFile(CPartFile* SwapTo, bool bIgnoreNoNeeded)
{
	EMULE_TRY

//	Check if we are downloading something from client
	if (GetDownloadState() == DS_DOWNLOADING)
		return false;

	POSITION		SwapToPos = 0;
	POSITION		cur_FilePos;
	CPartFile	   *pCurFile;
	uint32			dwLastNNSCheckedTime = ::GetTickCount();
	const uint16	uPriorityFactorsArray[3] = {1,2,3};
	uint16			uA4AFRating;
	uint16			uBestA4AFRating = 0;

	//	Check: does client have another files?
	if (!m_otherRequestsList.IsEmpty())
	{
	//	Check: target file is specified? if is not, then find another one
		if (!SwapTo)
		{
			for (cur_FilePos = m_otherRequestsList.GetHeadPosition(); cur_FilePos != NULL; )
			{
				pCurFile = m_otherRequestsList.GetNext(cur_FilePos);
			//	Get first high priority files

				if ( pCurFile != m_pReqPartFile && g_eMuleApp.m_pDownloadQueue->IsInDLQueue(pCurFile)
				     && (pCurFile->GetStatus(false) == PS_READY || pCurFile->GetStatus(false) == PS_EMPTY) )
				{
				//	Check: is new file in NNS-Map?
					bool	bNNS = m_otherNoNeededMap.Lookup(pCurFile, dwLastNNSCheckedTime);

				//	Calculate an A4AF rating
					if (bNNS)
						uA4AFRating = uPriorityFactorsArray[pCurFile->GetPriority()]
							+ static_cast<uint16>((::GetTickCount() - dwLastNNSCheckedTime)/FILEREASKTIME);
					else
						uA4AFRating = uPriorityFactorsArray[pCurFile->GetPriority()] * 3;
					if (uA4AFRating > uBestA4AFRating)
					{
						SwapTo = pCurFile;
						SwapToPos = cur_FilePos ;
						uBestA4AFRating = uA4AFRating;
					}
				}
			} // end for
		} // list check
		else if (SwapTo->GetStatus(false) == PS_READY || SwapTo->GetStatus(false) == PS_EMPTY)
		{
			SwapToPos = m_otherRequestsList.Find(SwapTo); // otherwise get position with .Find()
		}

	//	now let's swap the file to another
		if (SwapTo != NULL && SwapToPos != 0)
		{

		//	to prevent loop swaping remember NNS-Part & time when it was switched
			if ( m_pReqPartFile != NULL && GetDownloadState() == DS_NONEEDEDPARTS)
			{
				m_otherNoNeededMap[m_pReqPartFile] = ::GetTickCount();
			}

		//	switch state to NONE
			SetDownloadState(DS_NONE);

			if (m_pbytePartStatuses)
			{
				delete[] m_pbytePartStatuses;
				m_pbytePartStatuses = 0;
			}

			m_uPartCount = 0;
			m_dwLastAskedTime = 0;
			m_uLastPartAsked = 0xFFFF;

		//	add to past comment
			if (m_pReqPartFile != NULL)
			{
				m_pReqPartFile->AddPastComment(this);
				m_eRating = PF_RATING_NONE;
				m_strComment = _T("");
			//	find current file position in sourceslots
				cur_FilePos = m_pReqPartFile->m_srcLists[m_iSourcesSlot].Find(this);

				if (cur_FilePos)
					m_pReqPartFile->m_srcLists[m_iSourcesSlot].RemoveAt(cur_FilePos); // remove current file

			//	add them in A4AF list
				m_pReqPartFile->m_A4AFsrcList.AddTail(this);

			//	add old file to m_otherRequestsList
				m_otherRequestsList.AddTail(m_pReqPartFile);

			//	let's update current information for the removed file
				m_pReqPartFile->NewSrcPartsInfo(); // eklmn: changes(00): NewSrcPartsInfo & UpdateAvailablePartsCount merged (CPU load)

			//	update download list (old file update)
				m_pReqPartFile->UpdateDisplayedInfo();

				m_pReqPartFile = NULL;	//reset before delete m_pbytePartStatuses
			}

		//	actual switch
			m_pReqPartFile = SwapTo;

		//	remove new file from OtherRequests list
			m_otherRequestsList.RemoveAt(SwapToPos);

		//	remove current source from A4AF list in new file
			POSITION	pos_A4AF = SwapTo->m_A4AFsrcList.Find(this);

			if (pos_A4AF != NULL)
				SwapTo->m_A4AFsrcList.RemoveAt(pos_A4AF);

		//	to prevent loop swaping remove a choosen part from NNS-Map
			m_otherNoNeededMap.RemoveKey(SwapTo);

		//	add to souceslots
			if (!SwapTo->m_srcLists[m_iSourcesSlot].Find(this))
				SwapTo->m_srcLists[m_iSourcesSlot].AddTail(this);

		//	remove from past comment
			SwapTo->RemovePastComment(this);	// #zegzav:pastcomment

		//	Update info for target file
			SwapTo->NewSrcPartsInfo();	// eklmn: changes(00): NewSrcPartsInfo & UpdateAvailablePartsCount merged (CPU load)

		//	update download list (new file update)
			SwapTo->UpdateDisplayedInfo();

		//	now update the client item
			UpdateDisplayedInfo(true);

			return true;
		}
	}

	EMULE_CATCH

	return false;
}
//eklmn: bugfix(08): A4AF switching
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskACK(uint16 nNewQR)
{
	EMULE_TRY

	m_bUDPPending = false;
	SetRemoteQueueRank(nNewQR);
	m_dwLastAskedTime = ::GetTickCount();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskFNF()
{
	EMULE_TRY

	m_bUDPPending = false;
//SyruS (0.29a) avoid premature deletion of 'this' client

	if (GetDownloadState() != DS_DOWNLOADING)
	{
		AddDebugLogLine(false, CString("UDP ANSWER FNF : %s"), GetUserName());
		//eklmn: try to swap the source to any file before remove them from DL-queue
		if (!SwapToAnotherFile(NULL))
			g_eMuleApp.m_pDownloadQueue->RemoveSource(this);
#ifdef OLD_SOCKETS_ENABLED

		if (!m_pRequestSocket)
			Disconnected();

#endif //OLD_SOCKETS_ENABLED

	}
	else
		AddDebugLogLine(false, CString("UDP ANSWER FNF : %s (client not removed)"), GetUserName());

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UDPReaskForDownload()
{
	EMULE_TRY

	ASSERT (m_pReqPartFile);

	if (!m_pReqPartFile || m_bUDPPending)
		return;

//	Maverick Disable UDP when connecting to a PROXY
	ProxySettings proxy = g_eMuleApp.m_pGlobPrefs->GetProxy();

#ifdef OLD_SOCKETS_ENABLED

	if ( m_byteEmuleVersion >= 0x23
		&& m_uUDPPort != 0
		&& g_eMuleApp.m_pGlobPrefs->GetUDPPort() != 0
		&& (!g_eMuleApp.m_pServerConnect->IsLowID() && !HasLowID() )
		&& !(m_pRequestSocket && m_pRequestSocket->IsConnected())
		&& (!proxy.m_bUseProxy) )
	{

	//don't use udp to ask for sources

		if (IsSourceRequestAllowed())
			return;

		m_bUDPPending = true;

	//	netwolf: complete sources UDP - BEGIN
		uint16 packetsize = 16;

		if (GetUDPVersion() >= 3)
			packetsize += 2;

	//	netwolf: complete sources UDP - END
		Packet		*response = new Packet(OP_REASKFILEPING, packetsize, OP_EMULEPROT);	// netwolf: complete sources UDP

		md4cpy(response->m_pcBuffer, m_pReqPartFile->GetFileHash());		// <filehash 16>

	//	netwolf: complete sources UDP - BEGIN
		if (GetUDPVersion() >= 3)
		{
			uint16		uNumCompleteSources = m_pReqPartFile->GetCompleteSourcesCount();

			memcpy2(response->m_pcBuffer + 16, &uNumCompleteSources, 2);	// <numcompletesources 2>
		}

	//	netwolf: complete sources UDP - END
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(response->m_dwSize);

		g_eMuleApp.m_pClientUDPSocket->SendPacket(response, GetIP(), GetUDPPort());
	}

#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Barry - Sets the vector of bool 'parts' to show the pending status of each part.
void CUpDownClient::ShowDownloadingParts(_Bvector& parts)
{
	EMULE_TRY

	parts.resize(m_uPartCount);

	for (POSITION pos = m_pendingBlocksList.GetHeadPosition(); pos != NULL; )
		parts[m_pendingBlocksList.GetNext(pos)->m_pRequestedBlock->m_dwStartOffset / PARTSIZE] = true;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UpdateDisplayedInfo(boolean force)
{
	g_eMuleApp.m_pDownloadList->UpdateSource(this);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	New functions by Mr.Fry
uint16 CUpDownClient::GetAvailableOfNeededPartsCount()
{
	EMULE_TRY

	uint16 uCount = 0;

	if (m_pbytePartStatuses != NULL && m_uPartCount > 0)
	{
		for (int i = 0; i < m_uPartCount ; i++)
			if (m_pbytePartStatuses[i] && !m_pReqPartFile->IsComplete(i * PARTSIZE, ((i + 1) * PARTSIZE) - 1))
				uCount++;

	//TRACE("Counted %u Chunks (parts) that %s has and I need (to download)",uCount,GetUserName());
	}

	return uCount;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef LATER
uint32 CUpDownClient::GetCurrentlyDownloadingPart()
{
	EMULE_TRY

	if (m_eDownloadState == DS_DOWNLOADING)
	{
		return		static_cast<uint32>(m_dwLastBlockOffset / PARTSIZE);
	}
	else
	{
		return 0xFFFF;
	}

	EMULE_CATCH

	return 0xFFFF;
}
#endif LATER
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ClearPendingBlocksList(void)
{
	EMULE_TRY

	Pending_Block_Struct*		pPendingBlock;

	while (!m_pendingBlocksList.IsEmpty() )
	{
		pPendingBlock = m_pendingBlocksList.RemoveHead();

		if (pPendingBlock)
		{
			if (pPendingBlock->m_pRequestedBlock)
			{
				CPartFile* pReqFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(pPendingBlock->m_pRequestedBlock->m_fileHash);

				pReqFile->RemoveBlockFromList(pPendingBlock->m_pRequestedBlock->m_dwStartOffset,
											pPendingBlock->m_pRequestedBlock->m_dwEndOffset);
				delete pPendingBlock->m_pRequestedBlock;
			}

			if (pPendingBlock->m_pZStream)
			{
				inflateEnd(pPendingBlock->m_pZStream);
				delete pPendingBlock->m_pZStream;
			}
			delete pPendingBlock;
		}
		pPendingBlock = NULL;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	If downloadBlocks are existing free block reserved in pPartFile
void CUpDownClient::ClearDownloadBlocksList(void)
{
	while (!m_downloadBlocksList.IsEmpty())
	{
		Requested_Block_Struct* pDLBlock =  m_downloadBlocksList.RemoveHead();
		if (pDLBlock)
		{
			CPartFile* pReqFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(pDLBlock->m_fileHash);

			pReqFile->RemoveBlockFromList(pDLBlock->m_dwStartOffset, pDLBlock->m_dwEndOffset);
			delete pDLBlock;
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
