// EmEngine.h: interface for the CEmEngine class.
//
//////////////////////////////////////////////////////////////////////

#define WM_TASK_UI WM_USER + 1000

#if !defined(AFX_EMENGINE_H__F77B635A_251D_4554_9D96_DACFCA968759__INCLUDED_)
#define AFX_EMENGINE_H__F77B635A_251D_4554_9D96_DACFCA968759__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "../Loggable.h"
#include "Sockets/TaskProcessorSockets.h"
#include "Other/TaskProcessorLogger.h"
#include "Sockets/TcpEngine.h"

class CServer;

enum EnumServerStates
{
	SERVER_DISCONNECTED,
	SERVER_CONNECTING,
	SERVER_WAITFORLOGIN,
	SERVER_CONNECTED
};

struct ServerState
{
	SOCKET		hSocket;
	ULONG		nAddr;
	USHORT		nPort;
	short		nConnState;
	long		nClientID;
};

class CEmEngine : public CLoggable
{
	HANDLE m_hStop; // termination event

	HWND m_hWndUI;

public:

	CEmEngine();
	~CEmEngine() { Uninit(); }

	bool Init(HWND hWndUI);
	void Uninit();

	bool InitializeDatabases();
	void CloseDatabases();

	// Public methods
	void ParseXmlCommand(CString sXml);

	void ConnectToServer(CString sAddr = "", ULONG ulPort = 0);
	void ConnectToAnyServer();
	void DisconnectFromServer(SOCKET hPrevServer = NULL);

	void SendSharedListToServer();

	// Data
	bool IsConnecting();
	bool IsConnected();
	SOCKET GetServerSocket(){ return m_stServer.hSocket; }
	ServerState* GetConnectedState(){ return &m_stServer; }
	long GetClientID(){ return m_stServer.nClientID; }
	bool IsLocalServer(ULONG nAddr, USHORT uPort);
	bool IsLowID();

	// Global preferences
	BYTE* GetUserHash();
	USHORT GetPort();
	CString GetUserNick();

	// Update Server Info
	void SetConnectionState(short nConnectionState);
	void SetClientID(long cID);
	void ConnectedTo(CEmClient_Server* pClient);

	inline bool IsShuttingDown() { return WAIT_TIMEOUT != WaitForSingleObject(m_hStop, 0); }

/*	inline void PushToUI(CTask* pTask)
	{
		if (pTask && !PostMessage(m_hWndUI, WM_TASK_UI, (WPARAM) pTask, 0))
		{
			ASSERT(FALSE);
			delete pTask;
		}
	}*/
	inline void PushToSocketsTP(CTask *pTask)
	{
		m_stSocketsTP.Push(pTask);
	}
	inline void PushToLogger(CTask *pTask)
	{
		m_stLogger.Push(pTask);
	}
	inline void KillClient(CEmClient *pClient)
	{
		m_stSocketsTP.KillClient(pClient);
	}
	inline bool SendOpCode(SOCKET hSocket, const COpCode &stOpCode, CEmClient* pClient)
	{
		if (pClient)
			m_stLogger.Post(stOpCode, pClient->GetType(), pClient->m_nClientID, TRUE);
		return m_stTcp.AllocSend(hSocket, stOpCode);
	}

	DbEnv	*m_pDbEnv;				// Database environment
	Db		*m_pDbServers;			// Servers list Db
	Db		*m_pDbServersAddr;		// Servers list primary index by addr/port/dynIP
	Db		*m_pDbServersStatic;	// Servers list secondary index by static
	Db		*m_pDbServersPriority;	// Servers list secondary index by priority

private:
	// Task processors
	CTaskProcessor_Main		m_stMainTP;
	CTaskProcessor_Sockets	m_stSocketsTP;
	CTaskProcessor_Logger	m_stLogger;
	CTcpEngine				m_stTcp;

	// Low-level classes might have want to access our private vars
	friend struct CTask_Tcp_Err;
	friend struct CTask_Tcp_Accepted;
	friend struct CTask_Tcp_Connected;
	friend struct CTask_Tcp_Web;
	friend struct CTask_Connect;
	friend struct CEmClient_Web;
	friend class CTaskProcessor_Sockets;
	friend class COpCode;
	friend struct CTask_LogOpcode;

	ServerState	m_stServer;
};

#endif // !defined(AFX_EMENGINE_H__F77B635A_251D_4554_9D96_DACFCA968759__INCLUDED_)
