#include "StdAfx.h"
#include "resource.h"
#include "OtherFunctions.h"
#include ".\keyboardshortcut.h"


short GetCodeFromPressedKeys(MSG* pMsg)
{
	short nCode = 0;

	nCode += MASK_ENABLED;
	if (GetAsyncKeyState(VK_MENU) < 0)
		nCode += MASK_ALT;
	if (GetAsyncKeyState(VK_CONTROL) < 0)
		nCode += MASK_CTRL;
	if (GetAsyncKeyState(VK_SHIFT) < 0)
		nCode += MASK_SHIFT;
	if ((GetAsyncKeyState(VK_LWIN) < 0) || (GetAsyncKeyState(VK_RWIN) < 0))
		nCode += MASK_WIN;
	nCode += static_cast<short>(pMsg->wParam);

	return nCode;
}

CString GetStringFromShortcutCode(short nCode, EnumShortcutStringPrefix ePrefixType, BOOL bReturnNoneIfDisabled)
{
	CString String;
	String = _T("");

	if ((nCode & MASK_ENABLED) != 0) 
	{
		if ((nCode & MASK_ALT) != 0)
			String += _T("Alt+");
		if ((nCode & MASK_CTRL) != 0)
			String += _T("Ctrl+");
		if ((nCode & MASK_SHIFT) != 0)
			String += _T("Shift+");
		if ((nCode & MASK_WIN) != 0)
			String +=_T("Win+");

		uchar byteKey = nCode & MASK_KEY;
		if (byteKey == VK_RETURN)	// Enter key
			String += GetResString(IDS_SHORTCUTS_ENTER_KEY);
		else if (byteKey == VK_SPACE) // Spacebar
			String += GetResString(IDS_SHORTCUTS_SPACE_KEY);
		else if (byteKey == VK_TAB) // Tab key
			String += GetResString(IDS_SHORTCUTS_TAB_KEY);
		else if (byteKey == VK_BACK) // Bakcspace key
			String += GetResString(IDS_SHORTCUTS_BACK_KEY);
		else if (byteKey == VK_INSERT) // Insert key
			String += GetResString(IDS_SHORTCUTS_INSERT_KEY);
		else if (byteKey == VK_DELETE) // Delete key
			String += GetResString(IDS_SHORTCUTS_DELETE_KEY);
		else if (byteKey == VK_HOME) // Home key
			String += GetResString(IDS_SHORTCUTS_HOME_KEY);
		else if (byteKey == VK_END) // End key
			String += GetResString(IDS_SHORTCUTS_END_KEY);
		else if (byteKey == VK_PRIOR) // Page Up key
			String += GetResString(IDS_SHORTCUTS_PAGEUP_KEY);
		else if (byteKey == VK_NEXT) // Page Down key
			String += GetResString(IDS_SHORTCUTS_PAGEDOWN_KEY);
		else if ((byteKey >= VK_F1) && (byteKey <= VK_F12)) // function keys
			String.AppendFormat(_T("F%i"), (byteKey-VK_F1)+1);
		else if ((byteKey >= 'A') && (byteKey <= 'Z')) // letter keys
			String.AppendChar(byteKey);
		else if ((byteKey >= '0') && (byteKey <= '9')) // number keys
			String.AppendChar(byteKey);
		else /*** Should never happen ***/
			String = _T("ERROR");
	}
	else if (bReturnNoneIfDisabled)
		String = GetResString(IDS_SHORTCUTS_NONE);

	if (!String.IsEmpty())
	{
		if (ePrefixType == SSP_TAB_PREFIX)
			return (_T("\t") + String);
		else if (ePrefixType == SSP_SPACE_PREFIX)
			return (_T(" ") + String);
	}
	
	return String;
}


CKeyboardShortcut::CKeyboardShortcut(void)
: m_bEnabled(false)
, m_bAlt(false)
, m_bCtrl(false)
, m_bShift(false)
, m_bWin(false)
, m_byteKey(VK_RETURN)
{
}

CKeyboardShortcut::CKeyboardShortcut(short nCode)
{
	SetCode(nCode);
}

CKeyboardShortcut::~CKeyboardShortcut(void)
{
}

void CKeyboardShortcut::SetCode(short nCode)
{
	m_bEnabled	= static_cast<bool>(nCode & MASK_ENABLED);
	m_bAlt		= static_cast<bool>(nCode & MASK_ALT);
	m_bCtrl		= static_cast<bool>(nCode & MASK_CTRL);
	m_bShift	= static_cast<bool>(nCode & MASK_SHIFT);
	m_bWin		= static_cast<bool>(nCode & MASK_WIN);
	m_byteKey	= static_cast<uchar>(nCode & MASK_KEY);
}
CString CKeyboardShortcut::GetKeyComboString(void)
{
	return GetStringFromShortcutCode(GetCode());
}
