//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//	ListenSocket.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ListenSocket.h"
#include "opcodes.h"
#include "KnownFile.h"
#include "SharedFileList.h"
#include "UploadQueue.h"
#include "updownclient.h"
#include "ClientList.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

#ifdef OLD_SOCKETS_ENABLED

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CClientReqSocket
CClientReqSocket::CClientReqSocket(CPreferences* in_prefs, CUpDownClient* in_client)
{
	m_pPrefs = in_prefs;
	m_pClient = in_client;
	if (in_client)
		m_pClient->m_pRequestSocket = this;
	g_eMuleApp.m_pListenSocket->AddSocket(this);
	ResetTimeOutTimer();
	m_bDeleteThis = false;
	m_dwDeleteTimer = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientReqSocket::~CClientReqSocket()
{
	if (m_pClient != NULL)
		m_pClient->m_pRequestSocket = NULL;
	m_pClient = NULL;
	g_eMuleApp.m_pListenSocket->RemoveSocket(this);

	DEBUG_ONLY (g_eMuleApp.m_pClientList->Debug_SocketDeleted(this));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::ResetTimeOutTimer()
{
	m_dwTimeoutTimer = ::GetTickCount();
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CheckTimeOut() checks the timeout timer, and, if it has expired, it disconnects the m_pRequestSocket
//	and resets the timeout timer. It returns true if the timer has expired, false if it has not.
bool CClientReqSocket::CheckTimeOut()
{
//	If the timeout period has been exceeded...
	if (::GetTickCount() - m_dwTimeoutTimer > CONNECTION_TIMEOUT)
	{
	//	Reset the timeout timer
		m_dwTimeoutTimer = ::GetTickCount();
		Disconnect();
		return true;
	}
	return false;
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::OnClose(int nErrorCode)
{
	ASSERT (g_eMuleApp.m_pListenSocket->IsValidSocket(this));
	CEMSocket::OnClose(nErrorCode);
	Disconnect();
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::Disconnect()
{
//	Tell the m_pRequestSocket not to receive any more notifications
	AsyncSelect(0);
	m_eConnectionState = ES_DISCONNECTED;
	if (m_pClient != NULL && !g_eMuleApp.m_pClientList->IsValidClient(m_pClient)) // obaldin: a quick fix for a strange bug
	{
		AddDebugLogLine(false, _T("Invalid client in CClientReqSocket::Disconnect()"));
		m_pClient = NULL;
	}
	if (m_pClient == NULL)
	{
		Safe_Delete();
	}
	else
	{
		if (m_pClient->Disconnected())
			m_pClient = NULL;
	}
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::TimedDelete()
{
//	It seems that MFC Sockets call socketfunctions after they are deleted, even if the m_pRequestSocket is closed
//	and select(0) is set. So we need to wait some time to make sure this doesn't happen.
	if (::GetTickCount() - m_dwDeleteTimer > 10000)
		delete this;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::Safe_Delete()
{
	ASSERT (g_eMuleApp.m_pListenSocket->IsValidSocket(this));
	AsyncSelect(0);
	m_dwDeleteTimer = ::GetTickCount();
	if (m_SocketData.hSocket != INVALID_SOCKET) // deadlake PROXYSUPPORT - changed to AsyncSocketEx
		ShutDown(2);
	if (m_pClient != NULL)
		m_pClient->m_pRequestSocket = NULL;
	m_pClient = NULL;
	m_eConnectionState = ES_DISCONNECTED;
	m_bDeleteThis = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientReqSocket::ProcessPacket(char *pcPacketBuf, uint32 dwPacketSize, EnumOpcodes eOpcode)
{
	EnumDLQState enumClientDLState;

	try
	{
		try
		{
			if (m_pClient == NULL && eOpcode != OP_HELLO)
			{
				throw GetResString(IDS_ERR_NOHELLO);
			}

			switch (eOpcode)
			{
				case OP_HELLOANSWER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					m_pClient->ProcessHelloAnswer(pcPacketBuf, dwPacketSize);
					if (m_pClient)
					{
						m_pClient->ConnectionEstablished();
						g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(m_pClient);
					}
					break;
				}
				case OP_HELLO:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					bool newClient = false;
					if (!m_pClient)
					{
					//	Create new client to save standard informations
						m_pClient = new CUpDownClient(this);
						newClient = true;
					}
				//	eklmn: originally we had try & catch here, but cause our EMULE_TRY & EMULE_CATCH
				//	function in ProcessHelloPacket, local trycatch did work. That means even in case of exception
				//	we process the client without full information... so i changed a function to get some info back
					if (!m_pClient->ProcessHelloPacket(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize))
					{
						if (newClient)
							safe_delete(m_pClient);
						throw CString(_T("Unable to process HelloPacket"));
					}
				//	now we check if we now this client already. if yes this m_pRequestSocket will
				//	be attached to the known client, the new client will be deleted
				//	and the var. "client" will point to the known client.
				//	if not we keep our new-constructed client ;)
					if (g_eMuleApp.m_pClientList->AttachToAlreadyKnown(&m_pClient, this))
					{
					//	update the old client informations
						m_pClient->ProcessHelloPacket(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);
						m_pClient->DisableL2HAC(); //<<-- enkeyDEV(th1) -L2HAC- lowid side
					}
					else
					{
						//eklmn: if new client was not created delete him
						if (!g_eMuleApp.m_pClientList->AddClient(m_pClient))
						{
							safe_delete(m_pClient);
							throw CString(_T("Unable to add client in ClientList"));
						}
						m_pClient->SetCommentDirty();
					}

					g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(m_pClient);

				//	send a response pcPacketBuf with standard informations
					if (m_pClient->GetClientSoft() == SO_EMULE && m_pClient->GetHashType() == SO_EMULE)
						m_pClient->SendMuleInfoPacket(false);

					m_pClient->SendHelloAnswer();
					if (m_pClient)
						m_pClient->ConnectionEstablished();

					break;
				}
				case OP_FILEREQUEST:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);

				//	IP banned, no answer for this request
					if (m_pClient->IsBanned())
					{
						break;
					}

				//	If we're filtering scanning (and misbehaving) clients and this client has asked for a
				//		non-existant file more than three times...
					if (g_eMuleApp.m_pGlobPrefs->IsScanFilterEnabled() && m_pClient->GetFailedFileRequests() >= 3)
					{
					//	Temporarily ban the client from the upload queue
						g_eMuleApp.m_pIPFilter->AddTemporaryBannedIP(m_pClient->GetIP());
					//	If we're logging countermeasures...
						if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
							AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Client %s (%s) added to filtered clients due to file scanning"), m_pClient->GetFullIP(), GetClientNameAndVersionString(m_pClient));
						Disconnect();
						break;
					}
				//	Since we don't process extended information it is better for compatibility with rest of the network
				//		to always process a filehash & extended information only if the file exists and was activated by tag ET_EXTENDEDREQUEST

					if (dwPacketSize >= 16)
					{
						if (!m_pClient->GetWaitStartTime())
							m_pClient->SetWaitStartTime();

						uchar		requestedFileHash[16];

						md4cpy(requestedFileHash, pcPacketBuf);

						CKnownFile	*m_pReqSharedFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(requestedFileHash);

					//	If we are not sharing the requested file...
						if (!m_pReqSharedFile)
						{
						//
						//	If we've just started a download we may want to use that client as a source.
						//		Moreover we need to check if the requested file exists in the DL queue
						//		in order to filter the clients correct way.

						//	Look for the requested file in the download queue
							CPartFile		*pPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(requestedFileHash);

						//	If we've found the file and it hasn't completed...
							if (pPartFile != NULL && pPartFile->IsPartFile())
							{
							//	... and we don't already have enough sources...
								if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > pPartFile->GetSourceCount())
								{
									g_eMuleApp.m_pDownloadQueue->CheckAndAddKnownSource(pPartFile, m_pClient);
								}
							}
						//	If we're not currently downloading the requested file...
							else
							{
								m_pClient->UpdateFailedFileRequests();
							}

						//	If the client is using our user hash (hash stealer)...
							if (md4cmp(g_eMuleApp.m_pGlobPrefs->GetUserHash(), m_pClient->GetUserHash()) == 0)
								AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Client %s (%s) is using our userhash"), m_pClient->GetUserName(), m_pClient->GetFullIP());

						/*
						eklmn: removed at 12.12.2003 in order to save some traffic and not send FILEREQANSNOFIL twice
						//	send file request no such file pcPacketBuf (0x48)
							Packet * replypacket = new Packet(OP_FILEREQANSNOFIL, 16);
							md4cpy(replypacket->m_pcBuffer, pcPacketBuf);
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(replypacket->m_dwSize);
							SendPacket(replypacket, true);
						*/
						}
					//	If we are sharing the requested file...
						else
						{
						//	... and we're downloading the file... (this could be a new source)
							if (m_pReqSharedFile->IsPartFile())
							{
							//	... and we don't already have enough sources...
								if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > static_cast<CPartFile*>(m_pReqSharedFile)->GetSourceCount())
								{
									g_eMuleApp.m_pDownloadQueue->CheckAndAddKnownSource((CPartFile*)m_pReqSharedFile, m_pClient);
								}
							}

						//	If this client is requesting this file for the first time...
							if (md4cmp(m_pClient->m_reqFileHash, pcPacketBuf) != 0)
							{
								m_pClient->SetCommentDirty();
							}
						//	Send filename etc
							m_pClient->ResetFailedFileRequests();
							m_pClient->SetUploadFileID(reinterpret_cast<BYTE*>(pcPacketBuf));

							CSafeMemFile		packetStream(128);

							packetStream.Write(m_pReqSharedFile->GetFileHash(), 16);

							uint16		uNameLength = _tcslen(m_pReqSharedFile->GetFileName());

							packetStream.Write(&uNameLength, 2);
							packetStream.Write(m_pReqSharedFile->GetFileName(), uNameLength);
						//	TODO: Don't let 'ProcessUpFileStatus' re-process the entire pcPacketBuf and search the fileid
						//	again in 'm_pSharedFilesList' -> waste of time.
							m_pClient->ProcessUpFileStatus(pcPacketBuf, dwPacketSize);

							Packet		*pcPacketBuf = new Packet(&packetStream);

							pcPacketBuf->m_eOpcode = OP_FILEREQANSWER;
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pcPacketBuf->m_dwSize);
							SendPacket(pcPacketBuf, true);
							m_pClient->SendCommentInfo(m_pReqSharedFile);
						}
					}
				//	If the packet is too small (malformed)...
					else
					{
						m_pClient->UpdateFailedFileRequests();
						throw CString(GetResString(IDS_ERR_WRONGPACKAGESIZE) + _T(" (OP_FILEREQUEST)"));
					}
					break;
				}
				case OP_FILEREQANSNOFIL:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					if (dwPacketSize == 16)
					{
					//	if that client do not have my file maybe has another different
						CPartFile* m_pReqPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID((uchar*)pcPacketBuf);
						if (!m_pReqPartFile)
							break;

						//eklmn: do nothing if none-file was requested
						if (!m_pClient->m_pReqPartFile)
							break;

					//	we try to swap to another file ignoring no needed parts files
						if (m_pClient)
							switch (m_pClient->GetDownloadState())
							{
								case DS_CONNECTED:
								case DS_ONQUEUE:
								case DS_NONEEDEDPARTS:
								if (!m_pClient->SwapToAnotherFile(NULL))
								{	//eklmn: bugfix(08): A4AF switching (???)
									g_eMuleApp.m_pDownloadQueue->RemoveSource(m_pClient, true);
								}
								break;
							}
						break;
					}
					throw CString(GetResString(IDS_ERR_WRONGPACKAGESIZE) + _T(" (OP_FILEREQANSNOFIL)"));
					break;
				}
				case OP_FILEREQANSWER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					m_pClient->ProcessFileInfo(pcPacketBuf, dwPacketSize);
					break;
				}
				case OP_FILESTATUS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					m_pClient->ProcessFileStatus(pcPacketBuf, dwPacketSize);
					break;
				}
				case OP_STARTUPLOADREQ:
				// eklmn: the old eMule clients (<0.26) as well as original eDonkey don't send a hash (dwPacketSize is 0)
				// so we gonna add client to UL-queue anyway, but inside AddClientToQueue() will be checked
				// if any file was requested before.
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					if (dwPacketSize == 16)
					{
						uchar requestedFileHash[16];
						md4cpy(requestedFileHash, pcPacketBuf);
					//check if file exits
						CKnownFile* m_pReqSharedFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(requestedFileHash);
						if (m_pReqSharedFile)
						{
						//	if wer are downloading this file, this could be a new source
							if (m_pReqSharedFile->IsPartFile())
								if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > ((CPartFile*)m_pReqSharedFile)->GetSourceCount()) //<<--
									g_eMuleApp.m_pDownloadQueue->CheckAndAddKnownSource((CPartFile*)m_pReqSharedFile, m_pClient);

						//	check to see if this is a new file they are asking for
							if (md4cmp(m_pClient->GetUploadFileID(), pcPacketBuf) != 0)
								m_pClient->SetCommentDirty();

							m_pClient->ResetFailedFileRequests();
							m_pClient->SetUploadFileID((uchar*)pcPacketBuf);
							m_pClient->SendCommentInfo(m_pReqSharedFile);
						}
						else
						{
						//eklmn: if client still asking for not existing, he is definitely violating eDonkey protocol
							if (!g_eMuleApp.m_pDownloadQueue->GetFileByID(requestedFileHash))
								m_pClient->UpdateFailedFileRequests();

						//	Reply that requested file is not exist
							Packet * replypacket = new Packet(OP_FILEREQANSNOFIL, 16);
							md4cpy(replypacket->m_pcBuffer, pcPacketBuf);
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(replypacket->m_dwSize);
							SendPacket(replypacket, true);
							break;
						}
					}
					g_eMuleApp.m_pUploadQueue->AddClientToQueue(m_pClient);
					break;
				}
				case OP_QUEUERANK:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					uint32 rank;
					memcpy2(&rank, pcPacketBuf, 4);
					m_pClient->SetRemoteQueueRank(rank);
				//	eklmn: switch DL state in case newrank != 0
				//	if (rank && m_pClient->GetDownloadState() == DS_DOWNLOADING)
				//		m_pClient->SetDownloadState(DS_ONQUEUE);
					break;
				}
				case OP_ACCEPTUPLOADREQ:
				{
					enumClientDLState = m_pClient->GetDownloadState();
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
				//	modified by sivka [dont download if paused]
					if (m_pClient->m_pReqPartFile && g_eMuleApp.m_pDownloadQueue->IsInDLQueue(m_pClient->m_pReqPartFile))
					{
					//	If the client is in our download queue and the part file is downloading or waiting but
					//		isn't paused...
					//	DbT: ask for next block while downloading
						if (enumClientDLState == DS_DOWNLOADING)
						{
						//	eklmn: sometimes remote client sends us OP_ACCEPTUPLOADREQ during DS_DOWNLOADING state
						//	to prevent desynconization of request we gonna clear the list of pending bloks
							m_pClient->ClearPendingBlocksList();
							m_pClient->SendBlockRequests();
						}
					//	eklmn: ask while connecting, but don't ask for paused files
						else if ( (enumClientDLState == DS_CONNECTED || enumClientDLState == DS_ONQUEUE)
								&& !m_pClient->m_pReqPartFile->IsPaused()
								&& (m_pClient->m_pReqPartFile->GetStatus(false) == PS_READY || m_pClient->m_pReqPartFile->GetStatus(false) == PS_EMPTY))
						{
							m_pClient->SetDownloadState(DS_DOWNLOADING);
							m_pClient->SetLastDownPartAsked(0xFFFF); // eklmn: feature(00): ICR 2.3 reset part before new download
							m_pClient->SendBlockRequests();
						}
						else // added by sivka [dont download if paused]
						{
							Packet		*pcPacketBuf = new Packet(OP_CANCELTRANSFER);

							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pcPacketBuf->m_dwSize);
							m_pClient->m_pRequestSocket->SendPacket(pcPacketBuf, true, true); //eklmn: bugfix(05): Listen socket OP_ACCEPTUPLOADREQ
							m_pClient->SetDownloadState(DS_ONQUEUE);
						}
					}
					else
						throw CString(GetResString(IDS_ERR_WRONGFILEID) + _T(" (OP_ACCEPTUPLOADREQ => !m_pReqPartFile)"));
					break;
				}
				case OP_REQUESTPARTS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);

					CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(pcPacketBuf), dwPacketSize);

					uchar		reqfilehash[16];

					packetStream.Read(reqfilehash, 16);

					//reset previous value due to safety reason
					memzero(&m_sReqBlock1, sizeof(Requested_Block_Struct));
					memzero(&m_sReqBlock2, sizeof(Requested_Block_Struct));
					memzero(&m_sReqBlock3, sizeof(Requested_Block_Struct));
					
					packetStream.Read(&m_sReqBlock1.m_dwStartOffset, 4);
					packetStream.Read(&m_sReqBlock2.m_dwStartOffset, 4);
					packetStream.Read(&m_sReqBlock3.m_dwStartOffset, 4);
					packetStream.Read(&m_sReqBlock1.m_dwEndOffset, 4);
					packetStream.Read(&m_sReqBlock2.m_dwEndOffset, 4);
					packetStream.Read(&m_sReqBlock3.m_dwEndOffset, 4);
					md4cpy(&m_sReqBlock1.m_fileHash, reqfilehash);
					md4cpy(&m_sReqBlock2.m_fileHash, reqfilehash);
					md4cpy(&m_sReqBlock3.m_fileHash, reqfilehash);
					
					if (m_sReqBlock1.m_dwEndOffset - m_sReqBlock1.m_dwStartOffset != 0)
					{
						m_pClient->AddReqBlock(&m_sReqBlock1);
					}
					if (m_sReqBlock2.m_dwEndOffset - m_sReqBlock2.m_dwStartOffset != 0)
					{
						m_pClient->AddReqBlock(&m_sReqBlock2);
					}
					if (m_sReqBlock3.m_dwEndOffset - m_sReqBlock3.m_dwStartOffset != 0)
					{
						m_pClient->AddReqBlock(&m_sReqBlock3);
					}
					break;
				}
				case OP_CANCELTRANSFER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(m_pClient, ETS_CANCELED);
					g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, _T("%s: Upload session ended due to a cancelled transfer."), m_pClient->GetUserName());
					m_pClient->SetUploadFileID(NULL);
					break;
				}
				case OP_END_OF_DOWNLOAD:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					if (dwPacketSize >= 16 && !md4cmp(m_pClient->GetUploadFileID(), pcPacketBuf))
					{
						g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(m_pClient, ETS_END_OF_DOWNLOAD);
						AddDebugLogLine(false, _T("%s: Upload session ended due to a completed transfer."), m_pClient->GetUserName());
						m_pClient->SetUploadFileID(NULL);
					}
					break;
				}
				case OP_HASHSETREQUEST:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					if (dwPacketSize != 16)
						throw GetResString(IDS_ERR_WRONGHPACKAGESIZE);
					m_pClient->SendHashsetPacket(pcPacketBuf);
					break;
				}
				case OP_HASHSETANSWER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
				//if (m_pClient->GetDownloadState() != DS_REQHASHSET)
				//	throw CString("unwanted hashset");
					m_pClient->ProcessHashSet(pcPacketBuf, dwPacketSize);
					break;
				}
				case OP_SENDINGPART:	// i.e source sending part to us
				{
					m_pClient->UpdateLastBlockReceivedTime();
				//	Before we process the packet we need to check the client's download state.
				//	If the source sends us a packet while we're in an unwanted state (we
				//	haven't made a request) we switch him to an error state.
					enumClientDLState = m_pClient->GetDownloadState();
				//	If we're uploading to this client or there are already too many connections...
					if (enumClientDLState == DS_DOWNLOADING || enumClientDLState == DS_TOOMANYCONNS)
					{
					//	If we have made a part file request of this client and the requested file is Waiting...
						if ( m_pClient->m_pReqPartFile != NULL
						  && ( m_pClient->m_pReqPartFile->GetStatus() == PS_READY
						    || m_pClient->m_pReqPartFile->GetStatus() == PS_EMPTY ) )
						{
							m_pClient->ProcessBlockPacket(pcPacketBuf, dwPacketSize);
						//	If the file is now in a stopped state (paused/stopped/error)...
							if ( (m_pClient->m_pReqPartFile != NULL)
							  && ( m_pClient->m_pReqPartFile->GetStatus() == PS_PAUSED
							    || m_pClient->m_pReqPartFile->GetStatus() == PS_STOPPED
							    || m_pClient->m_pReqPartFile->GetStatus() == PS_ERROR ) )
							{
								Packet		*pPacket = new Packet(OP_CANCELTRANSFER);

								g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
								m_pClient->m_pRequestSocket->SendPacket(pPacket, true, true);
							}
						}
					//	If we didn't want this file (or don't now)...
						else
						{
							Packet		*pPacket = new Packet(OP_CANCELTRANSFER);

							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
							m_pClient->m_pRequestSocket->SendPacket(pPacket, true, true);
							m_pClient->SetDownloadState(m_pClient->m_pReqPartFile == NULL ? DS_NONE : DS_ONQUEUE);
						}
					}
					else
						m_pClient->SetDownloadState(DS_ERROR);
					break;
				}
				case OP_OUTOFPARTREQS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);
					if (m_pClient->GetDownloadState() == DS_DOWNLOADING)
					{
						m_pClient->SetDownloadState(DS_ONQUEUE);
					//	this request ends in an endless loop !!!
					//	OP_REQUESTPARTS <-> OP_OUTOFPARTREQS
						//m_pClient->SendBlockRequests();
					}
					break;
				}
				case OP_SETREQFILEID:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadFileRequest(dwPacketSize);

					if (m_pClient->GetFailedFileRequests() >= 3 && g_eMuleApp.m_pGlobPrefs->IsScanFilterEnabled())
					{
						g_eMuleApp.m_pIPFilter->AddTemporaryBannedIP(m_pClient->GetIP());
						if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
							AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Client %s (%s) added to filtered clients due to file scanning (%s)"), m_pClient->GetFullIP(), GetClientNameAndVersionString(m_pClient));
						Disconnect();
						break;
					}

					if (dwPacketSize == 16)
					{
						if (!m_pClient->GetWaitStartTime())
							m_pClient->SetWaitStartTime();

						CKnownFile* m_pReqSharedFile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)pcPacketBuf);

						if (!m_pReqSharedFile)
						{
						//	If we've just started a download we may want to use that client as a source.
						//  Moreover we need to check if the requested file exists in the DL queue
						//  in order to filter the clients the correct way.
							CPartFile* pPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID((uchar*)pcPacketBuf);
							if (pPartFile && pPartFile->IsPartFile())
							{
								if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > pPartFile->GetSourceCount()) //<<--
									g_eMuleApp.m_pDownloadQueue->CheckAndAddKnownSource(pPartFile, m_pClient);
							}
							else
								m_pClient->UpdateFailedFileRequests();

						//	send file request no such file pcPacketBuf (0x48)
							Packet* replypacket = new Packet(OP_FILEREQANSNOFIL, 16);
							md4cpy(replypacket->m_pcBuffer, pcPacketBuf);
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(replypacket->m_dwSize);
							SendPacket(replypacket, true);
							break;
						}

					//	eklmn: reset Failed Request Counter
						m_pClient->ResetFailedFileRequests();
					//	if we are downloading this file, this could be a new source
						if (m_pReqSharedFile->IsPartFile())
							if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > ((CPartFile*)m_pReqSharedFile)->GetSourceCount())
								g_eMuleApp.m_pDownloadQueue->CheckAndAddKnownSource((CPartFile*)m_pReqSharedFile, m_pClient);

					//	check to see if this is a new file they are asking for
						if (md4cmp(m_pClient->GetUploadFileID(), pcPacketBuf) != 0)
							m_pClient->SetCommentDirty();

					//send filestatus
						m_pClient->SetUploadFileID((uchar*)pcPacketBuf);
					//md4cpy(m_pClient->m_reqFileHash,pcPacketBuf);
						CSafeMemFile packetStream(16 + 16);
						packetStream.Write(m_pReqSharedFile->GetFileHash(), 16);
						if (m_pReqSharedFile->IsPartFile())
							((CPartFile*)m_pReqSharedFile)->WritePartStatus(&packetStream);

						else if (m_pReqSharedFile->GetJumpstartEnabled())
							m_pReqSharedFile->WriteJumpstartPartStatus(m_pClient, &packetStream);

						else if (m_pReqSharedFile->HasHiddenParts())
							m_pReqSharedFile->WritePartStatus(&packetStream);

						else
						{
							uint32 null = 0;
							packetStream.Write(&null, 3);
						}

						Packet* pcPacketBuf = new Packet(&packetStream);
						pcPacketBuf->m_eOpcode = OP_FILESTATUS;
						g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pcPacketBuf->m_dwSize);
						SendPacket(pcPacketBuf, true);
					}
					else
						throw CString(GetResString(IDS_ERR_WRONGPACKAGESIZE) + _T(" (OP_SETREQFILEID)"));
					break;
				}
				case OP_CHANGE_CLIENT_ID:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					CSafeMemFile packetStream((BYTE*)pcPacketBuf, dwPacketSize);
					uint32 nNewUserID;
					packetStream.Read(&nNewUserID, 4);
					uint32 nNewServerIP;
					packetStream.Read(&nNewServerIP, 4);
					if (nNewUserID < 0x1000000)
					{ // client changed server and gots a LowID
						CServer * pNewServer = g_eMuleApp.m_pServerList->GetServerByIP(nNewServerIP);
						if (pNewServer != NULL)
						{
							m_pClient->SetUserID(nNewUserID); // update UserID only if we know the server
							m_pClient->SetServerIP(nNewServerIP);
							m_pClient->SetServerPort(pNewServer->GetPort());
						}
					}
					else if (nNewUserID == m_pClient->GetIP())
					{ // client changed server and gots a HighID(IP)
						m_pClient->SetUserID(nNewUserID);
						CServer* pNewServer = g_eMuleApp.m_pServerList->GetServerByIP(nNewServerIP);
						if (pNewServer != NULL)
						{
							m_pClient->SetServerIP(nNewServerIP);
							m_pClient->SetServerPort(pNewServer->GetPort());
						}
					}
					break;
				}

				case OP_CHANGE_SLOT:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
				//	sometimes sent by Hybrid
					break;
				}
				case OP_MESSAGE:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);

					uint16		uMessageLen;
					memcpy2(&uMessageLen, pcPacketBuf, 2);
				//
				//	Verify the packet size
					if (uMessageLen + 2 != dwPacketSize)
						throw CString("invalid message pcPacketBuf");

					char		*strMessage = new char[uMessageLen + 1];
					memcpy2(strMessage, pcPacketBuf + 2, uMessageLen);
					strMessage[uMessageLen] = 0;

					g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.ProcessMessage(m_pClient, strMessage);
					delete[] strMessage;
					break;
				}
				case OP_ASKSHAREDFILES:
				{
				//	Client wants to know what we have in share, let's see if we allow him to know that
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);

				//	IP banned, no answer for this request
					if (m_pClient->IsBanned())
					{
						break;
					}

					CPtrList	list;

					if ( g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_EVERYBODY
					  || (g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_FRIENDS && m_pClient->IsFriend()) )
					{
						CCKey		bufKey;
						CKnownFile	*cur_file;

						for (POSITION pos = g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetStartPosition(); pos != NULL;)
						{
							g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetNextAssoc(pos, bufKey, cur_file);
							list.AddTail((void*&)cur_file);
						}
						AddDebugLogLine(false, IDS_REQ_SHAREDFILES, m_pClient->GetUserName(), m_pClient->GetUserID(), GetResString(IDS_ACCEPTED));
					}
					else
					{
						AddDebugLogLine(false, IDS_REQ_SHAREDFILES, m_pClient->GetUserName(), m_pClient->GetUserID(), GetResString(IDS_DENIED));
					}
				//
				//	Build the reply packet and send it.
					uint32			dwNumFiles = list.GetCount();
					CSafeMemFile	packetStream(80);

					packetStream.Write(&dwNumFiles, 4);
					while (list.GetCount())
					{
						g_eMuleApp.m_pSharedFilesList->WriteToOfferedFilePacket(*static_cast<CKnownFile*>(list.GetHead()), packetStream, NULL);
						list.RemoveHead();
					}

					Packet		*pReplyPacket = new Packet(&packetStream);

					pReplyPacket->m_eOpcode = OP_ASKSHAREDFILESANSWER;
					g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
					SendPacket(pReplyPacket, true, true);
					break;
				}
				case OP_ASKSHAREDFILESANSWER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					m_pClient->ProcessSharedFileList(pcPacketBuf, dwPacketSize);
					break;
				}
				case OP_ASKSHAREDDIRS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					ASSERT(dwPacketSize == 0);

				//	IP banned, no answer for this request
					if (m_pClient->IsBanned())
					{
						break;
					}

					if ( g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_EVERYBODY
					  || (g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_FRIENDS && m_pClient->IsFriend()) )
					{
						AddDebugLogLine(false, IDS_SHAREDREQ1, m_pClient->GetUserName(), m_pClient->GetUserID(), GetResString(IDS_ACCEPTED));

						CString		strDir, strTest;

					//	Virtual folder list
						CArray<CString, CString&>	astrFolders;
						CMapStringToString			&mapVDirs = g_eMuleApp.m_pSharedFilesList->GetSharedVDirForList();
						POSITION					pos = mapVDirs.GetStartPosition();

						while (pos)
						{
							CString		strKey, strDir;

							mapVDirs.GetNextAssoc(pos, strKey, strDir);

							int			i = 0;

							while ((i < astrFolders.GetCount()) && (strKey.CompareNoCase(astrFolders.GetAt(i)) > 0))
								i++;
							astrFolders.InsertAt(i, strKey);
						}
					//
					//	Build the reply packet
						CSafeMemFile		packetStream(80);
						uint32				dwNumDirs = astrFolders.GetCount();

						packetStream.Write(&dwNumDirs, 4);
					//	For each directory in the array...
						for (int ix = 0; ix < astrFolders.GetCount(); ix++)
						{
							const CString		&strDir = astrFolders.GetAt(ix);
							uint16				uCnt = strDir.GetLength();

							packetStream.Write(&uCnt, 2);
							packetStream.Write(static_cast<LPCTSTR>(strDir), uCnt);
						}

						Packet		*pReplyPacket = new Packet(&packetStream);

						pReplyPacket->m_eOpcode = OP_ASKSHAREDDIRSANS;
						g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
						SendPacket(pReplyPacket, true, true);
					}
					else
					{
						AddDebugLogLine(false, IDS_SHAREDREQ1, m_pClient->GetUserName(), m_pClient->GetUserID(), GetResString(IDS_DENIED));

						Packet		*pReplyPacket = new Packet(OP_ASKSHAREDDENIEDANS);

						g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
						SendPacket(pReplyPacket, true, true);
					}
					break;
				}
				case OP_ASKSHAREDFILESDIR:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);

				//	IP banned, no answer for this request
					if (m_pClient->IsBanned())
					{
						break;
					}

					CSafeMemFile		packetStream(reinterpret_cast<uchar*>(pcPacketBuf), dwPacketSize);

					uint16		uDirNameLen;
					packetStream.Read(&uDirNameLen, 2);

					CString		strReqDir, strReqVDir;
					packetStream.Read(strReqVDir.GetBuffer(uDirNameLen), uDirNameLen);
					strReqVDir.ReleaseBuffer(uDirNameLen);
					PathRemoveBackslash(strReqVDir.GetBuffer());
					strReqVDir.ReleaseBuffer();

				//	Retrieve real dir from virtual dir
					CMapStringToString		&mapVDirs = g_eMuleApp.m_pSharedFilesList->GetSharedVDirForList();

					bool		bReqVDir = mapVDirs.Lookup(strReqVDir, strReqDir);

					if ( (bReqVDir)
					  && ( g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_EVERYBODY
					    || (g_eMuleApp.m_pGlobPrefs->CanSeeShares() == SEE_SHARE_FRIENDS && m_pClient->IsFriend()) ) )
					{
						AddDebugLogLine(false, IDS_SHAREDREQ2, m_pClient->GetUserName(), m_pClient->GetUserID(), strReqVDir, strReqDir, GetResString(IDS_ACCEPTED));
						ASSERT(packetStream.GetPosition() == packetStream.GetLength());

						CCKey		bufKey;
						CKnownFile	*pKnownFile;

						CTypedPtrList<CPtrList, CKnownFile*>	list;

						for (POSITION pos = g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetStartPosition(); pos != NULL;)
						{
							g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);

							CString		strSharedFileDir(pKnownFile->GetPath());

							PathRemoveBackslash(strSharedFileDir.GetBuffer());
							strSharedFileDir.ReleaseBuffer();
							if (strReqDir.CompareNoCase(strSharedFileDir) == 0)
								list.AddTail(pKnownFile);
						}

					//	Currently we are sending each shared directory, even if it does not contain any files.
					//	Because of this we also have to send an empty shared files list..
						/*if (list.GetCount())*/
						{
							CSafeMemFile		packetStream(80);

							uint16		uDirNameLen = strReqVDir.GetLength();
							packetStream.Write(&uDirNameLen, 2);

							packetStream.Write(static_cast<LPCTSTR>(strReqVDir), uDirNameLen);

							uint32		dwNumFiles = list.GetCount();
							packetStream.Write(&dwNumFiles, 4);

							while (list.GetCount())
							{
								g_eMuleApp.m_pSharedFilesList->WriteToOfferedFilePacket(*list.GetHead(), packetStream, NULL);
								list.RemoveHead();
							}

							Packet		*pReplyPacket = new Packet(&packetStream);

							pReplyPacket->m_eOpcode = OP_ASKSHAREDFILESDIRANS;
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
							SendPacket(pReplyPacket, true, true);
						}
					}
				//	If the requester isn't allowed to see our shares...
					else
					{
						AddDebugLogLine(false, IDS_SHAREDREQ2, m_pClient->GetUserName(), m_pClient->GetUserID(), strReqVDir, strReqDir, GetResString(IDS_DENIED));

						Packet		*pReplyPacket = new Packet(OP_ASKSHAREDDENIEDANS);

						g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
						SendPacket(pReplyPacket, true, true);
					}
					break;
				}
				case OP_ASKSHAREDDIRSANS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					if (m_pClient->GetFileListRequested() == 1)
					{
						CSafeMemFile		packetStream(reinterpret_cast<uchar*>(pcPacketBuf), dwPacketSize);

						uint32		dwNumDirs;
						packetStream.Read(&dwNumDirs, 4);

						for (UINT i = 0; i < dwNumDirs; i++)
						{
							uint16		uDirNameLen;
							packetStream.Read(&uDirNameLen, 2);

							CString		strDirName;
							packetStream.Read(strDirName.GetBuffer(uDirNameLen), uDirNameLen);
							strDirName.ReleaseBuffer(uDirNameLen);

						//	Better send the received and untouched directory string back to that client
						//	PathRemoveBackslash(strDirName.GetBuffer());
						//	strDirName.ReleaseBuffer();
							AddDebugLogLine(false, IDS_SHAREDANSW, m_pClient->GetUserName(), m_pClient->GetUserID(), strDirName);

							CSafeMemFile		packetStream2(80);

							uDirNameLen = strDirName.GetLength();
							packetStream2.Write(&uDirNameLen, 2);
							packetStream2.Write(static_cast<LPCTSTR>(strDirName), uDirNameLen);

							Packet		*pReplyPacket = new Packet(&packetStream2);

							pReplyPacket->m_eOpcode = OP_ASKSHAREDFILESDIR;
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pReplyPacket->m_dwSize);
							SendPacket(pReplyPacket, true, true);
						}

						ASSERT(packetStream.GetPosition() == packetStream.GetLength());
						m_pClient->SetFileListRequested(dwNumDirs);
					}
					else
					{
						AddDebugLogLine(false, IDS_SHAREDANSW2, m_pClient->GetUserName(), m_pClient->GetUserID());
					}

					break;
				}
				case OP_ASKSHAREDFILESDIRANS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);

					CSafeMemFile		packetStream(reinterpret_cast<uchar*>(pcPacketBuf), dwPacketSize);
				//
				//	Get the directory name length
					uint16		uDirNameLen;
					packetStream.Read(&uDirNameLen, 2);
				//
				//	Get the directory name
					CString		strDirName;
					packetStream.Read(strDirName.GetBuffer(uDirNameLen), uDirNameLen);
					strDirName.ReleaseBuffer(uDirNameLen);
					PathRemoveBackslash(strDirName.GetBuffer());
					strDirName.ReleaseBuffer();

					if (m_pClient->GetFileListRequested() > 0)
					{
						AddDebugLogLine(false, IDS_SHAREDINFO1, m_pClient->GetUserName(), m_pClient->GetUserID(), strDirName);
						m_pClient->ProcessSharedFileList(pcPacketBuf + packetStream.GetPosition(), dwPacketSize - packetStream.GetPosition(), strDirName);
						if (m_pClient->GetFileListRequested() == 0)
						{
							AddDebugLogLine(false, IDS_SHAREDINFO2, m_pClient->GetUserName(), m_pClient->GetUserID());
						}
					}
					else
					{
						AddDebugLogLine(false, IDS_SHAREDANSW3, m_pClient->GetUserName(), m_pClient->GetUserID(), strDirName);
					}
					break;
				}
				case OP_ASKSHAREDDENIEDANS:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					ASSERT(dwPacketSize == 0);
					AddLogLine(true, IDS_SHAREDREQDENIED, m_pClient->GetUserName(), m_pClient->GetUserID());
					m_pClient->SetFileListRequested(0);
					break;
				}
				case OP_UNKNOWNx6A:
				case OP_UNKNOWNx7A:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
					break;
				}
				default:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(dwPacketSize);
//					AddDebugLogLine(false, RGB_LOG_WARNING + _T("Unknown eOpcode: %i %x from %s"), eOpcode, eOpcode, GetClientNameAndVersionString(m_pClient));
//					PacketToDebugLogLine(RGB_LOG_WARNING + _T("Unknown eOpcode"), pcPacketBuf, dwPacketSize, eOpcode);
				}
			}
		}
		catch (CFileException * error)
		{
			OUTPUT_DEBUG_TRACE();
			error->Delete();	//mf
			throw GetResString(IDS_ERR_INVALIDPACKAGE);
		}
		catch (CMemoryException * error)
		{
			OUTPUT_DEBUG_TRACE();
			error->Delete();
			throw CString(_T("Memory exception"));
		}
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
		if (m_pClient)
		{
			m_pClient->SetDownloadState(DS_ERROR);
		//	TODO write this into a debugfile
			AddDebugLogLine(false, GetResString(IDS_ERR_CLIENTERROR) + _T(" (%s)"),
				m_pClient->GetUserName(), m_pClient->GetFullIP(), error.GetBuffer(), GetClientNameAndVersionString(m_pClient) );
//			PacketToDebugLogLine(RGB_LOG_WARNING + _T("Unknown client eOpcode"), pcPacketBuf, dwPacketSize, eOpcode);
		}
//		else
//		{
//			AddDebugLogLine(false, IDS_ERR_BADCLIENTACTION, error.GetBuffer());
//			PacketToDebugLogLine(RGB_LOG_WARNING + _T("Unknown protocol eOpcode"), pcPacketBuf, dwPacketSize, eOpcode);
//		}

		Disconnect();
		return false;
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientReqSocket::ProcessExtPacket(char* packet, uint32 size, EnumOpcodes opcode)
{
	try
	{
		try
		{
			if (!m_pClient)
				throw GetResString(IDS_ERR_UNKNOWNCLIENTACTION);

			EnumDLQState enumClientDLState = m_pClient->GetDownloadState();

			switch (opcode)
			{
				case OP_EMULEINFO:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(size);
					m_pClient->ProcessMuleInfoPacket(reinterpret_cast<BYTE*>(packet), size);
					m_pClient->SendMuleInfoPacket(true);
					break;
				}
				case OP_EMULEINFOANSWER:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(size);
					m_pClient->ProcessMuleInfoPacket(reinterpret_cast<BYTE*>(packet), size);
					break;
				}
				case OP_SECIDENTSTATE:
				{
					m_pClient->ProcessSecIdentStatePacket(reinterpret_cast<BYTE*>(packet), size);
					if (m_pClient->GetSecureIdentState() == IS_SIGNATURENEEDED)
					{
						m_pClient->SendSignaturePacket();
					}
					else if (m_pClient->GetSecureIdentState() == IS_KEYANDSIGNEEDED)
					{
						m_pClient->SendPublicKeyPacket();
						m_pClient->SendSignaturePacket();
					}
					break;
				}
				case OP_PUBLICKEY:
				{
					if (m_pClient->IsBanned())
					{
						break;
					}
					m_pClient->ProcessPublicKeyPacket((uchar*)packet, size);
					break;
				}
				case OP_SIGNATURE:
				{
					if (m_pClient->IsBanned())
					{
						break;
					}
					m_pClient->ProcessSignaturePacket((uchar*)packet, size);
#ifdef DEBUG_SHOW_SECUREID

					DEBUG_ONLY(AddDebugLogLine(false, "-->Signature Received"));
#endif

					break;
				}

				case OP_COMPRESSEDPART:
				{
					m_pClient->UpdateLastBlockReceivedTime();
					// eklmn: before we process a packet we need to check Download State
					// if client send us a packet in unwanted state (we don't request (need) this packet)
					// we gonna switch him to error state.
					if (enumClientDLState == DS_DOWNLOADING || enumClientDLState == DS_TOOMANYCONNS)
					{
						if (m_pClient->m_pReqPartFile && (m_pClient->m_pReqPartFile->GetStatus() == PS_READY || m_pClient->m_pReqPartFile->GetStatus() == PS_EMPTY))
						{
							m_pClient->ProcessBlockPacket(packet, size, true);
							if ( (m_pClient->m_pReqPartFile != NULL) &&
							     ( m_pClient->m_pReqPartFile->GetStatus() == PS_PAUSED
							       || m_pClient->m_pReqPartFile->GetStatus() == PS_STOPPED
							       || m_pClient->m_pReqPartFile->GetStatus() == PS_ERROR ) )
							{
								Packet * packet = new Packet(OP_CANCELTRANSFER);
								g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
								m_pClient->m_pRequestSocket->SendPacket(packet, true, true);
							}
						}
						else
						{
							Packet* packet = new Packet(OP_CANCELTRANSFER);
							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
							m_pClient->m_pRequestSocket->SendPacket(packet, true, true);
							m_pClient->SetDownloadState(m_pClient->m_pReqPartFile == NULL ? DS_NONE : DS_ONQUEUE);
						}
					}
					else
						m_pClient->SetDownloadState(DS_ERROR);
					break;
				}
				case OP_QUEUERANKING:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(size);
					if (size != 12)
						throw GetResString(IDS_ERR_BADSIZE);
					uint16 newrank;
					memcpy2(&newrank, packet, 2);
					m_pClient->SetRemoteQueueFull(false);
					m_pClient->SetRemoteQueueRank(newrank);
				//	eklmn: switch DL state in case newrank != 0
					if (newrank && enumClientDLState == DS_DOWNLOADING)
						m_pClient->SetDownloadState(DS_ONQUEUE);
					break;
				}
				case OP_REQUESTSOURCES:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadSourceExchange(size);
					if (m_pClient->GetSourceExchangeVersion() >= 1)
					{
						if (size != 16)
							throw GetResString(IDS_ERR_BADSIZE);

					//first check shared file list, then download list
						CKnownFile	   *file = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)packet);

						if (!file)
							file = g_eMuleApp.m_pDownloadQueue->GetFileByID((uchar*)packet);

						if (file)
						{
							DWORD		dwTimePassed = ::GetTickCount() - m_pClient->GetLastSrcReqTime() + CONNECTION_LATENCY;
							bool		bNeverAskedBefore = m_pClient->GetLastSrcReqTime() == 0;

							if (
							//	if not complete and file is rare, allow once every 10 minutes
							    ( file->IsPartFile()
							      && ((CPartFile*)file)->GetSourceCount() <= RARE_FILE * 2
							      && (bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASK) )
							//	OR if file is not rare or if file is complete, allow every 90 minutes
							    || (bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASK * MINCOMMONPENALTY) )
							{

								m_pClient->SetLastSrcReqTime();

								Packet	   *tosend = file->CreateSrcInfoPacket(m_pClient);

								if (tosend)
								{
									g_eMuleApp.m_pUploadQueue->AddUpDataOverheadSourceExchange(tosend->m_dwSize);
									SendPacket(tosend, true, true);
								//	AddDebugLogLine(false, "RCV:Source Request User(%s) File(%s)", m_pClient->GetUserName(), file->GetFileName());
								}
							}
						}
					}
					break;
				}
				case OP_ANSWERSOURCES:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadSourceExchange(size);

					try
					{
						CSafeMemFile		packetStream(reinterpret_cast<BYTE*>(packet), size);

						uchar				hash[16];

						packetStream.Read(hash, 16);

						CKnownFile			*pKnownFile = g_eMuleApp.m_pDownloadQueue->GetFileByID(reinterpret_cast<BYTE*>(packet));

						if (pKnownFile != NULL)
						{
							if (pKnownFile->IsPartFile())
							{
							//set the client's answer time
								m_pClient->SetLastSrcAnswerTime();
							//and set the pKnownFile's last answer time
								((CPartFile*)pKnownFile)->SetLastAnsweredTime();

							//v1:- eklmn: support Source Exchange v2
								((CPartFile*)pKnownFile)->AddClientSources(&packetStream, m_pClient->GetSourceExchangeVersion());
							}
						}
					}
					catch (CFileException * error)
					{
						OUTPUT_DEBUG_TRACE();
						error->Delete();
						throw GetResString(IDS_ERR_INVALIDPACKAGE);
					}
					break;
				}
				case OP_FILEDESC:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(size);
					m_pClient->ProcessMuleCommentPacket(packet, size);
					break;
				}
				/*
				case OP_VERIFYUPSREQ:{
					CUpDownClient* vuclient = g_eMuleApp.m_pClientList->VUGetRandomClient();
					if (vuclient){
						Packet* answer = new Packet(OP_VERIFYUPSANSWER,6,OP_EMULEPROT);
						uint32 ip = vuclient->GetIP();
						memcpy2(answer->m_pcBuffer,ip,4);
						uint16 udpport = vuclient->GetUDPPort();
						memcpy2(answer->m_pcBuffer,ip,4);
						g_eMuleApp.m_pUploadQueue->AddUpDataRateMSOverhead(packet->m_dwSize);
						SendPacket(answer,true,true);
					}
				}
				case OP_VERIFYUPSANSWER:{
					if (size != 12)
						throw CString("invalid size (OP_VERIFYUPSANSWER)");
				}
				*/
				default:
				{
					g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(size);
				//	AddDebugLogLine(false, RGB_LOG_WARNING + _T("unknown extended opcode: %i %x from %s"), opcode, opcode, GetClientNameAndVersionString(m_pClient));
				}
			}
		}
		catch (CFileException * error)
		{
			OUTPUT_DEBUG_TRACE();
			error->Delete();
			throw GetResString(IDS_ERR_INVALIDPACKAGE);
		}
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
		if (!m_pClient)
			AddDebugLogLine(false, IDS_ERR_BADCLIENTACTION, error.GetBuffer());
		else
			AddDebugLogLine(false, GetResString(IDS_ERR_BADCLIENTACTION) + _T(" (%s)"), error.GetBuffer(), GetClientNameAndVersionString(m_pClient));
		if (m_pClient)
			m_pClient->SetDownloadState(DS_ERROR);
		Disconnect();
		return false;
	}
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::OnInit()
{
	byte tv = 1;
//SetSockOpt(SO_DONTLINGER,&tv,sizeof(BOOL));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::OnSend(int nErrorCode)
{
	ResetTimeOutTimer();
	CEMSocket::OnSend(nErrorCode);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::OnError(int nErrorCode)
{

	if (m_pClient)
		AddDebugLogLine(false, IDS_ERR_BADCLIENT2, m_pClient->GetUserName(), m_pClient->GetFullIP(), nErrorCode);
	else
		AddDebugLogLine(false, IDS_ERR_BADCLIENTACTION, "CClientReqSocket::OnError");
	Disconnect();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::PacketReceived(Packet* packet)
{
//v- eklmn: feature(01): Filter clients, which are violating protocol
#ifdef OLD_SOCKETS_ENABLED
	SOCKADDR_IN sockAddr;
	memzero(&sockAddr, sizeof(sockAddr));
	uint32 nSockAddrLen = sizeof(sockAddr);
	GetPeerName((SOCKADDR*) & sockAddr, (int*) & nSockAddrLen);
	uint32 UserIP = sockAddr.sin_addr.S_un.S_addr;
	CString FullUserIP(inet_ntoa(sockAddr.sin_addr));
	if (g_eMuleApp.m_pIPFilter->IsFiltered(UserIP))
	{
		/*
		if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
			AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Filtered: %s %s "), FullUserIP,g_eMuleApp.m_pIPFilter->GetLastHit());
		*/
		InterlockedIncrement(&g_eMuleApp.m_lIncomingFiltered);
		InterlockedIncrement(&g_eMuleApp.m_lTotalFiltered);	//g_eMuleApp.stat_filteredclients++;
		//eklmn: reset client to NULL in order don't let CUpDownClient::Disconnected be processed for a client which is not in the list of clients.
		m_pClient = NULL;
		Disconnect();
		return;
	}
#endif //OLD_SOCKETS_ENABLED
//^- eklmn: feature(01): Filter clients, which are violating protocol

	switch (packet->m_eProtocol)
	{
		case OP_EDONKEYPROT:
			ProcessPacket(packet->m_pcBuffer, packet->m_dwSize, packet->m_eOpcode);
			break;
		case OP_PACKEDPROT:
			if (!packet->UnpackPacket())
			{
				ASSERT (false);
				break;
			}
		case OP_EMULEPROT:
			ProcessExtPacket(packet->m_pcBuffer, packet->m_dwSize, packet->m_eOpcode);
			break;
		default:
			//eklmn: let's free a socket from buggy and unknown clients
//			AddDebugLogLine(false, _T("Received unknown client TCP packet; IP=%s  protocol=0x%02x  opcode=0x%02x  size=%u"), FullUserIP, packet->m_eProtocol, packet->m_eOpcode, packet->m_dwSize);
			if (m_pClient)
				m_pClient->SetDownloadState(DS_ERROR);
			Disconnect();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientReqSocket::OnReceive(int nErrorCode)
{
	ResetTimeOutTimer();
	CEMSocket::OnReceive(nErrorCode);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientReqSocket::Create()
{
	g_eMuleApp.m_pListenSocket->AddConnection();
	BOOL result = CAsyncSocketEx::Create(0, SOCK_STREAM, FD_WRITE | FD_READ | FD_CLOSE); // deadlake PROXYSUPPORT - changed to AsyncSocketEx
	OnInit();
	return result;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//	CListenSocket member functions

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CListenSocket constructor
CListenSocket::CListenSocket(CPreferences *in_pPrefs)
{
	m_pPrefs = in_pPrefs;
	m_iNumOpenSockets = 0;
	m_iMaxConnectionsReachedCount = 0;
	m_iOpenSocketsInterval = 0;
	InterlockedExchange(&m_lNumPendingConnections, 0);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CListenSocket::~CListenSocket()
{
	Close();
	KillAllSockets();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CListenSocket::StartListening()
{
	m_bListening = true;

	return(this->Create(m_pPrefs->GetPort(), SOCK_STREAM, FD_ACCEPT) && this->Listen());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::RestartListening()
{
	m_bListening = true;
	if (InterlockedDecrement(&m_lNumPendingConnections) >= 0)
	{
		OnAccept(0);
	}
	else
	{
		InterlockedExchange(&m_lNumPendingConnections, 0);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::StopListening()
{
	m_bListening = false;
	m_iMaxConnectionsReachedCount++;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	OnAccept() is called to notify the request socket that it can accept pending connection requests by calling
//	Accept().
void CListenSocket::OnAccept(int nErrorCode)
{
	if (nErrorCode == 0)
	{
		InterlockedIncrement(&m_lNumPendingConnections);
	//	If the number of pending connections was less than 0 (which obviously shouldn't happen)...
		if (InterlockedCompareExchange(&m_lNumPendingConnections, 1, 0) == 0)
		{
			//ASSERT(false);
			//m_lNumPendingConnections = 1;
		}
		if (TooManySockets(true) && !g_eMuleApp.m_pServerConnect->IsConnecting())
		{
			StopListening();
			return;
		}
		else if (m_bListening == false)
			RestartListening(); //If the client is still at maxconnections, this will allow it to go above it.. But if you don't, you will get a lowID on all servers.

		while (InterlockedDecrement(&m_lNumPendingConnections) >= 0)
		{
			//m_lNumPendingConnections--;
			CClientReqSocket* newclient = new CClientReqSocket(m_pPrefs);
			if (!Accept(*newclient))
				newclient->Safe_Delete();
			else
			{
				newclient->AsyncSelect(FD_WRITE | FD_READ | FD_CLOSE);
				newclient->OnInit();
			}
			AddConnection();
		}
	//	if (TooManySockets(true) && !g_eMuleApp.m_pServerConnect->IsConnecting())
	//	StopListening();
	}

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Process() takes care of cleaning up invalid, deleted, and timed out sockets from the open request socket list.
void CListenSocket::Process()
{
	POSITION	pos2;

	m_iOpenSocketsInterval = 0;
	m_iNumOpenSockets = 0;

//	For each open request socket...
	for (POSITION pos1 = m_openSocketList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_openSocketList.GetNext(pos1);

		CClientReqSocket	*pSocket = m_openSocketList.GetAt(pos2);

		m_iNumOpenSockets++;

	//	If the request socket is marked for deletion...
		if (pSocket->m_bDeleteThis)
		{
			if (pSocket->m_SocketData.hSocket != INVALID_SOCKET) // deadlake PROXYSUPPORT - changed to AsyncSocketEx
			{
				pSocket->Close();
			}
			else
			{
				pSocket->TimedDelete();
			}
		}
		else
		{
		//	If the request socket has timed out, disconnect it.
			m_openSocketList.GetAt(pos2)->CheckTimeOut();
		}
	}
//	If we're not currently listening for new connections and we're either just connecting to a server
//	or we're not within 5 connections of max then start listening again.
	if ((GetNumOpenSockets() + 5 < m_pPrefs->GetMaxConnections() || g_eMuleApp.m_pServerConnect->IsConnecting()) && !m_bListening)
		RestartListening();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::RecalculateStats()
{
	memzero(m_ConnectionStates, 6);
	POSITION pos1, pos2;
	for (pos1 = m_openSocketList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_openSocketList.GetNext(pos1);
		CClientReqSocket* cur_sock = m_openSocketList.GetAt(pos2);
		switch (cur_sock->GetConState())
		{
			case ES_DISCONNECTED:
			m_ConnectionStates[0]++;
			break;
			case ES_NOTCONNECTED:
			m_ConnectionStates[1]++;
			break;
			case ES_CONNECTED:
			m_ConnectionStates[2]++;
			break;
		}
	}
}

void CListenSocket::AddSocket(CClientReqSocket* toadd)
{
	m_openSocketList.AddTail(toadd);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::RemoveSocket(CClientReqSocket *pSocket)
{
	POSITION	pos;

	pos = m_openSocketList.Find(pSocket);
	if (pos != NULL)
		m_openSocketList.RemoveAt(pos);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	DeleteSocket() schedules the socket 'pSocket' for deletion if it's in the open socket list.
bool CListenSocket::DeleteSocket(CClientReqSocket *pSocket)
{
	bool		bDeleted = false;

	if (pSocket != NULL && IsValidSocket(pSocket))
	{
		pSocket->Safe_Delete();
		bDeleted = true;
	}

	return bDeleted;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void CListenSocket::KillAllSockets()
{
	for (POSITION pos = m_openSocketList.GetHeadPosition(); pos != NULL; pos = m_openSocketList.GetHeadPosition())
	{
		CClientReqSocket * cur_socket = m_openSocketList.GetAt(pos);
		if (cur_socket->m_pClient)
			delete cur_socket->m_pClient;
		else
			delete cur_socket;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::AddConnection()
{
	m_iOpenSocketsInterval++;
	m_iNumOpenSockets++;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	TooManySockets() returns true if the number of open sockets is greater than the number specified
//	in Preferences. If 'bIgnoreInterval' is false, it will also return true if the number of
//	new connections since the last Process() call is greater than the "connections per 5 secs"
//	setting in Preferences.
bool CListenSocket::TooManySockets(bool bIgnoreInterval)
{
	if ( GetNumOpenSockets() > m_pPrefs->GetMaxConnections()
	     || ( m_iOpenSocketsInterval > (g_eMuleApp.m_pGlobPrefs->GetMaxConPerFive() * g_eMuleApp.m_pdlgEmule->m_dlgStatistics.GetMaxConPerFiveModifier())
	          && !bIgnoreInterval ) )
	{
		return true;
	}
	else
		return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsValidSocket() returns true if 'pSocket' is in the open request socket list.
bool CListenSocket::IsValidSocket(CClientReqSocket *pSocket)
{
	return m_openSocketList.Find(pSocket);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CListenSocket::Debug_ClientDeleted(CUpDownClient* deleted)
{
	POSITION pos1, pos2;
	for (pos1 = m_openSocketList.GetHeadPosition();(pos2 = pos1) != NULL;)
	{
		m_openSocketList.GetNext(pos1);
		CClientReqSocket* cur_sock = m_openSocketList.GetAt(pos2);
		if (!AfxIsValidAddress(cur_sock, sizeof(CClientReqSocket)))
		{
			AfxDebugBreak();
		}
		if (cur_sock->m_pClient == deleted)
		{
			AfxDebugBreak();
		}
	}
}
#endif //OLD_SOCKETS_ENABLED
