//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#pragma once
#include "opcodes.h"
#include "emule.h"
#include "types.h"
#include "preferences.h"
#include "KnownFile.h"
#include "knownfilelist.h"
#include "SharedFilesCtrl.h"
#include "sockets.h"
#include "partfile.h"
#include "mapkey.h"
#include "lancast.h"	// LANCAST (moosetea) - header include
#include "afxmt.h"
#include "Loggable.h"

struct UnknownFile_Struct
{
	CString		m_strFileName;
	CString		m_strDirectory;
};

class CKnownFileList;

class CSharedFileList : public CLoggable
{
	friend class CSharedFilesCtrl;
	friend class CClientReqSocket;
	friend class CWebServer;

public:
	CMutex				m_mutexList;
	CKnownFileList	   *m_pKnownFileList;

protected:
	CMutex				m_HashMapMutex;

private:
	CPreferences	   *m_pPrefs;
	CServerConnect	   *m_pServerConnect;
	CSharedFilesCtrl   *m_pOutput;
	CLanCast		   *m_pLanCast;		// LANCAST (moosetea)
	bool				m_bDelayReload;
	CMapStringToString	m_SharedVDirForList;

public:
					CSharedFileList(CPreferences* in_prefs,CServerConnect* in_server, CKnownFileList* in_filelist);
				   ~CSharedFileList();
	void			SendListToServer();
	void			Reload(bool sendtoserver = true);
	void			SafeAddKnownFile(CKnownFile* toadd, bool bOnlyAdd = false);
	void			SetOutputCtrl(CSharedFilesCtrl* in_ctrl);
	void			RemoveFile(CKnownFile* toremove);
	CKnownFile	   *GetFileByID(const uchar* filehash);
	short			GetFilePriorityByID(uchar* filehash);
	void			WriteToOfferedFilePacket(CKnownFile& cur_file, CMemFile& files, CServer* pServer);
	uint64			GetDatasize(uint64 &qwLargest);
	uint16			GetCount()	{return m_mapSharedFiles.GetCount(); }
	void			UpdateItem(CKnownFile* pKnownFile,bool bResort=true);
	void			NextLANBroadcast();
	bool			FindUnknownFile(CString strDirectory, CString strFileName);

	CMapStringToString	   &GetSharedVDirForList()		{ return m_SharedVDirForList; }		// #zegzav:vdir4list

private:
	void			FindSharedFiles();
	void			AddFilesFromDirectory(const CString& strDirectory);
	void			HashNextFile();
//	UI calls
	void			UpdateOfferedItems();

//	virtual dir for Hybrid client to client shared files listing
	void			BuildSharedVDirForList();
	bool			AddSharedVDirForList(bool bIncomplete, const CString &strDir);

#ifdef NEW_SOCKETS
public:
#endif //NEW_SOCKETS
	CMap<CCKey,CCKey&,CKnownFile*,CKnownFile*>		m_mapSharedFiles;
	CTypedPtrList<CPtrList, UnknownFile_Struct*>	m_waitingForHashList;

};

class CHashFileThread : public CWinThread
{
	DECLARE_DYNCREATE(CHashFileThread)

private:
	CSharedFileList	   *m_pOwner;
	CString				m_strDirectory;
	CString				m_strFileName;
	CPartFile		   *m_pPartFile;

public:
	virtual	BOOL		InitInstance() { return true; }
	virtual int			Run();
	void				SetValues( CSharedFileList *pOwner, LPCTSTR strDirectory, LPCTSTR strFileName,
								   CPartFile* in_pPartFile = NULL );

protected:
						CHashFileThread();
};
