// Ini2.cpp: Ini-File-Interface by bond006 <rene.landgrebe@gmx.de>
// ****************************************************************
// This new interface replaces the old interface which was too slow
// Last Update: 08/24/2003

#include "stdafx.h"
#include "ini2.h"
#include "otherfunctions.h"

CIni::CIni(const CString& strFileName)
{
	m_strFileName=strFileName;
	m_strDefCategory=_T("");

	OpenFile();
	ReadData();
}

CIni::~CIni()
{
	CloseFile();
}

void CIni::OpenFile()
{
	if(m_hFile==CFile::hFileNull)
		BOOL bRetCode = Open((LPCTSTR)m_strFileName,CFile::modeCreate | CFile::modeNoTruncate | CFile::modeReadWrite | CFile::shareDenyWrite);
}

void CIni::CloseFile()
{
	if(m_hFile!=CFile::hFileNull)
		Close();
}

void CIni::ReadData()
{
	if(m_hFile!=CFile::hFileNull)
	{
		char tempBuffer[131072]; // change this value to support larger files than 128KB
		memzero(tempBuffer, sizeof(tempBuffer));
		Read(tempBuffer,sizeof(tempBuffer));
		m_strFileBuffer = tempBuffer;

		// Cut off blank lines at the end of the file
		m_strFileBuffer.TrimRight(_T("\r\n"));
	}
}

void CIni::SaveAndClose()
{
	if(m_hFile!=CFile::hFileNull)
	{
		SetLength(0);
		m_strFileBuffer.TrimRight(_T("\r\n"));
		Write(m_strFileBuffer,m_strFileBuffer.GetLength());
	}

	CloseFile();
}

void CIni::AddCategory(const CString& strCategoryName)
{
	if(!CategoryExist(strCategoryName))
	{
		if(m_strFileBuffer.GetLength()!=0)
			m_strFileBuffer += _T("\r\n");

		m_strFileBuffer += _T("[");
		m_strFileBuffer += strCategoryName;
		m_strFileBuffer += _T("]");
	}
}

void CIni::SetDefaultCategory(const CString& strCategoryName)
{

	if(!CategoryExist(strCategoryName))
	{
		AddCategory(strCategoryName);
		m_strDefCategory = strCategoryName;
	}
	else
		m_strDefCategory = strCategoryName;
}

void CIni::DeleteCategory(const CString& strCategoryName)
{
	if(CategoryExist(strCategoryName))
	{
		CString catSearchString = _T("[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		int nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

		// Check if the category is valid
		if(nCatFirstCharPos!=0)
		{
			catSearchString = _T("\r\n[");
			catSearchString += strCategoryName;
			catSearchString += _T("]");

			nCatFirstCharPos -= 2;
			nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);
			nCatFirstCharPos += 2;
		}

		int nCatLastCharPos = m_strFileBuffer.Find(_T("\r\n["),(nCatFirstCharPos + strCategoryName.GetLength() + 1));

		if(nCatLastCharPos == (-1))
			nCatLastCharPos = (m_strFileBuffer.GetLength());

		m_strFileBuffer.Delete(nCatFirstCharPos,(nCatLastCharPos - nCatFirstCharPos));
	}
}

void CIni::DeleteEntry(const CString& strCategoryName, const CString &strEntryName)
{
	if(EntryExist(strCategoryName,strEntryName))
	{
		CString catSearchString = _T("[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		int nEntryFirstCharPos = 0;
		int nEntryLastCharPos = 0;
		int nCatFirstCharPos = 0;
		nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

		// Check if the category is valid
		if(nCatFirstCharPos!=0)
		{
			catSearchString = _T("\r\n[");
			catSearchString += strCategoryName;
			catSearchString += _T("]");

			nCatFirstCharPos -= 2;
			nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);
			nCatFirstCharPos += 2;
		}

		CString entrySearchString = _T("\r\n");
		entrySearchString += strEntryName;
		entrySearchString += _T("=");

		nEntryFirstCharPos = m_strFileBuffer.Find(entrySearchString,(nCatFirstCharPos + strCategoryName.GetLength() + 1));
		nEntryFirstCharPos += 2;

		nEntryLastCharPos = m_strFileBuffer.Find(_T("\r\n"),nEntryFirstCharPos);

		if(nEntryLastCharPos==(-1))
			nEntryLastCharPos = m_strFileBuffer.GetLength();
		else
			nEntryLastCharPos += 2; // remove entry including "\r\n"

		m_strFileBuffer.Delete(nEntryFirstCharPos,(nEntryLastCharPos - nEntryFirstCharPos));
	}
}

CString CIni::GetString(const CString& strEntryName, const CString& defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return temp;
}

double CIni::GetDouble(const CString& strEntryName, double defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return _tstof(temp);
}

float CIni::GetFloat(const CString& strEntryName, float defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return (float)_tstof(temp);
}

int CIni::GetInt(const CString& strEntryName, int defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return _tstoi(temp);
}

uint32 CIni::GetUInt32(const CString& strEntryName, uint32 defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return _tstoi(temp);
}

uint64 CIni::GetUInt64(const CString& strEntryName, uint64 defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return _tstoi64(temp);
}

WORD CIni::GetWORD(const CString& strEntryName, WORD defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return (WORD)_tstoi(temp);
}

bool CIni::GetBool(const CString& strEntryName, bool defValue, const CString& strCategoryName)
{
	CString finalCatName;
	CString temp;

	if(strCategoryName!=_T(""))
	{
		finalCatName = strCategoryName;
		temp = GetValue(finalCatName,strEntryName);
	}
	else
	{
		finalCatName = m_strDefCategory;
		temp = GetValue(finalCatName,strEntryName);
	}

	if(temp==_T("") && EntryExist(finalCatName,strEntryName)==false)
		return defValue;
	else
		return (_tstoi(temp)!=0);
}

void CIni::SetString(const CString& strEntryName, const CString& strValue, const CString& strCategoryName)
{
	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::SetDouble(const CString& strEntryName, double dValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%g"),dValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::SetFloat(const CString& strEntryName, float fValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%g"),fValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::SetInt(const CString& strEntryName, int nValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%d"),nValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}


void CIni::SetUInt32(const CString& strEntryName, uint32 nValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%I32u"),nValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::SetUInt64(const CString& strEntryName, uint64 nValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%I64u"),nValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}


void CIni::SetWORD(const CString& strEntryName, WORD wValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%u"),wValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::SetBool(const CString& strEntryName, bool bValue, const CString& strCategoryName)
{
	CString strValue;
	strValue.Format(_T("%d"),bValue);

	if(strCategoryName!=_T(""))
		SetValue(strCategoryName,strEntryName,strValue);
	else
		SetValue(m_strDefCategory,strEntryName,strValue);
}

void CIni::GetArray(CString* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int nPos = 0;
	int nLastPos = 0;
	int nEscapePos = 0;

	// bugfix: using escape characters for "\" and "," in strings - 08/25/03
	if(temp.Find(_T('\\')) != (-1))
	{
		for(int i=0;i<nElements;i++)
		{
			nLastPos = nPos;
			nPos = temp.Find(_T(','), nLastPos);
			nEscapePos = temp.Find(_T('\\'), nLastPos);

			if(nPos==(-2))
				pToArray[i] = _T("");

			if(nEscapePos==(-1) && nPos!=(-2))
			{
				if(nPos == (-1))
				{
					pToArray[i] = temp.Mid(nLastPos,(temp.GetLength() - nLastPos));
					nPos = -2;
				}
				else
				{
					pToArray[i] = (nPos >= 0) ? temp.Mid(nLastPos,(nPos - nLastPos)) : _T("");
					nPos++;
				}
			}
			else if(nEscapePos!=(-1) && nPos!=(-2))
			{
				if(nPos < nEscapePos)
				{
					if(nPos == (-1))
					{
						pToArray[i] = temp.Mid(nLastPos,(temp.GetLength() - nLastPos));
						nPos = -2;
					}
					else
					{
						pToArray[i] = (nPos >= 0) ? temp.Mid(nLastPos,(nPos - nLastPos)) : _T("");
						nPos++;
					}
				}
				else
				{
					int saveBegPos = nLastPos;

					while(nPos > nEscapePos && nEscapePos != (-1))
					{
						nLastPos = nEscapePos + 2;
						nPos = temp.Find(_T(','), nLastPos);
						nEscapePos = temp.Find(_T('\\'), nLastPos);
					}

					if(nPos == (-1))
					{
						pToArray[i] = temp.Mid(saveBegPos,(temp.GetLength() - saveBegPos));
						nPos = -2;
					}
					else
					{
						pToArray[i] = (nPos >= 0) ? temp.Mid(saveBegPos,(nPos - saveBegPos)) : _T("");
						nPos++;
					}
				}
			}

			pToArray[i].Replace(_T("\\,"),_T(","));
			pToArray[i].Replace(_T("\\\\"),_T("\\"));
		}
	}
	else
	{
		for(int i=0;i<nElements;i++)
			pToArray[i] = (nPos >= 0) ? temp.Tokenize(_T(","),nPos) : _T("");
	}
}

void CIni::GetArray(double* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int nPos = 0;

	for(int i=0;i<nElements;i++)
		pToArray[i] = (nPos >= 0) ? _tstof(temp.Tokenize(_T(","),nPos)) : 0.;
}

void CIni::GetArray(float* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int nPos = 0;

	for(int i=0;i<nElements;i++)
		pToArray[i] = (nPos >= 0) ? (float)_tstof(temp.Tokenize(_T(","),nPos)) : 0.f;
}

void CIni::GetArray(int* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int nPos = 0;

	for(int i=0;i<nElements;i++)
		pToArray[i] = (nPos >= 0) ? _tstoi(temp.Tokenize(_T(","),nPos)) : 0;
}

void CIni::GetArray(WORD* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int nPos = 0;

	for(int i=0;i<nElements;i++)
		pToArray[i] = (nPos >= 0) ? (WORD)_tstoi(temp.Tokenize(_T(","),nPos)) : 0;
}

void CIni::GetArray(bool* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp;
	
	if(strCategoryName==_T(""))
		temp = GetString(strEntryName,_T(""),m_strDefCategory);
	else
		temp = GetString(strEntryName,_T(""),strCategoryName);

	int token = 0;
	int nPos = 0;

	for(int i=0;i<nElements;i++)
	{
		token = (nPos >= 0) ? _tstoi(temp.Tokenize(_T(","),nPos)) : false;

		if(token!=0)
			pToArray[i] = true;
		else
			pToArray[i] = false;
	}
}

void CIni::SetArray(CString* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");

	for(int i=0;i<nElements;i++)
	{
		// bugfix: using escape characters for "\" and "," in strings - 08/25/03
		pToArray[i].Replace(_T("\\"),_T("\\\\"));
		pToArray[i].Replace(_T(","),_T("\\,"));

		temp += pToArray[i];
		temp += _T(",");
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetArray(double* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");
	CString strValue = _T("");

	for(int i=0;i<nElements;i++)
	{
		strValue.Format(_T("%g"),pToArray[i]);
		temp += strValue;
		temp += _T(',');
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetArray(float* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");
	CString strValue = _T("");

	for(int i=0;i<nElements;i++)
	{
		strValue.Format(_T("%g"),pToArray[i]);
		temp += strValue;
		temp += _T(',');
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetArray(int* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");
	CString strValue = _T("");

	for(int i=0;i<nElements;i++)
	{
		strValue.Format(_T("%d"),pToArray[i]);
		temp += strValue;
		temp += _T(',');
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetArray(WORD* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");
	CString strValue = _T("");

	for(int i=0;i<nElements;i++)
	{
		strValue.Format(_T("%u"),pToArray[i]);
		temp += strValue;
		temp += _T(',');
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetArray(bool* pToArray, int nElements, const CString& strEntryName, const CString& strCategoryName)
{
	CString temp = _T("");
	CString strValue = _T("");

	for(int i=0;i<nElements;i++)
	{
		strValue.Format(_T("%d"),pToArray[i]);
		temp += strValue;
		temp += _T(',');
	}

	temp.TrimRight(_T(','));

	if(strCategoryName==_T(""))
		SetString(strEntryName,temp,m_strDefCategory);
	else
		SetString(strEntryName,temp,strCategoryName);
}

void CIni::SetValue(const CString& strCategoryName, const CString& strEntryName, const CString& strValue)
{
	if(EntryExist(strCategoryName,strEntryName))
	{
		CString catSearchString = _T("[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		int nValueFirstCharPos = 0;
		int nEntryFirstCharPos = 0;
		int nEntryLastCharPos = 0;
		int nCatFirstCharPos = 0;

		nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

		// Check if the category is valid
		if(nCatFirstCharPos!=0)
		{
			catSearchString = _T("\r\n[");
			catSearchString += strCategoryName;
			catSearchString += _T("]");

			nCatFirstCharPos -= 2;
			nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);
			nCatFirstCharPos += 2;
		}

		CString entrySearchString = _T("\r\n");
		entrySearchString += strEntryName;
		entrySearchString += _T("=");

		nEntryFirstCharPos = m_strFileBuffer.Find(entrySearchString,(nCatFirstCharPos + strCategoryName.GetLength() + 1));
		nEntryFirstCharPos += 2;
		nEntryLastCharPos = m_strFileBuffer.Find(_T("\r\n"),nEntryFirstCharPos);

		if(nEntryLastCharPos==(-1))
			nEntryLastCharPos = m_strFileBuffer.GetLength();

		nValueFirstCharPos = nEntryFirstCharPos + strEntryName.GetLength() + 1;

		// Delete an existing value and insert a new value
		m_strFileBuffer.Delete(nValueFirstCharPos,(nEntryLastCharPos - nValueFirstCharPos));
		m_strFileBuffer.Insert(nValueFirstCharPos,strValue);
	}
	else
	{
		if(CategoryExist(strCategoryName))
		{
			CString catSearchString = _T("[");
			catSearchString += strCategoryName;
			catSearchString += _T("]");

			int nCatFirstCharPos = 0;
			int nCatLastCharPos = 0;

			nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

			// Check if the category is valid
			if(nCatFirstCharPos!=0)
			{
				catSearchString = _T("\r\n[");
				catSearchString += strCategoryName;
				catSearchString += _T("]");

				nCatFirstCharPos -= 2;
				nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);
				nCatFirstCharPos += 2;
			}

			nCatLastCharPos = m_strFileBuffer.Find(_T("\r\n["),(nCatFirstCharPos + strCategoryName.GetLength() + 1));

			CString entryString = _T("");

			if(nCatLastCharPos==(-1))
			{
				nCatLastCharPos = m_strFileBuffer.GetLength();
				entryString = _T("\r\n");
				entryString += strEntryName;
				entryString += _T("=");
				entryString += strValue;
			}
			else
			{
				nCatLastCharPos += 2;
				entryString = strEntryName;
				entryString += _T("=");
				entryString += strValue;
				entryString += _T("\r\n");
			}

			m_strFileBuffer.Insert(nCatLastCharPos,entryString);
		}
	}
}

CString CIni::GetValue(const CString& strCategoryName, const CString& strEntryName)
{
	if(ValueExist(strCategoryName,strEntryName))
	{
		CString catSearchString = _T("[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		int nValueFirstCharPos = 0;
		int nEntryFirstCharPos = 0;
		int nEntryLastCharPos = 0;
		int nCatFirstCharPos = 0;

		nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

		// Check if the category is valid
		if(nCatFirstCharPos!=0)
		{
			catSearchString = _T("\r\n[");
			catSearchString += strCategoryName;
			catSearchString += _T("]");

			nCatFirstCharPos -= 2;
			nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);
			nCatFirstCharPos += 2;
		}

		CString entrySearchString = _T("\r\n");
		entrySearchString += strEntryName;
		entrySearchString += _T("=");

		nEntryFirstCharPos = m_strFileBuffer.Find(entrySearchString,(nCatFirstCharPos + strCategoryName.GetLength() + 1));
		nEntryFirstCharPos += 2;
		nEntryLastCharPos = m_strFileBuffer.Find(_T("\r\n"),nEntryFirstCharPos);

		if(nEntryLastCharPos==(-1))
			nEntryLastCharPos = m_strFileBuffer.GetLength();

		nValueFirstCharPos = nEntryFirstCharPos + strEntryName.GetLength() + 1;

		return m_strFileBuffer.Mid(nValueFirstCharPos,(nEntryLastCharPos - nValueFirstCharPos));
	}
	else
		return _T("");
}

bool CIni::CategoryExist(const CString& strCategoryName)
{
	CString catSearchString = _T("[");
	catSearchString += strCategoryName;
	catSearchString += _T("]");

	int nFindPos = m_strFileBuffer.Find(catSearchString);

	if(nFindPos==(-1))
		return false;

	// Check if the category is valid
	if(nFindPos!=0)
	{
		catSearchString = _T("\r\n[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		nFindPos -= 2;
		nFindPos = m_strFileBuffer.Find(catSearchString,nFindPos);

		if(nFindPos==(-1))
			return false;
	}

	return true;
}

bool CIni::EntryExist(const CString& strCategoryName, const CString& strEntryName)
{
	CString catSearchString = _T("[");
	catSearchString += strCategoryName;
	catSearchString += _T("]");

	int nEntryFirstCharPos = 0;
	int nCatFirstCharPos = 0;
	int nCatLastCharPos = 0;
	nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

	if(nCatFirstCharPos==(-1))
		// Category does not exist
		return false;

	// Check if the category is valid
	if(nCatFirstCharPos!=0)
	{
		catSearchString = _T("\r\n[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		nCatFirstCharPos -= 2;
		nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);

		if(nCatFirstCharPos==(-1))
			return false;

		nCatFirstCharPos += 2;
	}

	nCatLastCharPos = m_strFileBuffer.Find(_T("\r\n["),(nCatFirstCharPos + strCategoryName.GetLength() + 1));

	if(nCatLastCharPos==(-1))
		// No following category exists - end of category is end of file
		nCatLastCharPos = m_strFileBuffer.GetLength();
	else
		nCatLastCharPos += 2;

	CString entrySearchString = _T("\r\n");
	entrySearchString += strEntryName;
	entrySearchString += _T("=");

	nEntryFirstCharPos = m_strFileBuffer.Find(entrySearchString,(nCatFirstCharPos + strCategoryName.GetLength() + 1));
	
	if(nEntryFirstCharPos!=(-1))
		nEntryFirstCharPos += 2;
	else
		return false;

	if(nEntryFirstCharPos<nCatFirstCharPos || nEntryFirstCharPos>nCatLastCharPos)
		// The entry we're searching for does not exist in the given category
		return false;

	return true;
}

bool CIni::ValueExist(const CString& strCategoryName, const CString& strEntryName)
{
	CString catSearchString = _T("[");
	catSearchString += strCategoryName;
	catSearchString += _T("]");

	int nEntryFirstCharPos = 0;
	int nCatFirstCharPos = 0;
	int nCatLastCharPos = 0;
	nCatFirstCharPos = m_strFileBuffer.Find(catSearchString);

	if(nCatFirstCharPos==(-1))
		// Category does not exist
		return false;

	// Check if the category is valid
	if(nCatFirstCharPos!=0)
	{
		catSearchString = _T("\r\n[");
		catSearchString += strCategoryName;
		catSearchString += _T("]");

		nCatFirstCharPos -= 2;
		nCatFirstCharPos = m_strFileBuffer.Find(catSearchString,nCatFirstCharPos);

		if(nCatFirstCharPos==(-1))
			return false;

		nCatFirstCharPos += 2;
	}

	nCatLastCharPos = m_strFileBuffer.Find(_T("\r\n["),(nCatFirstCharPos + strCategoryName.GetLength() + 1));

	if(nCatLastCharPos==(-1))
		// No following category exists - end of category is end of file
		nCatLastCharPos = m_strFileBuffer.GetLength();
	else
		nCatLastCharPos += 2;

	CString entrySearchString = _T("\r\n");
	entrySearchString += strEntryName;
	entrySearchString += _T("=");

	nEntryFirstCharPos = m_strFileBuffer.Find(entrySearchString,(nCatFirstCharPos + strCategoryName.GetLength() + 1));

	if(nEntryFirstCharPos!=(-1))
		nEntryFirstCharPos += 2;
	else
		return false;

	if(nEntryFirstCharPos<nCatFirstCharPos || nEntryFirstCharPos>nCatLastCharPos)
		// The entry we're searching for does not exist in the given category
		return false;

	int nEndOfLine = m_strFileBuffer.Find(_T("\r\n"),nEntryFirstCharPos);

	if(nEndOfLine==(-1))
		nEndOfLine = m_strFileBuffer.GetLength();

	if((nEntryFirstCharPos + strEntryName.GetLength() + 1) == nEndOfLine)
		return false;

	return true;
}